/************************************************************
thorse  ibmer       1415 2006-06-13 17:37 run
-rwxrwxr
 * HMMER - Biological sequence analysis with profile HMMs
 * Copyright (C) 1992-2003 Washington University School of Medicine
 * All Rights Reserved
 *
 *     This source code is distributed under the terms of the
 *     GNU General Public License. See the files COPYING and LICENSE
 *     for details.
 ************************************************************/

/* hmmpfam.c
 * SRE, Mon Aug 25 17:03:14 1997 [Denver]
 *
 * Search a single sequence against an HMM database.
 * Conditionally includes PVM parallelization when HMMER_PVM is defined
 *    at compile time; hmmpfam --pvm runs the PVM version.
 *
 * CVS $Id: hmmpfam.c,v 1.2 2008/03/16 14:52:26 cpsosa Exp $
 */

#define HMMER_MPI

#define cluster 8     // size of autonomous groups

#include "config.h"		/* compile-time configuration constants */
#include "squidconf.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <float.h>
#ifdef HMMER_THREADS
#include <pthread.h>
#endif
#ifdef HMMER_PVM
#include <pvm3.h>
#endif
#ifdef HMMER_MPI
#include <mpi.h>

//hitol is the number of hits a master needs before flushing the hit buffer back to the supermaster.
//countol is the number of sequences the master need to processes before checking if hitol has
//been reached.

#define hitol 64
#define countol 4096
#endif

#include "squid.h"		/* general sequence analysis library    */
#include "structs.h"		/* data structures, macros, #define's   */
#include "funcs.h"		/* function declarations                */
#include "globals.h"		/* alphabet global variables            */

const int TRACECOM = 8;
const int TRACECOM1 = 9;
const int TRACECOM2 = 10;
const int TRACECOM3 = 11;

static char banner[] = "hmmpfam - search one or more sequences against HMM database";

static char usage[]  = "\
Usage: hmmpfam [-options] <hmm database> <sequence file or database>\n\
  Available options are:\n\
   -h        : help; print brief help on version and usage\n\
   -n        : nucleic acid models/sequence (default protein)\n\
   -A <n>    : sets alignment output limit to <n> best domain alignments\n\
   -E <x>    : sets E value cutoff (globE) to <x>; default 10\n\
   -T <x>    : sets T bit threshold (globT) to <x>; no threshold by default\n\
   -Z <n>    : sets Z (# models) for E-value calculation\n\
";

static char experts[] = "\
   --acc         : use HMM accession numbers instead of names in output\n\
   --compat      : make best effort to use last version's output style\n\
   --cpu <n>     : run <n> threads in parallel (if threaded)\n\
   --cut_ga      : use Pfam GA gathering threshold cutoffs\n\
   --cut_nc      : use Pfam NC noise threshold cutoffs\n\
   --cut_tc      : use Pfam TC trusted threshold cutoffs\n\
   --domE <x>    : sets domain Eval cutoff (2nd threshold) to <x>\n\
   --domT <x>    : sets domain T bit thresh (2nd threshold) to <x>\n\
   --forward     : use the full Forward() algorithm instead of Viterbi\n\
   --informat <s>: sequence file is in format <s>\n\
   --null2       : turn OFF the post hoc second null model\n\
   --pvm         : run on a PVM (Parallel Virtual Machine) cluster\n\
   --mpi         : run in parallel using MPI (Message Passing Interface)\n\
   --xnu         : turn ON XNU filtering of query protein sequence\n\
\n";


static struct opt_s OPTIONS[] = {
  { "-h",        TRUE,  sqdARG_NONE },
  { "-n",        TRUE,  sqdARG_NONE },
  { "-A",        TRUE,  sqdARG_INT  },
  { "-E",        TRUE,  sqdARG_FLOAT},
  { "-T",        TRUE,  sqdARG_FLOAT},
  { "-Z",        TRUE,  sqdARG_INT  },
  { "--acc",     FALSE, sqdARG_NONE },
  { "--compat",  FALSE, sqdARG_NONE },
  { "--cpu",     FALSE, sqdARG_INT  },
  { "--cut_ga",  FALSE, sqdARG_NONE },
  { "--cut_nc",  FALSE, sqdARG_NONE },
  { "--cut_tc",  FALSE, sqdARG_NONE },
  { "--domE",    FALSE, sqdARG_FLOAT},
  { "--domT",    FALSE, sqdARG_FLOAT},
  { "--forward", FALSE, sqdARG_NONE },
  { "--informat",FALSE, sqdARG_STRING},
  { "--null2",   FALSE, sqdARG_NONE },
  { "--pvm",     FALSE, sqdARG_NONE },
  { "--mpi",     FALSE, sqdARG_NONE },
  { "--xnu",     FALSE, sqdARG_NONE },
};
#define NOPTIONS (sizeof(OPTIONS) / sizeof(struct opt_s))


static void main_loop_serial(char *hmmfile, HMMFILE *hmmfp, char *seq, SQINFO *sqinfo,
			     struct threshold_s *thresh, int do_xnu, int do_forward, int do_null2,
			     struct tophit_s *ghit, struct tophit_s *dhit, int *nhmm);
static void main_loop_pvm(char *hmmfile, HMMFILE *hmmfp, char *seq, SQINFO *sqinfo,
			  struct threshold_s *thresh, int do_xnu, int do_forward, int do_null2,
			  struct tophit_s *ghit, struct tophit_s *dhit, int *ret_nhmm);
static void main_loop_threaded(char *hmmfile, HMMFILE *hmmfp, char *seq, SQINFO *sqinfo,
			       struct threshold_s *thresh, int do_xnu,
			       int do_forward, int do_null2, int num_threads,
			       struct tophit_s *ghit, struct tophit_s *dhit, int *nhmm);
static void main_loop_mpi_master(char *hmmfile, HMMFILE *hmmfp, char *seq,
        SQINFO *sqinfo, struct threshold_s *thresh,
        int do_xnu, int do_forward, int do_null2, struct transporter pack,
        struct tophit_s *ghit,
        struct tophit_s *dhit, int *ret_nhmm, int num_nodes,
        MPI_Comm subcomm, MPI_Comm mastercomm, int real_rank);

static void main_loop_mpi_slave(char *hmmfile, HMMFILE *hmmfp, char *seq,
        struct threshold_s thresh, int do_xnu,
        int do_forward, int do_null2, struct transporter pack, int nhmm, int rank, char *index,
        MPI_Comm subcomm, MPI_Comm mastercomm, char *seqfile, int format);

static void main_loop_mpi(char *hmmfile, HMMFILE *hmmfp, char *seq,
        SQINFO *sqinfo, struct threshold_s *thresh,
        int do_xnu, int do_forward, int do_null2, struct transporter pack,
        struct tophit_s *ghit,
        struct tophit_s *dhit, int *ret_nhmm, int num_nodes,
        MPI_Comm mastercomm, MPI_Comm subcomm, FILE * indexfile, int num_mast, int off,
        int num_seq, char * index, char * seqfile, int format);

#ifdef HMMER_MPI

//Don't worry about the transporter, it's not important anymore.  It's simply
//a device for shuttling variables in between functions and is very
//inefficient.

struct transporter
{
    int be_backwards;
    int show_acc;
    int format;
    char *seqfile;
};

/*These functions are obsolete.

char* mpiRecvString(int rank, MPI_Comm comm, int tag1, int tag2)
{
    int len;
    char * buffer;
    MPI_Status temp;
    MPI_Status *useless;
    MPI_Recv(&len, 1, MPI_INT, rank, tag1, comm, useless);
    if(len==0) return NULL;
    buffer = malloc((len+4)*sizeof(char));
    MPI_Recv(buffer, len+1, MPI_CHAR, rank, tag2, comm, useless);
    buffer[len] = '\0';
    return buffer;
}

void mpiSendString(char* sendee, int rank, MPI_Comm comm, int tag1, int tag2)
{       
    int len = (sendee==NULL ? 0 : strlen(sendee));
    MPI_Send(&len, 1, MPI_INT, rank, tag1, comm);
    if(sendee==NULL) return;
    MPI_Send(sendee, len, MPI_CHAR, rank, tag2, comm);
}

*/

// mpiSendSeq() and mpiRecvSeq() are used for the master to update the current
// sequence on the worker nodes.
/* Not only are these functions inefficient, these should probably be put
   separate .h file.
*/

void mpiSendSeq(char * seq, SQINFO * sqinfo, int rank, MPI_Comm comm)
{       
    int len = sqinfo->len;
    MPI_Send(&len, 1, MPI_INT, rank, 990, comm);
    MPI_Send(seq, len, MPI_CHAR, rank, 991, comm);
    MPI_Send(sqinfo, sizeof(SQINFO), MPI_BYTE, rank, 992, comm);
}

void mpiRecvSeq(char ** seq, SQINFO * sqinfo, int rank, MPI_Comm comm)
{       
    int len;
    MPI_Status temp;
    MPI_Recv(&len, 1, MPI_INT, rank, 990, comm, &temp);
    *seq = malloc(len*sizeof(char));
    MPI_Recv(*seq, len, MPI_CHAR, rank, 991, comm, &temp);
    MPI_Recv(sqinfo, sizeof(SQINFO)+1, MPI_BYTE, rank, 992, comm, &temp);
}

int packInt(unsigned char * buffer, int num)
{
    memcpy(buffer, &num, sizeof(int));
    return sizeof(int);
}

int unpackInt(unsigned char * buffer, int *num)
{
    memcpy( num, buffer, sizeof(int));
    return sizeof(int);
}

int mpiPackString(char* buffer, char* string)
{
    if(string == NULL)
    {
        return packInt(buffer, 0);
    }
    else
    {
        int len = strlen(string);
        memcpy(buffer + packInt(buffer, len), string, len);
        return len+sizeof(int);
    }
}

int mpiUnpackString(char* buffer, char**string)
{
    int len;
    unpackInt(buffer, &len);
    if(len==0)
    {
        *string = NULL;
        return sizeof(int);
    }
    else
    {
        *string = malloc(sizeof(char)*(len+1));
        memcpy(*string, buffer+sizeof(int), len);
        (*string)[len]='\0';
        return len+sizeof(int);
    }
}

// These functions are no longer used.
/*
void mpiSendHMM(struct plan7_s *hmm, int rank, MPI_Comm comm)
{
    int k, M;
    int sendflags;		 // HMM flags to send

    char *buffer = malloc(sizeof(char)*1024*1024);

    sendflags = hmm->flags;
    sendflags &= ~PLAN7_HASBITS; // no log odds scores sent
    sendflags &= ~PLAN7_HASDNA;	 // no DNA scores sent 
    int count = 0;

    count += packInt(buffer, hmm->M);
    count += packInt(buffer + count, sendflags);

    count += mpiPackString(buffer + count, hmm->name);
    if(hmm->flags & PLAN7_DESC) count += mpiPackString(buffer+count, hmm->desc);
    if(hmm->flags & PLAN7_RF) count += mpiPackString(buffer+count, hmm->rf);
    if(hmm->flags & PLAN7_CS) count += mpiPackString(buffer+count, hmm->cs);
    count += mpiPackString(buffer+count, hmm->comlog);
    count += packInt(buffer+count, hmm->nseq);
    count += mpiPackString(buffer+count, hmm->ctime);
    if(hmm->flags & PLAN7_MAP)
    {
        memcpy(buffer+count, hmm->map, sizeof(int)*(hmm->M+1));
        count+=sizeof(int)*(hmm->M+1);
    }
    count += packInt(buffer+count, hmm->checksum);

    for(k=1; k<hmm->M; k++)
    {
        memcpy(buffer+count, hmm->t[k], 7*sizeof(float));
        count += sizeof(float)*7;
    }

    for(k=1; k<= hmm->M; k++)
    {
        memcpy(buffer+count, hmm->mat[k], Alphabet_size*sizeof(float));
        count+=Alphabet_size*sizeof(float);
    }

    for(k=1; k<hmm->M; k++)
    {
        memcpy(buffer+count,  hmm->ins[k], Alphabet_size*sizeof(float));
        count+=Alphabet_size*sizeof(float);
    }
    memcpy(buffer+count, (&(hmm->tbd1)), sizeof(float));
    count+=sizeof(float);
    for(k=0; k<4; k++)
    {
        memcpy(buffer+count, hmm->xt[k], 2*sizeof(float));
        count+=2*sizeof(float);
    }

    memcpy(buffer+count, hmm->begin, sizeof(float)*(hmm->M+1));
    count+=(hmm->M+1)*sizeof(float);
    memcpy(buffer+count, hmm->end, sizeof(float)*(hmm->M+1));
    count+=(hmm->M+1)*sizeof(float);
    memcpy(buffer+count, hmm->null, sizeof(float)*Alphabet_size);
    count+=Alphabet_size*sizeof(float);
    memcpy(buffer+count, (&(hmm->p1)), sizeof(float));
    count+=sizeof(float);
    if(hmm->flags & PLAN7_STATS)
    {
        memcpy(buffer+count, &(hmm->mu), sizeof(float));
        count+=sizeof(float);
        memcpy(buffer+count, &(hmm->lambda), sizeof(float));
        count+=sizeof(float);
    }

    
    MPI_Send(&count, 1, MPI_INT, rank, 997, comm);
    MPI_Send(buffer, count, MPI_BYTE, rank, 998, comm);

    free(buffer);
}

struct plan7_s * mpiRecvHMM(int rank, MPI_Comm comm)
{
    int M, i, k;
    int sendflags;

    struct plan7_s * hmm;
    int count1, count2=0;
    MPI_Status temp;
    
    char *buffer = malloc(sizeof(char)*1024*1024);
    
    MPI_Recv(&count1, 1, MPI_INT, rank, 997, comm, MPI_STATUS_IGNORE);
    MPI_Recv(buffer, count1, MPI_BYTE, rank, 998, comm, &temp);

    count2+=unpackInt(buffer, &M);
    hmm = AllocPlan7(M);
    count2+=unpackInt(buffer+count2, &sendflags);
    hmm->flags = sendflags;
   
    count2+= mpiUnpackString(buffer+count2, &(hmm->name));
    if (hmm->flags & PLAN7_DESC) count2+=mpiUnpackString(buffer+count2, &(hmm->desc));
    if (hmm->flags & PLAN7_RF) count2+=mpiUnpackString(buffer+count2, &(hmm->rf));
    if (hmm->flags & PLAN7_CS) count2+=mpiUnpackString(buffer+count2, &(hmm->cs));
    count2+=mpiUnpackString(buffer+count2, &(hmm->comlog));
    count2+=unpackInt(buffer+count2, &(hmm->nseq));
    count2+=mpiUnpackString(buffer+count2, &(hmm->ctime));
    if (hmm->flags & PLAN7_MAP) 
    {
        memcpy(hmm->map, buffer+count2, sizeof(int)*(M+1));
        count2+=sizeof(int)*(M+1);
    }
    
    count2+=unpackInt(buffer+count2, &(hmm->checksum));
    
    for (k = 1; k < M; k++)
    {
        memcpy(hmm->t[k], buffer+count2, 7*sizeof(float));
        count2+=7*sizeof(float);
    }
    for (k = 1; k <= M; k++)
    {
        memcpy(hmm->mat[k], buffer+count2, Alphabet_size*sizeof(float));
        count2+=Alphabet_size*sizeof(float);
    }

    for (k = 1; k < M; k++)
    {
        memcpy(hmm->ins[k], buffer+count2, Alphabet_size*sizeof(float));
        count2+=Alphabet_size*sizeof(float);
    }

    memcpy(&(hmm->tbd1), buffer+count2, sizeof(float));
    count2+=sizeof(float);
    for (k = 0; k < 4; k++)
    {
        memcpy(hmm->xt[k], buffer+count2, sizeof(float)*2);
        count2+=sizeof(float)*2;
    }

    memcpy(hmm->begin, buffer+count2, sizeof(float)*(M+1));
    count2+=sizeof(float)*(M+1);
    memcpy(hmm->end, buffer+count2, sizeof(float)*(M+1));
    count2+=sizeof(float)*(M+1);
    memcpy(hmm->null, buffer+count2, sizeof(float)*Alphabet_size);
    count2+=sizeof(float)*Alphabet_size;
    memcpy(&(hmm->p1), buffer+count2, sizeof(float));
    count2+=sizeof(float);
    if (hmm->flags & PLAN7_STATS)
    {
        memcpy(&(hmm->mu), buffer+count2, sizeof(float));
        count2+=sizeof(float);
        memcpy(&(hmm->lambda), buffer+count2, sizeof(float));
    }
    free(buffer);
    return hmm;
}
*/

int mpiSendTrace(struct p7trace_s *tr, int target, MPI_Comm comm)
{       
    if (MPI_Send(&(tr->tlen), 1, MPI_INT, target, 550, comm) != MPI_SUCCESS) return 0;
    if (MPI_Send(tr->statetype, tr->tlen, MPI_BYTE, target, 551, comm) != MPI_SUCCESS) return 0;
    if (MPI_Send(tr->nodeidx, tr->tlen, MPI_INT, target, 552, comm) !=MPI_SUCCESS) return 0;
    if (MPI_Send(tr->pos, tr->tlen, MPI_INT, target, 553, comm) !=MPI_SUCCESS) return 0;
    return 1;
}

struct p7trace_s * mpiRecvTrace(int target, MPI_Comm comm)
{ 	
    struct p7trace_s * tr;
    int tlen;
    MPI_Status temp;
    MPI_Status * useless = &temp;

    MPI_Recv(&tlen, 1, MPI_INT, target, 550, comm, useless);
    P7AllocTrace(tlen, &tr);
    if (MPI_Recv(tr->statetype, tlen, MPI_BYTE, target, 551, comm, useless) != MPI_SUCCESS) 
        { P7FreeTrace(tr); return NULL;}
    if (MPI_Recv(tr->nodeidx, tlen, MPI_INT, target, 552, comm, useless) != MPI_SUCCESS) 
        { P7FreeTrace(tr); return NULL;}
    if (MPI_Recv(tr->pos, tlen, MPI_INT, target, 553, comm, useless) != MPI_SUCCESS) 
        { P7FreeTrace(tr); return NULL;}
    tr->tlen = tlen;
    return tr;
}

/* These functions are no longer used, but could be used for experimentation
   You'll also have to uncomment the mpiRecvString and mpiSendString functions

//  mpiRecvAli()
//  Receives a fancyali_s struct from rank in comm.

struct fancyali_s * mpiRecvAli(int rank, MPI_Comm comm)
{       
    struct fancyali_s *ans;
    MPI_Status temp;
    ans = malloc(sizeof(struct fancyali_s));
    MPI_Recv(ans, sizeof(struct fancyali_s), MPI_BYTE, rank, 119, comm, &temp);
    ans->rfline = mpiRecvString(rank, comm, 120, 121);
    ans->csline = mpiRecvString(rank, comm, 122, 123);
    ans->model = mpiRecvString(rank, comm, 124, 125);
    ans->mline = mpiRecvString(rank, comm, 126, 127);
    ans->query = mpiRecvString(rank, comm, 128, 129);
    ans->target = mpiRecvString(rank, comm, 130, 131);
    ans->aseq = mpiRecvString(rank, comm, 200, 201);
    return ans;
}

//  mpiRecvAli()
//  Receives a fancyali_s struct from rank in comm.

void mpiSendAli(struct fancyali_s * ans, int rank, MPI_Comm comm)
{        
    MPI_Send(ans, sizeof(struct fancyali_s), MPI_BYTE, rank, 119, comm);
    mpiSendString(ans->rfline, rank, comm, 120, 121);
    mpiSendString(ans->csline, rank, comm, 122, 123);
    mpiSendString(ans->model, rank, comm, 124, 125);
    mpiSendString(ans->mline, rank, comm, 126, 127);
    mpiSendString(ans->query, rank, comm, 128, 129);
    mpiSendString(ans->target, rank, comm, 130, 131);
    mpiSendString(ans->aseq, rank, comm, 200, 201);
}

void mpiRecvHit(struct hit_s * hit, int rank, MPI_Comm comm)
{       
    fprintf(stderr, "");     //for some reason this is necessary, don't ask me
    hit->name = mpiRecvString(rank, comm, 113, 114);
    hit->acc = mpiRecvString(rank, comm, 115, 116);
    hit->desc = mpiRecvString(rank, comm, 117, 118);
    if((hit->ali)!=NULL)
    {
        fprintf(stderr, "");   //this too
        hit->ali = mpiRecvAli(rank, comm);
    }
}

void mpiSendHit(struct hit_s * hit, int rank, MPI_Comm comm)
{       
    mpiSendString(hit->name, rank, comm, 113, 114);
    mpiSendString(hit->acc, rank, comm, 115, 116);
    mpiSendString(hit->desc, rank, comm, 117, 118);
    if((hit->ali)!=NULL)
        mpiSendAli(hit->ali, rank, comm);
}
*/

void mpiPackHits(struct hit_s * hit, int num, int rank, MPI_Comm comm)
{
    char * buffer = (char *) malloc(1024*1024*10);
    int count = 0;
    
    int temp;
    MPI_Comm_rank(MPI_COMM_WORLD, &temp);
    for(int i=0; i<num; i++)
    {
        int len;
        count+=mpiPackString(buffer+count, (hit+i)->name);
        count+=mpiPackString(buffer+count, (hit+i)->acc);
        count+=mpiPackString(buffer+count, (hit+i)->desc);
        if((hit+i)->ali == NULL)
        {
            count+=packInt(buffer+count, 0);
        }
        else
        {
            if((hit+i)->ali->rfline==NULL)
            {
                count+=packInt(buffer+count, -1);
            }
            else
            {
                len = strlen((hit+i)->ali->rfline);
                count+=packInt(buffer+count, len);
                memcpy(buffer+count, (hit+i)->ali->rfline, len);
                count+=len;
            }
            count+=mpiPackString(buffer+count, (hit+i)->ali->csline);
            count+=mpiPackString(buffer+count, (hit+i)->ali->model);
            count+=mpiPackString(buffer+count, (hit+i)->ali->mline);
            count+=mpiPackString(buffer+count, (hit+i)->ali->aseq);
            count+=mpiPackString(buffer+count, (hit+i)->ali->query);
            count+=mpiPackString(buffer+count, (hit+i)->ali->target);
            count+=packInt(buffer+count, (hit+i)->ali->len);
            count+=packInt(buffer+count, (hit+i)->ali->sqfrom);
            count+=packInt(buffer+count, (hit+i)->ali->sqto);
        }
    }

    MPI_Send(&count, 1, MPI_INT, rank, 113, comm);
    MPI_Send(buffer, count, MPI_BYTE, rank, 114, comm);

    free(buffer);
}

void mpiUnpackHits(struct hit_s * hit, int num, int rank, MPI_Comm comm)
{
    int count;

    MPI_Recv(&count, 1, MPI_INT, rank, 113, comm, MPI_STATUS_IGNORE);

    char * buffer = (char *) malloc(count);
    MPI_Status temp;
    MPI_Recv(buffer, count, MPI_BYTE, rank, 114, comm, &temp);
    count = 0;

    for(int i=0; i<num; i++)
    {
        int len;
        count+=mpiUnpackString(buffer+count, &((hit+i)->name));
        count+=mpiUnpackString(buffer+count, &((hit+i)->acc));
        count+=mpiUnpackString(buffer+count, &((hit+i)->desc));
        count+=unpackInt(buffer+count, &len);
        if(len==0)
            (hit+i)->ali = NULL;
        else
        {
            (hit+i)->ali = (struct fancyali_s *) malloc(sizeof(struct fancyali_s));
            if(len == -1)
            {
                (hit+i)->ali->rfline = NULL;
            }
            else
            {
                (hit+i)->ali->rfline = (char *) malloc(sizeof(char)*(len+1));
                memcpy((hit+i)->ali->rfline, buffer+count, len);
                ((hit+i)->ali->rfline)[len] = '\0';
                count+=len;
            }
            count+=mpiUnpackString(buffer+count, &((hit+i)->ali->csline));
            count+=mpiUnpackString(buffer+count, &((hit+i)->ali->model));
            count+=mpiUnpackString(buffer+count, &((hit+i)->ali->mline));
            count+=mpiUnpackString(buffer+count, &((hit+i)->ali->aseq));
            count+=mpiUnpackString(buffer+count, &((hit+i)->ali->query));
            count+=mpiUnpackString(buffer+count, &((hit+i)->ali->target));
            count+=unpackInt(buffer+count, &((hit+i)->ali->len));
            count+=unpackInt(buffer+count, &((hit+i)->ali->sqfrom));
            count+=unpackInt(buffer+count, &((hit+i)->ali->sqto));
        }
    }
    free(buffer);
}
#endif

#ifdef HMMER_THREADS
/* POSIX threads version:
 * the threads share a workpool_s structure amongst themselves,
 * for obtaining locks on input HMM file and output histogram and
 * tophits structures.
 */
struct workpool_s {
    /* Shared configuration resources that don't change:
     */
    char          *hmmfile;       /* name of HMM file                */
    unsigned char *dsq;           /* digitized query sequence        */
    char  *seqname;               /* sequence name                   */
    int    L;			/* length of dsq                   */
    int    do_forward;		/* TRUE to score using Forward     */
    int    do_null2;		/* TRUE to apply null2 correction  */
    struct threshold_s *thresh;   /* score/evalue cutoff information */

    /* Shared (mutex-protected) input resources:
     */
    HMMFILE *hmmfp;               /* ptr to open HMM file           */
    int nhmm;			/* number of HMMs searched so far */
    pthread_mutex_t input_lock;   /* mutex for locking input        */

    /* Shared (mutex-protected) output resources:
     */
    struct tophit_s *ghit;        /* per-sequence top hits */
    struct tophit_s *dhit;        /* per-domain top hits */
    pthread_mutex_t output_lock;  /* mutex for locking output */

    /* Thread pool information
     */
    pthread_t *thread;            /* our pool of threads */
    int        num_threads;       /* number of threads   */
};

static struct workpool_s *workpool_start(char *hmmfile, HMMFILE *hmmfp,
        unsigned char *dsq, char *seqname, int L,
        int do_forward, int do_null2,
        struct threshold_s *thresh,
        struct tophit_s *ghit, struct tophit_s *dhit,
        int num_threads);
static void  workpool_stop(struct workpool_s *wpool);
static void  workpool_free(struct workpool_s *wpool);
static void *worker_thread(void *ptr);
#endif /* HMMER_THREADS */



int
main(int argc, char **argv)
{
    char            *hmmfile;	/* file to read HMMs from                  */
    HMMFILE         *hmmfp=NULL;       /* opened hmmfile for reading              */
    char            *seqfile;     /* file to read target sequence from       */
    SQFILE          *sqfp;        /* opened seqfile for reading              */
    int              format;	/* format of seqfile                       */
    char              *seq;	/* target sequence                         */
    SQINFO             sqinfo;	/* optional info for seq                   */
    struct fancyali_s *ali;	/* an alignment for display                */
    struct tophit_s   *ghit;      /* list of top hits and alignments for seq  */
    struct tophit_s   *dhit;	/* list of top hits/alignments for domains  */

    float   sc;			/* log-odds score in bits                  */
    double  pvalue;		/* pvalue of an HMM score                  */
    double  evalue;		/* evalue of an HMM score                  */
    double  motherp;		/* pvalue of a whole seq HMM score         */
    float   mothersc;		/* score of a whole seq parent of domain   */
    int     sqfrom, sqto;		/* coordinates in sequence                 */
    int     hmmfrom, hmmto;	/* coordinate in HMM                       */
    char   *name, *acc, *desc;    /* hit HMM name, accession, description            */
    int     hmmlen;		/* length of HMM hit                       */
    int     nhmm;			/* number of HMMs searched                 */
    int     domidx;		/* number of this domain                   */
    int     ndom;			/* total # of domains in this seq          */
    int     namewidth;		/* max width of printed HMM name           */
    int     descwidth;		/* max width of printed description        */

    int    Alimit;		/* A parameter limiting output alignments   */
    struct threshold_s thresh;    /* contains all threshold (cutoff) info     */

    char *optname;                /* name of option found by Getopt()         */
    char *optarg;                 /* argument found by Getopt()               */
    int   optind;                 /* index in argv[]                          */
    int   do_forward;		/* TRUE to use Forward() not Viterbi()      */
    int   do_nucleic;		/* TRUE to do DNA/RNA instead of protein    */
    int   do_null2;		/* TRUE to adjust scores with null model #2 */
    int   do_pvm;			/* TRUE to run on PVM                       */
    int   do_xnu;			/* TRUE to do XNU filtering                 */
    int   do_mpi;
    int   be_backwards;		/* TRUE to be backwards-compatible in output*/
    int   show_acc;		/* TRUE to sub HMM accessions for names     */
    int   i;
    int   nreported;

    int   num_threads;            /* number of worker threads */
    int   threads_support;	/* TRUE if threads support compiled in */
    int   pvm_support;		/* TRUE if PVM support compiled in     */
    int   mpi_support;
    int   num_nodes;
    int   rank, real_num;
    char * index;
    FILE * indexfile;
    int ranks[][3] = {{0,0,1},{0,0,1}};
    MPI_Comm subcomm, mastercomm;
    MPI_Group wgroup, subgroup, empgroup, mastergroup;

    /***********************************************
     * Parse command line
     ***********************************************/

    format      = SQFILE_UNKNOWN;	/* default: autodetect format w/ Babelfish */
    do_forward  = FALSE;
    do_nucleic  = FALSE;
    do_null2    = TRUE;
    do_pvm      = FALSE;
    do_xnu      = FALSE;
    do_mpi      = FALSE;
    be_backwards= FALSE;
    show_acc    = FALSE;

    pvm_support     = FALSE;
    threads_support = FALSE;
    mpi_support     = FALSE;
    num_threads     = 0;
    rank            = 0;
#ifdef HMMER_THREADS
    num_threads     = ThreadNumber();
    threads_support = TRUE;
#endif
#ifdef HMMER_PVM
    pvm_support     = TRUE;
#endif
#ifdef HMMER_MPI
    mpi_support     = TRUE;
    MPI_Init(&argc, &argv);
    MPI_Comm_size(MPI_COMM_WORLD, &num_nodes);
    MPI_Comm_rank(MPI_COMM_WORLD, &rank);

    MPI_Comm_group(MPI_COMM_WORLD, &wgroup);

    ranks[0][0] = rank - rank%cluster;
    if(num_nodes - ranks[0][0] > cluster -1)
        ranks[0][1] = ranks[0][0] + cluster -1;
    else ranks[0][1] = num_nodes-1;

    real_num = ranks[0][1] - ranks[0][0] + 1;

    if(num_nodes>1)
    {
        MPI_Group_range_incl(wgroup, 1, ranks, &subgroup);
        MPI_Comm_create(MPI_COMM_WORLD, subgroup, &subcomm); //communicator 
                                              //containg master-slave group

        ranks[1][0] = 1;
        ranks[1][1] = num_nodes;
        ranks[1][2] = cluster;
        ranks[0][0] = 0;
        ranks[0][1] = 0;
        ranks[0][2] = 1;

        MPI_Group_range_incl(wgroup, 2, ranks, &mastergroup);
        MPI_Comm_create(MPI_COMM_WORLD, mastergroup, &mastercomm);
        //communicator containing all masters
   }


#endif

    Alimit         = INT_MAX;	/* no limit on alignment output     */
    thresh.globE   = 10.0;	/* use a reasonable Eval threshold; */
    thresh.globT   = -FLT_MAX;	/*   but no bit threshold,          */
    thresh.domT    = -FLT_MAX;	/*   no domain bit threshold,       */
    thresh.domE    = FLT_MAX;     /*   and no domain Eval threshold.  */
    thresh.autocut = CUT_NONE;	/*   and no Pfam cutoffs used.      */
    thresh.Z       = 0;		/* Z not preset, so determined by # of HMMs */
    while (Getopt(argc, argv, OPTIONS, NOPTIONS, usage,
                &optind, &optname, &optarg))  {
        if      (strcmp(optname, "-n")        == 0) do_nucleic     = TRUE;
        else if (strcmp(optname, "-A")        == 0) Alimit         = atoi(optarg);
        else if (strcmp(optname, "-E")        == 0) thresh.globE   = atof(optarg);
        else if (strcmp(optname, "-T")        == 0) thresh.globT   = atof(optarg);
        else if (strcmp(optname, "-Z")        == 0) thresh.Z       = atoi(optarg);
        else if (strcmp(optname, "--acc")     == 0) show_acc       = TRUE;
        else if (strcmp(optname, "--compat")  == 0) be_backwards   = TRUE;
        else if (strcmp(optname, "--cpu")     == 0) num_threads    = atoi(optarg);
        else if (strcmp(optname, "--cut_ga")  == 0) thresh.autocut = CUT_GA;
        else if (strcmp(optname, "--cut_nc")  == 0) thresh.autocut = CUT_NC;
        else if (strcmp(optname, "--cut_tc")  == 0) thresh.autocut = CUT_TC;
        else if (strcmp(optname, "--domE")    == 0) thresh.domE    = atof(optarg);
        else if (strcmp(optname, "--domT")    == 0) thresh.domT    = atof(optarg);
        else if (strcmp(optname, "--forward") == 0) do_forward     = TRUE;
        else if (strcmp(optname, "--null2")   == 0) do_null2       = FALSE;
        else if (strcmp(optname, "--pvm")     == 0) do_pvm         = TRUE;
        else if (strcmp(optname, "--xnu")     == 0) do_xnu         = TRUE;
        else if (strcmp(optname, "--mpi")     == 0) do_mpi         = TRUE;
        else if (strcmp(optname, "--informat") == 0) {
            format = String2SeqfileFormat(optarg);
            if (format == SQFILE_UNKNOWN && rank==0)
                Die("unrecognized sequence file format \"%s\"", optarg);
        }
        else if (strcmp(optname, "-h")      == 0 && rank==0) {
            HMMERBanner(stdout, banner);
            puts(usage);
            puts(experts);
            exit(0);
        }
        if (argc - optind != 2 && rank==0)
            Die("Incorrect number of arguments.\n%s\n", usage);
    }
    hmmfile = argv[optind++];
    seqfile = argv[optind++];

    if(rank==0)
    {   if (do_pvm && ! pvm_support)
            Die("PVM support is not compiled into HMMER; --pvm doesn't work.");
        if (num_threads && ! threads_support)
            Die("Posix threads support is not compiled into HMMER; --cpu doesn't have any effect");
        if (do_mpi && ! mpi_support)
            Die("MPI support is not compiled into HMMER; --mpi doesn't work.");

        /* Try to work around inability to autodetect from a pipe or .gz:
         * assume FASTA format
         */
    }
    if (format == SQFILE_UNKNOWN &&
            (Strparse("^.*\\.gz$", seqfile, 0) || strcmp(seqfile, "-") == 0))
        format = SQFILE_FASTA;

        /***********************************************
         * Open sequence database (must be in curr directory);
         * get target sequence.
         ***********************************************/
    double start_time;
    if(rank==0)
    {
        if (do_nucleic) SetAlphabet(hmmNUCLEIC);
        else            SetAlphabet(hmmAMINO);
        printf("Setting start time\n");
        start_time = MPI_Wtime();
    }

    index = malloc((sizeof(char))*(strlen(seqfile)+3));
    i=strlen(seqfile);
    strcpy(index, seqfile);
    index[i]='.';
    index[i+1]='i';
    index[i+2]='\0';

    if(rank==0)
    {
        if (do_nucleic && do_xnu)
            Die("You can't use -n and --xnu together: I can't xnu DNA data.");

        if ((sqfp = SeqfileOpen(seqfile, format, NULL)) == NULL)
            Die("Failed to open sequence file %s\n%s\n", seqfile, usage);
#ifdef HMMER_MPI
        if((indexfile = fopen(index, "r"))==NULL)
        {   printf("Creating index file: %s, this will only be done once:\n", index);
            printf("|-----------------------------------------------|\n");
            indexfile = fopen(index, "w");
            printf("Done open\n");
            fprintf(indexfile, "            \n");
            i=0;
            fprintf(indexfile, "%11d\n", ftell(sqfp->f));
            while (ReadSeq(sqfp, sqfp->format, &seq, &sqinfo))
            {   
                fprintf(indexfile, "%11d\n", ftell(sqfp->f));
                i++;
                FreeSequence(seq, &sqinfo);
            }
            rewind(indexfile);
            fprintf(indexfile, "%11d", i);
            fclose(indexfile);
        }
        indexfile = fopen(index, "r");

#endif
         /***********************************************
         * Open HMM database (might be in HMMERDB or current directory)
         ***********************************************/
    }
    if(rank==0 || rank%cluster==1)
        if ((hmmfp = HMMFileOpen(hmmfile, "HMMERDB")) == NULL)
            Die("Failed to open HMM database %s\n%s", hmmfile, usage);
    if(rank==0)
    {
        /***********************************************
         * Show the banner
         ***********************************************/

        HMMERBanner(stdout, banner);
        printf(   "HMM file:                 %s\n", hmmfile);
        printf(   "Sequence file:            %s\n", seqfile);
        if (do_pvm)
            printf( "PVM:                      ACTIVE\n");
        if (do_mpi)
            printf( "MPI:                      ACTIVE\n");
        printf("- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -\n");
    }

    if (mpi_support && do_mpi)
    {
         struct transporter pack;
         pack.be_backwards = be_backwards;
         pack.show_acc = show_acc;
         pack.format = format;
         pack.seqfile = seqfile;
         if(rank==0)
         {
             main_loop_mpi(hmmfile, hmmfp, seq, &sqinfo, &thresh,
                do_xnu, do_forward, do_null2, pack, ghit, dhit, &nhmm,
                num_nodes, mastercomm, subcomm, indexfile, 0, 0, 0, index,
                seqfile, format);
             printf("OK\n");
         }
         else if(rank%cluster==1)
         {
             main_loop_mpi_master(hmmfile, hmmfp, seq, &sqinfo, &thresh,
                do_xnu, do_forward, do_null2, pack, ghit, dhit, &nhmm,
                num_nodes, subcomm, mastercomm, rank);
         }
         else
         {
            main_loop_mpi_slave(hmmfile, hmmfp, seq, thresh, do_xnu, do_forward,
                do_null2, pack, nhmm, rank, index, subcomm, mastercomm, seqfile, format);
        }
        MPI_Barrier(MPI_COMM_WORLD);
        if(rank==0)
        {
            printf("Finished, time is %lf\n", MPI_Wtime() - start_time);
        }
         MPI_Finalize();
         return 0;

    }
        /***********************************************
         * Search each HMM against each sequence
         ***********************************************/
    else if(rank==0)
    {
        while (ReadSeq(sqfp, format, &seq, &sqinfo))
        {
            ghit = AllocTophits(20);   /* keeps full seq scores */
            dhit = AllocTophits(20);   /* keeps domain scores   */

            /* 1. Search sequence against all HMMs.
             *    Significant scores+alignments accumulate in ghit, dhit.
             */
            if (pvm_support && do_pvm)
                main_loop_pvm(hmmfile, hmmfp, seq, &sqinfo,
                        &thresh, do_xnu, do_forward, do_null2,
                        ghit, dhit, &nhmm);
            else if (threads_support && num_threads)
                main_loop_threaded(hmmfile, hmmfp, seq, &sqinfo,
                        &thresh, do_xnu, do_forward, do_null2, num_threads,
                        ghit, dhit, &nhmm);
            else
                main_loop_serial(hmmfile, hmmfp, seq, &sqinfo,
                        &thresh, do_xnu, do_forward, do_null2,
                        ghit, dhit, &nhmm);
            
            HMMFileRewind(hmmfp);

            /* set Z for good now that we're done */
            if (!thresh.Z) thresh.Z = nhmm;

            /* 2. (Done searching all HMMs for this query seq; start output)
             *    Report the overall sequence hits, sorted by significance.
             */
            if (be_backwards)
            {
                printf("Query:  %s  %s\n", sqinfo.name,
                        sqinfo.flags & SQINFO_DESC ? sqinfo.desc : "");
            }
            else
            {
                printf("\nQuery sequence: %s\n", sqinfo.name);
                printf("Accession:      %s\n", sqinfo.flags &SQINFO_ACC ? sqinfo.acc  : "[none]");
                printf("Description:    %s\n", sqinfo.flags &SQINFO_DESC? sqinfo.desc : "[none]");
            }
            /* We'll now sort the global hit list by evalue...
             * (not score! that was bug #12. in hmmpfam, score and evalue are not
             *  monotonic.)
             */
            FullSortTophits(ghit);
            namewidth = MAX(8, TophitsMaxName(ghit)); /* must print whole name, no truncation  */
            descwidth = MAX(52-namewidth, 11);        /* may truncate desc, but avoid neg len! */

            printf("\nScores for sequence family classification (score includes all domains):\n");
            printf("%-*s %-*s %7s %10s %3s\n", namewidth, "Model", descwidth, "Description", "Score", "E-value", " N ");
            printf("%-*s %-*s %7s %10s %3s\n", namewidth, "--------", descwidth, "-----------", "-----", "-------", "---");
            for (i = 0, nreported = 0; i < ghit->num; i++)
            {
                char *safedesc;
                GetRankedHit(ghit, i,
                        &pvalue, &sc, NULL, NULL,
                        &name, &acc, &desc,
                        NULL, NULL, NULL,           /* seq positions */
                        NULL, NULL, NULL,           /* HMM positions */
                        NULL, &ndom,                /* domain info   */
                        NULL);	                   /* alignment info*/

                evalue = pvalue * (double) thresh.Z;

                /* safedesc is a workaround for an apparent Linux printf()
                 * bug with the *.*s format. dbmalloc crashes with a memchr() ptr out of bounds
                 * flaw if the malloc'ed space for desc is short. The workaround
                 * is to make sure the ptr for *.* has a big malloc space.
                 */
                if (desc != NULL && strlen(desc) < 80)
                {
                    safedesc = MallocOrDie(sizeof(char) * 80);
                    strcpy(safedesc, desc);
                }
                else safedesc = Strdup(desc);

                /* sneaky trick warning:
                 * if we're using dynamic Pfam score cutoffs (GA, TC, NC),
                 * then the list of hits is already correct and does not
                 * need any score cutoffs. Unset the thresholds. They'll
                 * be reset in the main_loop if we still have sequences
                 * to process.
                 */
                if (thresh.autocut != CUT_NONE) {
                    thresh.globE = thresh.domE = FLT_MAX;
                    thresh.globT = thresh.domT = -FLT_MAX;
                }

                if (evalue <= thresh.globE && sc >= thresh.globT)
                {
                    printf("%-*s %-*.*s %7.1f %10.2g %3d\n",
                            namewidth,
                            (show_acc && acc != NULL) ?  acc : name,
                            descwidth, descwidth, safedesc != NULL ? safedesc : "",
                            sc, evalue, ndom);
                    nreported++;
                }
                free(safedesc);
            }
            if (nreported == 0) printf("\t[no hits above thresholds]\n");

            /* 3. Report domain hits (sorted on sqto coordinate)
             */
            FullSortTophits(dhit);
            namewidth = MAX(8, TophitsMaxName(dhit)); /* must print whole name, no truncation  */

            printf("\nParsed for domains:\n");
            printf("%-*s %7s %5s %5s    %5s %5s    %7s %8s\n",
                    namewidth, "Model", "Domain ", "seq-f", "seq-t", "hmm-f", "hmm-t", "score", "E-value");
            printf("%-*s %7s %5s %5s    %5s %5s    %7s %8s\n",
                    namewidth, "--------", "-------", "-----", "-----", "-----", "-----", "-----", "-------");

            for (i = 0, nreported = 0; i < dhit->num; i++)
            {
                GetRankedHit(dhit, i,
                        &pvalue, &sc, &motherp, &mothersc,
                        &name, &acc, NULL,
                        &sqfrom, &sqto, NULL,
                        &hmmfrom, &hmmto, &hmmlen,
                        &domidx, &ndom,
                        NULL);
                evalue = pvalue * (double) thresh.Z;

                /* Does the "mother" (complete) sequence satisfy global thresholds? */
                if (motherp * (double)thresh. Z > thresh.globE || mothersc < thresh.globT)
                    continue;
                else if (evalue <= thresh.domE && sc >= thresh.domT) {
                    printf("%-*s %3d/%-3d %5d %5d %c%c %5d %5d %c%c %7.1f %8.2g\n",
                            namewidth,
                            (show_acc && acc != NULL) ?  acc : name,
                            domidx, ndom,
                            sqfrom, sqto,
                            sqfrom == 1 ? '[' : '.', sqto == sqinfo.len ? ']' : '.',
                            hmmfrom, hmmto,
                            hmmfrom == 1 ? '[':'.',  hmmto == hmmlen ? ']' : '.',
                            sc, evalue);
                    nreported++;
                }
            }
            if (nreported == 0) printf("\t[no hits above thresholds]\n");


            /* 3. Alignment output, also by domain.
             *    dhits is already sorted and namewidth is set, from above code.
             *    Number of displayed alignments is limited by Alimit parameter;
             *    also by domE (evalue threshold), domT (score theshold).
             */
            if (Alimit != 0)
            {
                printf("\nAlignments of top-scoring domains:\n");
                for (i = 0, nreported = 0; i < dhit->num; i++)
                {
                    if (nreported == Alimit) break; /* limit to Alimit output alignments */
                    GetRankedHit(dhit, i,
                            &pvalue, &sc, &motherp, &mothersc,
                            &name, &acc, NULL,
                            &sqfrom, &sqto, NULL,              /* seq position info  */
                            &hmmfrom, &hmmto, &hmmlen,         /* HMM position info  */
                            &domidx, &ndom,                    /* domain info        */
                            &ali);	                      /* alignment info     */
                    evalue = pvalue * (double) thresh.Z;

                    if (motherp * (double) thresh.Z > thresh.globE || mothersc < thresh.globT)
                        continue;
                    else if (evalue <= thresh.domE && sc >= thresh.domT)
                    {
                        printf("%s: domain %d of %d, from %d to %d: score %.1f, E = %.2g\n",
                                (show_acc && acc != NULL) ?  acc : name,
                                domidx, ndom, sqfrom, sqto, sc, evalue);
                        PrintFancyAli(stdout, ali);
                        nreported++;
                    }
                }
                if (nreported == 0)      printf("\t[no hits above thresholds]\n");
                if (nreported == Alimit) printf("\t[output cut off at A = %d top alignments]\n", Alimit);
            }


            printf("//\n");
        }
        FreeSequence(seq, &sqinfo);
        FreeTophits(ghit);
        FreeTophits(dhit);
        HMMFileRewind(hmmfp);
    }
    /***********************************************
     * Clean-up and exit.
     ***********************************************/
    SeqfileClose(sqfp);
    HMMFileClose(hmmfp);
    SqdClean();
#ifdef HMMER_MPI
    MPI_Finalize();
#endif
    return 0;
}


/* Function: main_loop_serial()
 * Date:     SRE, Fri Aug  7 13:46:48 1998 [St. Louis]
 *
 * Purpose:  Search a sequence against an HMM database;
 *           main loop for the serial (non-PVM, non-threads)
 *           version.
 *
 *           On return, ghit and dhit contain info for all hits
 *           that satisfy the set thresholds. If an evalue
 *           cutoff is used at all, the lists will be overestimated --
 *           because the evalue will be underestimated until
 *           we know the final Z. (Thus the main program must recheck
 *           thresholds before printing any results.) If only
 *           score cutoffs are used, then the lists are correct,
 *           and may be printed exactly as they come (after
 *           appropriate sorting, anyway). This is especially
 *           important for dynamic thresholding using Pfam
 *           score cutoffs -- the main caller cannot afford to
 *           rescan the HMM file just to get the GA/TC/NC cutoffs
 *           back out for each HMM, and neither do I want to
 *           burn the space to store them as I make a pass thru
 *           Pfam.
 *
 * Args:     hmmfile - name of HMM file
 *           hmmfp   - open HMM file (and at start of file)
 *           dsq     - digitized sequence
 *           sqinfo  - ptr to SQINFO optional info for dsq
 *           thresh  - score/evalue threshold information
 *           do_xnu     - TRUE to apply XNU filter to sequence
 *           do_forward - TRUE to use Forward() scores
 *           do_null2   - TRUE to adjust scores w/ ad hoc null2 model
 *           num_threads- number of threads, if threaded
 *           ghit    - global hits list
 *           dhit    - domain hits list
 *           ret_nhmm    - number of HMMs searched.
 *
 * Returns:  (void)
 */
static void
main_loop_serial(char *hmmfile, HMMFILE *hmmfp, char *seq, SQINFO *sqinfo,
        struct threshold_s *thresh, int do_xnu, int do_forward, int do_null2,
        struct tophit_s *ghit, struct tophit_s *dhit, int *ret_nhmm)
{
    unsigned char     *dsq;       /* digitized sequence                      */
    int                nhmm;	/* number of HMMs searched                 */
    struct plan7_s    *hmm;       /* current HMM to search with              */
    struct p7trace_s  *tr;	/* traceback of alignment                  */
    struct dpmatrix_s *mx;        /* growable DP matrix                      */
    float   sc;                   /* an alignment score                      */
    double  pvalue;		/* pvalue of an HMM score                  */
    double  evalue;		/* evalue of an HMM score                  */

    /* Prepare sequence.
     */
    SQD_DPRINTF1(("main_loop_serial:\n"));

    dsq = DigitizeSequence(seq, sqinfo->len);
    if (do_xnu && Alphabet_type == hmmAMINO) XNU(dsq, sqinfo->len);

    /*
     * We'll create for at least N=300xM=300, and thus consume at least 1 MB,
     * regardless of RAMLIMIT -- this helps us avoid reallocating some weird
     * asymmetric matrices.
     *
     * We're growable in both M and N, because inside of P7SmallViterbi,
     * we're going to be calling P7Viterbi on subpieces that vary in size,
     * and for different models.
     */
    mx = CreatePlan7Matrix(300, 300, 25, 25);

    nhmm = 0;
    while (HMMFileRead(hmmfp, &hmm)) {
        if (hmm == NULL)
            Die("HMM file %s may be corrupt or in incorrect format; parse failed", hmmfile);
        P7Logoddsify(hmm, !(do_forward));
        SQD_DPRINTF1(("   ... working on HMM %s\n", hmm->name));
        SQD_DPRINTF1(("   ... mx is now M=%d by N=%d\n", mx->maxM, mx->maxN));

        if (! SetAutocuts(thresh, hmm))
            Die("HMM %s did not contain the GA, TC, or NC cutoffs you needed",
                    hmm->name);

        /* Score sequence, do alignment (Viterbi), recover trace
         */
        if (P7ViterbiSpaceOK(sqinfo->len, hmm->M, mx))
        {
            SQD_DPRINTF1(("   ... using normal P7Viterbi(); size ~%d MB\n",
                        P7ViterbiSize(sqinfo->len, hmm->M)));
            sc = P7Viterbi(dsq, sqinfo->len, hmm, mx, &tr);
        }
        else
        {
            SQD_DPRINTF1(("   ... using P7SmallViterbi(); size ~%d MB\n",
                        P7ViterbiSize(sqinfo->len, hmm->M)));
            sc = P7SmallViterbi(dsq, sqinfo->len, hmm, mx, &tr);
        }
        /* Implement do_forward; we'll override the whole_sc with a P7Forward()
         * calculation.
         * HMMER is so trace- (alignment-) dependent that this gets a bit hacky.
         * Some important implications:
         *   1) if --do_forward is selected, the domain (Viterbi) scores do not
         *      necessarily add up to the whole sequence (Forward) score.
         *   2) The implementation of null2 for a Forward score is undefined,
         *      since the null2 correction is trace-dependent. As a total hack,
         *      we use a null2 correction derived from the whole trace
         *      (which was the behavior of HMMER 2.1.1 and earlier, anyway).
         *      This could put the sum of domain scores and whole seq score even
         *      further in disagreement.
         *
         * Note that you can't move the Forward calculation into
         * PostprocessSignificantHit(). The Forward score will exceed the
         * Viterbi score, so you can't apply thresholds until you
         * know the Forward score. Also, since PostprocessSignificantHit()
         * is wrapped by a mutex in the threaded implementation,
         * you'd destroy all useful parallelism if PostprocessSignificantHit()
         * did anything compute intensive.
         */
        if (do_forward) {
            sc = P7Forward(dsq, sqinfo->len, hmm, NULL);
            if (do_null2) sc -= TraceScoreCorrection(hmm, tr, dsq);
        }

        /* Store scores/pvalue for each HMM aligned to this sequence, overall
         */
        pvalue = PValue(hmm, sc);
        evalue = thresh->Z ? (double) thresh->Z * pvalue : (double) nhmm * pvalue;
        if (sc >= thresh->globT && evalue <= thresh->globE) {
            sc = PostprocessSignificantHit(ghit, dhit,
                    tr, hmm, dsq, sqinfo->len,
                    sqinfo->name, NULL, NULL, /* won't need acc or desc even if we have 'em */
                    do_forward, sc,
                    do_null2,
                    thresh,
                    TRUE); /* TRUE -> hmmpfam mode */
        }
        P7FreeTrace(tr);
        FreePlan7(hmm);
        nhmm++;
    }

    FreePlan7Matrix(mx);
    free(dsq);
    *ret_nhmm = nhmm;
    return;
}


#ifdef HMMER_PVM
/*****************************************************************
 * PVM specific functions
 ****************************************************************/

/* Function: main_loop_pvm()
 * Date:     SRE, Fri Aug  7 13:58:34 1998 [St. Louis]
 *
 * Purpose:  Search a sequence against an HMM database;
 *           main loop for the PVM version.
 *
 * Args:     hmmfile - name of HMM file
 *           hmmfp   - open HMM file (and at start of file)
 *           seq     - sequence to search against
 *           sqinfo  - ptr to SQINFO optional info for dsq
 *           thresh  - score/evalue threshold settings
 *           do_xnu     - TRUE to apply XNU filter to sequence
 *           do_forward - TRUE to use Forward() scores
 *           do_null2   - TRUE to adjust scores w/ ad hoc null2 model
 *           ghit    - global hits list
 *           dhit    - domain hits list
 *           nhmm    - number of HMMs searched.
 *
 * Returns:  (void)
 */
static void
main_loop_pvm(char *hmmfile, HMMFILE *hmmfp, char *seq, SQINFO *sqinfo,
        struct threshold_s *thresh, int do_xnu, int do_forward, int do_null2,
        struct tophit_s *ghit, struct tophit_s *dhit, int *ret_nhmm)
{
    struct plan7_s   *hmm;        /* HMM that was searched with */
    struct p7trace_s *tr;         /* a traceback structure */
    unsigned char    *dsq;        /* digitized sequence */
    float  sc;			/* score of an HMM match */
    int    master_tid;		/* master's ID */
    int   *slave_tid;             /* array of slave IDs */
    int   *hmmlist;               /* array of hmm indexes being worked on by slaves */
    int    nslaves;	       	/* number of slaves in virtual machine   */
    int    nhmm;			/* number of HMMs searched */
    int    slaveidx;		/* index of a slave wanting work         */
    int    slave, msg;
    int    sent_trace;		/* TRUE if slave sent us a trace */
    char   slavename[32];		/* name of HMM that slave actually did */
    double pvalue;		/* pvalue of HMM score */
    int    arglen;

    /* Sanity checks.
     */
    if (hmmfp->ssi == NULL)
        Die("HMM file %s needs an SSI index to use PVM. See: hmmindex.", hmmfile);

    /* Prepare sequence.
     */
    dsq = DigitizeSequence(seq, sqinfo->len);
    if (do_xnu && Alphabet_type == hmmAMINO) XNU(dsq, sqinfo->len);

    /* Initialize PVM
     */
    master_tid = pvm_mytid();
#if DEBUGLEVEL >= 1
    pvm_catchout(stderr);		/* catch output for debugging */
#endif
    SQD_DPRINTF1(("Spawning slaves...\n"));
    PVMSpawnSlaves("hmmpfam-pvm", &slave_tid, &nslaves);
    hmmlist   = MallocOrDie(sizeof(int) * nslaves);
    SQD_DPRINTF1(("Spawned a total of %d slaves...\n", nslaves));

    /* Initialize the slaves
     */
    SQD_DPRINTF1(("Broadcasting to %d slaves...\n", nslaves));
    pvm_initsend(PvmDataDefault);
    arglen = strlen(hmmfile);
    pvm_pkint(&arglen, 1, 1);
    pvm_pkstr(hmmfile);
    pvm_pkint(&(sqinfo->len), 1, 1);
    pvm_pkstr(seq);
    pvm_pkfloat(&(thresh->globT), 1, 1);
    pvm_pkdouble(&(thresh->globE), 1, 1);
    pvm_pkint(&(thresh->Z), 1, 1);
    pvm_pkint((int *)&(thresh->autocut), 1, 1);
    pvm_pkint(&do_xnu, 1, 1);
    pvm_pkint(&do_forward, 1, 1);
    pvm_pkint(&do_null2, 1, 1);
    pvm_pkint(&Alphabet_type, 1, 1);
    pvm_mcast(slave_tid, nslaves, HMMPVM_INIT);
    SQD_DPRINTF1(("Slaves should be ready...\n"));
    /* get their OK codes. */
    PVMConfirmSlaves(slave_tid, nslaves);
    SQD_DPRINTF1(("Slaves confirm that they're ok...\n"));

    /* Load the slaves.
     * For efficiency reasons, we don't want the master to
     * load HMMs from disk until she absolutely needs them.
     */
    for (nhmm = 0; nhmm < nslaves && nhmm < hmmfp->ssi->nprimary; nhmm++) {
        pvm_initsend(PvmDataDefault);
        pvm_pkint(&nhmm, 1, 1);	/* side effect: also tells him what number he is. */
        pvm_send(slave_tid[nhmm], HMMPVM_WORK);
        hmmlist[nhmm] = nhmm;
    }
    SQD_DPRINTF1(("%d slaves are loaded\n", nhmm));


    /* Receive/send loop
     */
    for (; nhmm < hmmfp->ssi->nprimary; nhmm++)
    {
        /* check slaves before blocking */
        PVMCheckSlaves(slave_tid, nslaves);
        /* receive output */
        SQD_DPRINTF1(("Waiting for a slave to give me output...\n"));
        pvm_recv(-1, HMMPVM_RESULTS);
        pvm_upkint(&slaveidx, 1, 1);     /* # of slave who's sending us stuff */
        pvm_upkstr(slavename);           /* name of HMM that slave did */
        pvm_upkfloat(&sc, 1, 1);         /* score   */
        pvm_upkdouble(&pvalue, 1, 1);    /* P-value */
        pvm_upkint(&sent_trace, 1, 1);   /* TRUE if trace is coming */
        tr = (sent_trace) ? PVMUnpackTrace() : NULL;
        SQD_DPRINTF1(("Slave %d finished %s for me...\n", slaveidx, slavename));

        /* send new work */
        pvm_initsend(PvmDataDefault);
        pvm_pkint(&nhmm, 1, 1);
        pvm_send(slave_tid[slaveidx], HMMPVM_WORK);
        SQD_DPRINTF1(("Assigned %d -> slave %d\n", nhmm, slaveidx));

        /* process output */
        /* 1b. Store scores/pvalue for each HMM aligned to this sequence, overall
         */
        SQD_DPRINTF1(("%15s : %2d : %f\n", slavename, slaveidx, sc));
        if (sent_trace)
        {
            /* now load the HMM, because the hit is significant */
            HMMFilePositionByIndex(hmmfp, hmmlist[slaveidx]);
            if (!HMMFileRead(hmmfp, &hmm))
            { Die("Unexpected failure to read HMM file %s", hmmfile); }
            if (hmm == NULL)
            { Die("HMM file %s may be corrupt; parse failed", hmmfile); }
            P7Logoddsify(hmm, TRUE);
            if (! SetAutocuts(thresh, hmm))
                Die("HMM %s did not contain your GA, NC, or TC cutoffs", hmm->name);

            sc = PostprocessSignificantHit(ghit, dhit,
                    tr, hmm, dsq, sqinfo->len,
                    sqinfo->name,
                    sqinfo->flags & SQINFO_ACC ? sqinfo->acc : NULL,
                    sqinfo->flags & SQINFO_DESC ? sqinfo->desc : NULL,
                    do_forward, sc,
                    do_null2,
                    thresh,
                    TRUE); /* TRUE -> hmmpfam mode */

            FreePlan7(hmm);
            P7FreeTrace(tr);
        }
        hmmlist[slaveidx] = nhmm;
    }

    /* Collect the output. all n slaves are still working, so wait for them.
     */
    for (slave = 0; slave < nslaves && slave < nhmm; slave++)
    {
        /* don't check slaves (they're exiting normally);
           window of vulnerability here to slave crashes */
        /* receive output */
        pvm_recv(-1, HMMPVM_RESULTS);
        pvm_upkint(&slaveidx, 1, 1);     /* slave who's sending us stuff */
        pvm_upkstr(slavename);
        pvm_upkfloat(&sc, 1, 1);         /* one score */
        pvm_upkdouble(&pvalue, 1, 1);	   /* P-value */
        pvm_upkint(&sent_trace, 1, 1);   /* TRUE if trace is coming */
        tr = (sent_trace) ? PVMUnpackTrace() : NULL;

        /* process output */
        SQD_DPRINTF1(("%15s : %2d : %f\n", slavename, slaveidx, sc));
        if (sent_trace)
        {
            /* now load the HMM, because the hit is significant */
            HMMFilePositionByIndex(hmmfp, hmmlist[slaveidx]);
            if (!HMMFileRead(hmmfp, &hmm))
            { Die("Unexpected failure to read HMM file %s", hmmfile);}
            if (hmm == NULL)
            { Die("HMM file %s may be corrupt; parse failed", hmmfile); }
            P7Logoddsify(hmm, TRUE);
            if (! SetAutocuts(thresh, hmm))
                Die("HMM %s did not contain your GA, NC, or TC cutoffs", hmm->name);

            sc = PostprocessSignificantHit(ghit, dhit,
                    tr, hmm, dsq, sqinfo->len,
                    sqinfo->name, NULL, NULL, /* won't need acc or desc even if we have 'em */
                    do_forward, sc,
                    do_null2,
                    thresh,
                    TRUE); /* TRUE -> hmmpfam mode */

            FreePlan7(hmm);
            P7FreeTrace(tr);
        }
    }

    /* Shut down all the slaves. (xref bug #15/STL5 p.66).
     */
    for (slave = 0; slave < nslaves; slave++)
    {
        pvm_initsend(PvmDataDefault);
        msg = -1;			/* the special "shut down now" flag */
        pvm_pkint(&msg, 1, 1);
        pvm_send(slave_tid[slave], HMMPVM_WORK);
    }

    /* Cleanup; quit the PVM; and return
     */
    free(slave_tid);
    free(hmmlist);
    free(dsq);

    *ret_nhmm = nhmm;
    return;
}
#else /* HMMER_PVM off; insert dummy stub function: */
static void
main_loop_pvm(char *hmmfile, HMMFILE *hmmfp, char *seq, SQINFO *sqinfo,
        struct threshold_s *thresh, int do_xnu, int do_forward, int do_null2,
        struct tophit_s *ghit, struct tophit_s *dhit, int *ret_nhmm)
{
    Die("no PVM support");
}
#endif /*HMMER_PVM*/


#ifdef HMMER_THREADS
/*****************************************************************
 * POSIX threads implementation.
 *
 * API:
 *      workpool_start()   (makes a workpool_s structure. Starts calculations.)
 *      workpool_stop()    (waits for threads to finish.)
 *      workpool_free()    (destroys the structure)
 *
 * Threads:
 *      worker_thread()    (the actual parallelized worker thread).
 *****************************************************************/
/* Function: main_loop_threaded()
 * Date:     SRE, Wed Feb 27 11:10:13 2002 [St. Louis]
 *
 * Purpose:  Search a sequence against an HMM database;
 *           main loop for the threaded version.
 *
 *           On return, ghit and dhit contain info for all hits
 *           that satisfy the set thresholds. If an evalue
 *           cutoff is used at all, the lists will be overestimated --
 *           because the evalue will be underestimated until
 *           we know the final Z. (Thus the main program must recheck
 *           thresholds before printing any results.) If only
 *           score cutoffs are used, then the lists are correct,
 *           and may be printed exactly as they come (after
 *           appropriate sorting, anyway). This is especially
 *           important for dynamic thresholding using Pfam
 *           score cutoffs -- the main caller cannot afford to
 *           rescan the HMM file just to get the GA/TC/NC cutoffs
 *           back out for each HMM, and neither do I want to
 *           burn the space to store them as I make a pass thru
 *           Pfam.
 *
 * Args:     hmmfile - name of HMM file
 *           hmmfp   - open HMM file (and at start of file)
 *           dsq     - digitized sequence
 *           sqinfo  - ptr to SQINFO optional info for dsq
 *           thresh  - score/evalue threshold information
 *           do_xnu     - TRUE to apply XNU filter to sequence
 *           do_forward - TRUE to use Forward() scores
 *           do_null2   - TRUE to adjust scores w/ ad hoc null2 model
 *           num_threads- number of threads, >= 1
 *           ghit    - global hits list
 *           dhit    - domain hits list
 *           ret_nhmm    - number of HMMs searched.
 *
 * Returns:  (void)
 */
static void
main_loop_threaded(char *hmmfile, HMMFILE *hmmfp, char *seq, SQINFO *sqinfo,
        struct threshold_s *thresh, int do_xnu, int do_forward, int do_null2,
        int num_threads,
        struct tophit_s *ghit, struct tophit_s *dhit, int *ret_nhmm)
{
    unsigned char     *dsq;       /* digitized sequence      */
    int                nhmm;	/* number of HMMs searched */
    struct workpool_s *wpool;     /* pool of worker threads  */

    /* Prepare sequence.
     */
    dsq = DigitizeSequence(seq, sqinfo->len);
    if (do_xnu && Alphabet_type == hmmAMINO) XNU(dsq, sqinfo->len);

    wpool = workpool_start(hmmfile, hmmfp, dsq, sqinfo->name, sqinfo->len,
            do_forward, do_null2, thresh,
            ghit, dhit, num_threads);
    workpool_stop(wpool);
    nhmm = wpool->nhmm;
    workpool_free(wpool);

    free(dsq);
    *ret_nhmm = nhmm;
    return;
}
/* Function: workpool_start()
 * Date:     SRE, Mon Sep 28 11:10:58 1998 [St. Louis]
 *
 * Purpose:  Initialize a workpool_s structure, and return it.
 *
 * Args:     hmmfile    - name of HMM file
 *           hmmfp      - open HMM file, at start
 *           dsq        - ptr to sequence to search
 *           seqname    - ptr to name of dsq
 *           L          - length of dsq
 *           do_forward - TRUE to score using Forward
 *           do_null2   - TRUE to apply null2 ad hoc correction
 *           threshold  - evalue/score threshold settings
 *           ghit       - per-seq hit list
 *           dhit       - per-domain hit list
 *           num_threads- number of worker threads to run.
 *
 * Returns:  ptr to struct workpool_s.
 *           Caller must wait for threads to finish with workpool_stop(),
 *           then free the structure with workpool_free().
 */
static struct workpool_s *
workpool_start(char *hmmfile, HMMFILE *hmmfp, unsigned char *dsq, char *seqname, int L,
        int do_forward, int do_null2, struct threshold_s *thresh,
        struct tophit_s *ghit, struct tophit_s *dhit,
        int num_threads)
{
    struct workpool_s *wpool;
    pthread_attr_t    attr;
    int i;
    int rtn;

    wpool             = MallocOrDie(sizeof(struct workpool_s));
    wpool->thread     = MallocOrDie(num_threads * sizeof(pthread_t));
    wpool->hmmfile    = hmmfile;
    wpool->dsq        = dsq;
    wpool->L          = L;
    wpool->seqname    = seqname;
    wpool->do_forward = do_forward;
    wpool->do_null2   = do_null2;
    wpool->thresh     = thresh;

    wpool->hmmfp      = hmmfp;
    wpool->nhmm       = 0;
    if ((rtn = pthread_mutex_init(&(wpool->input_lock), NULL)) != 0)
        Die("pthread_mutex_init FAILED; %s\n", strerror(rtn));

    wpool->ghit       = ghit;
    wpool->dhit       = dhit;
    if ((rtn = pthread_mutex_init(&(wpool->output_lock), NULL)) != 0)
        Die("pthread_mutex_init FAILED; %s\n", strerror(rtn));

    wpool->num_threads= num_threads;

    /* Create slave threads. See comments in hmmcalibrate.c at
     * this step regarding concurrency and system scope.
     */
    pthread_attr_init(&attr);
#ifndef __sgi
#ifdef HAVE_PTHREAD_ATTR_SETSCOPE
    pthread_attr_setscope(&attr, PTHREAD_SCOPE_SYSTEM);
#endif
#endif
#ifdef HAVE_PTHREAD_SETCONCURRENCY
    pthread_setconcurrency(num_threads+1);
#endif
    for (i = 0; i < num_threads; i++)
        if ((rtn = pthread_create(&(wpool->thread[i]), &attr,
                        worker_thread , (void *) wpool)) != 0)
            Die("Failed to create thread %d; return code %d\n", i, rtn);

    pthread_attr_destroy(&attr);
    return wpool;
}

/* Function: workpool_stop()
 * Date:     SRE, Thu Jul 16 11:20:16 1998 [St. Louis]
 *
 * Purpose:  Waits for threads in a workpool to finish.
 *
 * Args:     wpool -- ptr to the workpool structure
 *
 * Returns:  (void)
 */
    static void
workpool_stop(struct workpool_s *wpool)
{
    int i;
    /* wait for threads to stop */
    for (i = 0; i < wpool->num_threads; i++)
        if (pthread_join(wpool->thread[i],NULL) != 0)
            Die("pthread_join failed");
    return;
}

/* Function: workpool_free()
 * Date:     SRE, Thu Jul 16 11:26:27 1998 [St. Louis]
 *
 * Purpose:  Free a workpool_s structure, after the threads
 *           have finished.
 *
 * Args:     wpool -- ptr to the workpool.
 *
 * Returns:  (void)
 */
    static void
workpool_free(struct workpool_s *wpool)
{
    free(wpool->thread);
    free(wpool);
    return;
}


/* Function: worker_thread()
 * Date:     SRE, Mon Sep 28 10:48:29 1998 [St. Louis]
 *
 * Purpose:  The procedure executed by the worker threads.
 *
 * Args:     ptr  - (void *) that is recast to a pointer to
 *                  the workpool.
 *
 * Returns:  (void *)
 */
    void *
worker_thread(void *ptr)
{
    struct workpool_s *wpool;     /* our working threads structure   */
    struct plan7_s    *hmm;       /* an HMM to search with           */
    struct p7trace_s  *tr;        /* traceback from an alignment     */
    float  sc;			/* score of an alignment           */
    int    rtn;			/* a return code from pthreads lib */
    double pvalue;		/* P-value of score                */
    double evalue;		/* E-value of a score              */
    struct threshold_s thresh;	/* a local copy of thresholds      */
    struct dpmatrix_s *mx;        /* growable DP matrix              */

    wpool = (struct workpool_s *) ptr;
    /* Because we might dynamically change the thresholds using
     * Pfam GA/NC/TC cutoffs, we make a local copy of the threshold
     * structure in this thread.
     */
    thresh.globT   = wpool->thresh->globT;
    thresh.globE   = wpool->thresh->globE;
    thresh.domT    = wpool->thresh->domT;
    thresh.domE    = wpool->thresh->domE;
    thresh.autocut = wpool->thresh->autocut;
    thresh.Z       = wpool->thresh->Z;

    /*
     * We'll create for at least N=300xM=300, and thus consume at least 1 MB,
     * regardless of RAMLIMIT -- this helps us avoid reallocating some weird
     * asymmetric matrices.
     *
     * We're growable in both M and N, because inside of P7SmallViterbi,
     * we're going to be calling P7Viterbi on subpieces that vary in size,
     * and for different models.
     */
    mx = CreatePlan7Matrix(300, 300, 25, 25);

    for (;;) {

        /* 1. acquire lock on HMM input, and get
         *    the next HMM to work on.
         */
        /* acquire a lock */
        if ((rtn = pthread_mutex_lock(&(wpool->input_lock))) != 0)
            Die("pthread_mutex_lock failure: %s\n", strerror(rtn));

        if (! HMMFileRead(wpool->hmmfp, &hmm))
        {	/* we're done. release lock, exit thread */
            if ((rtn = pthread_mutex_unlock(&(wpool->input_lock))) != 0)
                Die("pthread_mutex_unlock failure: %s\n", strerror(rtn));
            FreePlan7Matrix(mx);
            pthread_exit(NULL);
        }
        wpool->nhmm++;
        SQD_DPRINTF1(("a thread is working on %s\n", hmm->name));
        /* release the lock */
        if ((rtn = pthread_mutex_unlock(&(wpool->input_lock))) != 0)
            Die("pthread_mutex_unlock failure: %s\n", strerror(rtn));

        if (hmm == NULL)
            Die("HMM file %s may be corrupt or in incorrect format; parse failed", wpool->hmmfile);
        P7Logoddsify(hmm, !(wpool->do_forward));

        if (!SetAutocuts(&thresh, hmm))
            Die("HMM %s did not have the right GA, NC, or TC cutoffs", hmm->name);

        /* 2. We have an HMM in score form.
         *    Score the sequence.
         */
        if (P7ViterbiSpaceOK(wpool->L, hmm->M, mx))
            sc = P7Viterbi(wpool->dsq, wpool->L, hmm, mx, &tr);
        else
            sc = P7SmallViterbi(wpool->dsq, wpool->L, hmm, mx, &tr);


        /* The Forward score override (see comments in serial vers)
         */
        if (wpool->do_forward) {
            sc  = P7Forward(wpool->dsq, wpool->L, hmm, NULL);
            if (wpool->do_null2)   sc -= TraceScoreCorrection(hmm, tr, wpool->dsq);
        }

        /* 3. Save the output in tophits structures, after acquiring a lock
         */
        if ((rtn = pthread_mutex_lock(&(wpool->output_lock))) != 0)
            Die("pthread_mutex_lock failure: %s\n", strerror(rtn));
        SQD_DPRINTF1(("model %s scores %f\n", hmm->name, sc));

        pvalue = PValue(hmm, sc);
        evalue = thresh.Z ? (double) thresh.Z * pvalue : (double) wpool->nhmm * pvalue;
        if (sc >= thresh.globT && evalue <= thresh.globE)
        {
            sc = PostprocessSignificantHit(wpool->ghit, wpool->dhit,
                    tr, hmm, wpool->dsq, wpool->L,
                    wpool->seqname,
                    NULL, NULL, /* won't need seq's acc or desc */
                    wpool->do_forward, sc,
                    wpool->do_null2,
                    &thresh,
                    TRUE); /* TRUE -> hmmpfam mode */
        }
        if ((rtn = pthread_mutex_unlock(&(wpool->output_lock))) != 0)
            Die("pthread_mutex_unlock failure: %s\n", strerror(rtn));

        P7FreeTrace(tr);
        FreePlan7(hmm);

    } /* end 'infinite' loop over HMMs in this thread */
}
#else /* HMMER_THREADS off; put in a dummy stub: */
static void
main_loop_threaded(char *hmmfile, HMMFILE *hmmfp, char *seq, SQINFO *sqinfo,
        struct threshold_s *thresh, int do_xnu, int do_forward, int do_null2,
        int num_threads,
        struct tophit_s *ghit, struct tophit_s *dhit, int *ret_nhmm)
{
    Die("no threads support");
}
#endif /* HMMER_THREADS */
#ifdef HMMER_MPI

//The following 4 functions are useless as they are used for the dynamic
//master division scheme which is no longer used, but are still in place
//because they are intricately tangled into the setup phase.  If one were
//to rewrite it, one would take out these functions altogether.

void masterCommCreate(int begin, int length, MPI_Comm * comm)
{
    MPI_Group group1, group2;
    int ranks[][3] = {{begin*cluster,begin*cluster,1}, {(begin-length+1)*cluster+1,begin*cluster+1,cluster}};
    int rank;

    MPI_Comm_rank(MPI_COMM_WORLD, &rank);
    MPI_Comm_group(MPI_COMM_WORLD, &group1);
    MPI_Group_range_incl(group1, 2, ranks, &group2);
    MPI_Comm_create(MPI_COMM_WORLD, group2, comm);
}

void submasterCommCreate(MPI_Status *status, MPI_Comm * comm, MPI_Comm mastercomm, MPI_Comm subcomm)
{
    int args[2], i, j=-4;
    int size;
    int rank;

    MPI_Comm_size(subcomm, &size);
    MPI_Comm_rank(mastercomm, &rank);
    MPI_Recv(args, 2, MPI_INT, 0, 501, mastercomm, status);

    for(i=0; i<size; i++)
    {
        if(i!=1 && (i!=0 || rank!=1))
        {
            MPI_Send(&j, 1, MPI_INT, i, 0, subcomm);
            MPI_Send(args, 2, MPI_INT, i, 1, subcomm);
        }
    }
    masterCommCreate(args[0], args[1], comm);
}

void slaveCommCreate(MPI_Comm *comm, MPI_Comm subcomm)
{
    int args[2], i;
    MPI_Status status;

    MPI_Recv(args, 2, MPI_INT, 1, 1, subcomm, &status);
    masterCommCreate(args[0], args[1], comm);
}

void callOtherMasters(int begin, int length, MPI_Comm * comm, MPI_Comm mastercomm)
{
    int i, j=-4;
    int args[] = {begin, length};
    int size;
    MPI_Comm_size(mastercomm, &size);

    for(i = 1; i<size; i++)
    {
        MPI_Send(&j, 1, MPI_INT, i, 500, mastercomm);
        MPI_Send(args, 2, MPI_INT, i, 501, mastercomm);
    }
    masterCommCreate(begin, length, comm);
}

/**********************************
The print_master() function:

Acts as load balancer and printer.  When a low-level master is ready to print,
he sends a message to the print master indicating either that he has more
queries to process or that he is done.  The master then prints to a buffer and
sends it to the printer.  Meanwhile, the printer checks its query queue to
discover if any queries remain to be processed.  If so, he gives the next query
to the master if it is done, and then prints the sent buffer.  Single node
output is needed because I/O across multiple nodes is tricky in time.

***********************************/

void print_master(MPI_Comm mastercomm, int masters, int next, int final, FILE * index)
{
    int hold;
    MPI_Status temp;
    while(masters)
    {
        char buffer[700000];
        MPI_Recv(&hold, 1, MPI_INT, MPI_ANY_SOURCE, 987, mastercomm, &temp);
        int offset;
        if(hold==1 && next<final)
        {
            fseek(index, 12*(next), SEEK_SET);
            fscanf(index, "%d", &offset);
            next++;
        }
        else if(hold==1)
        {
            masters--;
            offset = -1;
        }
        MPI_Recv(buffer, 700000, MPI_BYTE, temp.MPI_SOURCE, 985, mastercomm, &temp);
        if(hold==1)
        {
            MPI_Send(&offset, 1, MPI_INT, temp.MPI_SOURCE, 984, mastercomm);
        }
        buffer[temp.count] = 0;
        printf("%s\n", buffer);
    }
}

/*
Coordinator function: main_loop_mpi()
-------------------------------------

This function does either one of two things.  If there are enough query
sequences, it assigns each master a set of sequences and itself becomes the
print master.  Otherwise, it bands all the masters together to perform the
queries sequentially.

*/

static void main_loop_mpi(char * hmmfile, HMMFILE *hmmfp, char *seq,
        SQINFO *sqinfo, struct threshold_s *thresh,
        int do_xnu, int do_forward, int do_null2, struct transporter pack,
        struct tophit_s *ghit,
        struct tophit_s *dhit, int *ret_nhmm, int num_nodes,
        MPI_Comm mastercomm, MPI_Comm subcomm, FILE * indexfile, int num_mast, int off,
        int num_seq, char * index, char * seqfile, int format)
{
    unsigned char *dsq;
    int i, count = 0;
    MPI_Status temp;
    int my_rank, ranks[][3] = {{0,0,1}};
    int numseqs, masters, saveseqs, ratio, mastertemp, offset;
    MPI_Group subgroup;
    int hmmpermaster;
    int diff, sum, j;
    int totals, total, check, next;
    int     ndom;			/* total # of domains in this seq          */
    int     namewidth;		/* max width of printed HMM name           */
    int     descwidth;		/* max width of printed description        */
    float   sc;			/* log-odds score in bits                  */
    double  pvalue;		/* pvalue of an HMM score                  */
    double  evalue;		/* evalue of an HMM score                  */
    double  motherp;		/* pvalue of a whole seq HMM score         */
    float   mothersc;		/* score of a whole seq parent of domain   */
    int     sqfrom, sqto;		/* coordinates in sequence                 */
    int     hmmfrom, hmmto;	/* coordinate in HMM                       */
    char   *name, *acc, *desc;    /* hit HMM name, accession, description            */
    int     hmmlen;		/* length of HMM hit                       */
    int     nhmm;			/* number of HMMs searched                 */
    int     domidx;		/* number of this domain                   */
    int     current;
    int   nreported;
    int    Alimit= INT_MAX;		/* A parameter limiting output alignments   */
    int    arglen;
    int    be_backwards = pack.be_backwards;
    int    show_acc = pack.show_acc;
    struct fancyali_s *ali;	/* an alignment for display                */
    SQFILE *sqfp;
    MPI_Comm tempcomm = mastercomm;

    if(hmmfp==NULL)
        if ((hmmfp = HMMFileOpen(hmmfile, "HMMERDB")) == NULL)
            Die("Failed to open HMM database %s\n%s", hmmfile, usage);
    if (hmmfp->ssi == NULL)
        Die("HMM file %s needs an SSI index to use PVM. See: hmmindex.", hmmfile);
    if(indexfile==NULL)
        indexfile = fopen(index, "r");
    if ((sqfp = SeqfileOpen(seqfile, format, NULL)) == NULL)
        Die("Failed to open sequence file %s\n%s\n", seqfile, usage);
    //if num_mast = 0, this function is acting as a coordinator,
    //if num_mast = 1, this function is acting like a master, actually
    //handling data.
    if(num_mast==0)
    {
        MPI_Bcast(&do_xnu, 1, MPI_INT, 0, MPI_COMM_WORLD);
        MPI_Bcast(&do_forward, 1, MPI_INT, 0, MPI_COMM_WORLD);
        MPI_Bcast(&do_null2, 1, MPI_INT, 0, MPI_COMM_WORLD);
        MPI_Bcast(&Alphabet_type, 1, MPI_INT, 0, MPI_COMM_WORLD);

        MPI_Comm_size(mastercomm, &masters);
        masters--;
        mastertemp = masters;

        const double perc = 1;   /***** Load balancing default = .7 *****/

        fscanf(indexfile, "%d", &numseqs);
        int saveseqs2 = numseqs;
        const int RATIO = 2;
        if(numseqs/masters<RATIO) numseqs = 1;   /******* RATIO default = 2 *******/
        else numseqs = (int) (perc*numseqs);
        saveseqs = numseqs;
        while(numseqs < mastertemp)
        {
            if(numseqs > 0)
            {   ratio = mastertemp/numseqs;
                callOtherMasters(mastertemp-1, ratio, &tempcomm, mastercomm);
                MPI_Send(&ratio, 1, MPI_INT, mastertemp, 500, mastercomm);
                j=-5;
                for(i=mastertemp-1; i>mastertemp-ratio; i--)
                    MPI_Send(&j, 1, MPI_INT, i, 500, mastercomm);
                if(count==0)
                    MPI_Send(&count, 1, MPI_INT, mastertemp, 501, mastercomm);
                
                else
                {   fseek(indexfile, 12*(count), SEEK_SET);
                    fscanf(indexfile, "%d", &offset);
                    MPI_Send(&offset, 1, MPI_INT, mastertemp, 501, mastercomm);
                }
                i = 1;
                MPI_Send(&saveseqs2, 1, MPI_INT, mastertemp, 502, mastercomm);
                count++;
                numseqs--;
                mastertemp -= ratio;
            }
            if(numseqs <=0)
            {   
                j=-5;
                main_loop_mpi_slave(hmmfile, hmmfp, seq, *thresh, do_xnu,
                    do_forward, do_null2, pack, nhmm, 0, index, subcomm,
                    tempcomm, seqfile,
                    format);
            }
        }
        while(numseqs >= mastertemp && numseqs!=0)
        {
            ratio = numseqs/mastertemp;
            i=1;
            MPI_Send(&i, 1, MPI_INT, mastertemp, 500, mastercomm);
            if(count==0)
                MPI_Send(&count, 1, MPI_INT, mastertemp, 501, mastercomm);
            else
            {
                fseek(indexfile, 12*(count), SEEK_SET);
                fscanf(indexfile, "%d", &offset);
                MPI_Send(&offset, 1, MPI_INT, mastertemp, 501, mastercomm);
            }
            MPI_Send(&ratio, 1, MPI_INT, mastertemp, 502, mastercomm);
            if(mastertemp==1)
            {
                print_master(mastercomm, masters, saveseqs, saveseqs2, indexfile);
                return;
            }
            count+=ratio;
            numseqs-=ratio;
            mastertemp--;

        }
    }
    else
    {
        if(off!=0)
        {   
            fseek(sqfp->f, off, SEEK_SET);
            ReadSeq(sqfp, sqfp->format, &seq, sqinfo);
        }
        mastertemp = num_mast;
        
        while (ReadSeq(sqfp, format, &seq, sqinfo) && num_seq>0)
         {
            num_seq--;
            num_mast = mastertemp;
            dhit = AllocTophits(4000); //normally this would only have to allocate ~20 hits
                                      // however there is a memory corruptions somewhere preventing
                                      // reallocation that would normally occur.
            ghit = AllocTophits(4000);
            totals = hmmfp->ssi->nprimary;
            total = totals;
            hmmpermaster = totals/num_mast;
            diff = totals - num_mast*hmmpermaster;
            j=0;
            current = 0;
            for(i=0; i<num_mast; i++)
            {   
                if(i==0) hmmpermaster+=diff;
                MPI_Send(&j, 1, MPI_INT, i+1, 500, mastercomm);
                MPI_Send(&current, 1, MPI_INT, i+1, 501, mastercomm);
                MPI_Send(&hmmpermaster, 1, MPI_INT, i+1, 502, mastercomm);
                mpiSendSeq(seq, sqinfo, i+1, mastercomm);
                if(i==0) hmmpermaster-=diff;
                if(current ==0) current +=diff;
                current +=hmmpermaster;
            }
            while(num_mast)
            {   MPI_Recv(&j, 1, MPI_INT, MPI_ANY_SOURCE, 151, mastercomm, &temp);
                if(j==1 || j==3)
                {
                    MPI_Recv(dhit->unsrt + dhit->num, 500*sizeof(struct hit_s),
                        MPI_BYTE, temp.MPI_SOURCE, 152, mastercomm,
                        &temp);
                    next = temp.count/sizeof(struct hit_s);
                    mpiUnpackHits(dhit->unsrt + dhit->num, next, temp.MPI_SOURCE,
                        mastercomm);
                    dhit->num += next;
                }
                if(j==2 || j==3)
                {
                    MPI_Recv(ghit->unsrt + ghit->num, 500*sizeof(struct hit_s),
                        MPI_BYTE, temp.MPI_SOURCE, 152, mastercomm,
                        &temp);
                    next = temp.count/sizeof(struct hit_s);
                    mpiUnpackHits(ghit->unsrt + ghit->num, next, temp.MPI_SOURCE,
                        mastercomm);
                    ghit->num += next;
                }
                if(j==3)
                    num_mast--;
            }

            if (!thresh->Z) thresh->Z = totals;

            /* 2. (Done searching all HMMs for this query seq; start output)
             *    Report the overall sequence hits, sorted by significance.
             */
            if (be_backwards)
            {
                printf("Query:  %s  %s\n", sqinfo->name,
                        sqinfo->flags & SQINFO_DESC ? sqinfo->desc : "");
            }
            else
            {
                printf("\nQuery sequence: %s\n", sqinfo->name);
                printf("Accession:      %s\n", sqinfo->flags &SQINFO_ACC ? sqinfo->acc  : "[none]");
                printf("Description:    %s\n", sqinfo->flags &SQINFO_DESC ? sqinfo->desc : "[none]");
            }
            /* We'll now sort the global hit list by evalue...
             * (not score! that was bug #12. in hmmpfam, score and evalue are not
             *  monotonic.)
             */
            FullSortTophits(ghit);
            namewidth = MAX(8, TophitsMaxName(ghit)); /* must print whole name, no truncation  */
            descwidth = MAX(52-namewidth, 11);        /* may truncate desc, but avoid neg len! */

            printf("\nScores for sequence family classification (score includes all domains):\n");
            printf("%-*s %-*s %7s %10s %3s\n", namewidth, "Model", descwidth, "Description", "Score", "E-value", " N ");
            printf("%-*s %-*s %7s %10s %3s\n", namewidth, "--------", descwidth, "-----------", "-----", "-------", "---");
            for (i = 0, nreported = 0; i < ghit->num; i++)
            {
                char *safedesc;
                GetRankedHit(ghit, i,
                        &pvalue, &sc, NULL, NULL,
                        &name, &acc, &desc,
                        NULL, NULL, NULL,           /* seq positions */
                        NULL, NULL, NULL,           /* HMM positions */
                        NULL, &ndom,                /* domain info   */
                        NULL);	                   /* alignment info*/

                evalue = pvalue * (double) thresh->Z;

                /* safedesc is a workaround for an apparent Linux printf()
                 * bug with the *.*s format. dbmalloc crashes with a memchr() ptr out of bounds
                 * flaw if the malloc'ed space for desc is short. The workaround
                 * is to make sure the ptr for *.* has a big malloc space.
                 */
                if (desc != NULL && strlen(desc) < 80)
                {
                    safedesc = MallocOrDie(sizeof(char) * 80);
                    strcpy(safedesc, desc);
                }
                else safedesc = Strdup(desc);

                /* sneaky trick warning:
                 * if we're using dynamic Pfam score cutoffs (GA, TC, NC),
                 * then the list of hits is already correct and does not
                 * need any score cutoffs. Unset the thresholds. They'll
                 * be reset in the main_loop if we still have sequences
                 * to process.
                 */
                if (thresh->autocut != CUT_NONE) {
                    thresh->globE = thresh->domE = FLT_MAX;
                    thresh->globT = thresh->domT = -FLT_MAX;
                }

                if (evalue <= thresh->globE && sc >= thresh->globT)
                {
                    printf("%-*s %-*.*s %7.1f %10.2g %3d\n",
                            namewidth,
                            (show_acc && acc != NULL) ?  acc : name,
                            descwidth, descwidth, safedesc != NULL ? safedesc : "",
                            sc, evalue, ndom);
                    nreported++;
                }
                free(safedesc);
            }
            if (nreported == 0) printf("\t[no hits above thresholds]\n");

            /* 3. Report domain hits (sorted on sqto coordinate)
             */
            FullSortTophits(dhit);
            namewidth = MAX(8, TophitsMaxName(dhit)); /* must print whole name, no truncation  */

            printf("\nParsed for domains:\n");
            printf("%-*s %7s %5s %5s    %5s %5s    %7s %8s\n",
                    namewidth, "Model", "Domain ", "seq-f", "seq-t", "hmm-f", "hmm-t", "score", "E-value");
            printf("%-*s %7s %5s %5s    %5s %5s    %7s %8s\n",
                    namewidth, "--------", "-------", "-----", "-----", "-----", "-----", "-----", "-------");

            for (i = 0, nreported = 0; i < dhit->num; i++)
            {
                GetRankedHit(dhit, i,
                        &pvalue, &sc, &motherp, &mothersc,
                        &name, &acc, NULL,
                        &sqfrom, &sqto, NULL,
                        &hmmfrom, &hmmto, &hmmlen,
                        &domidx, &ndom,
                        NULL);
                evalue = pvalue * (double) thresh->Z;

                /* Does the "mother" (complete) sequence satisfy global thresholds? */
                if (motherp * (double)thresh->Z > thresh->globE || mothersc < thresh->globT)
                    continue;
                else if (evalue <= thresh->domE && sc >= thresh->domT) {
                    printf("%-*s %3d/%-3d %5d %5d %c%c %5d %5d %c%c %7.1f %8.2g\n",
                            namewidth,
                            (show_acc && acc != NULL) ?  acc : name,
                            domidx, ndom,
                            sqfrom, sqto,
                            sqfrom == 1 ? '[' : '.', sqto == sqinfo->len ? ']' : '.',
                            hmmfrom, hmmto,
                            hmmfrom == 1 ? '[':'.',  hmmto == hmmlen ? ']' : '.',
                            sc, evalue);
                    nreported++;
                }
            }
            if (nreported == 0) printf("\t[no hits above thresholds]\n");


            /* 3. Alignment output, also by domain.
             *    dhits is already sorted and namewidth is set, from above code.
             *    Number of displayed alignments is limited by Alimit parameter;
             *    also by domE (evalue threshold), domT (score theshold).
             */
            if (Alimit != 0)
            {
                printf("\nAlignments of top-scoring domains:\n");
                for (i = 0, nreported = 0; i < dhit->num; i++)
                {
                    if (nreported == Alimit) break; /* limit to Alimit output alignments */
                    GetRankedHit(dhit, i,
                            &pvalue, &sc, &motherp, &mothersc,
                            &name, &acc, NULL,
                            &sqfrom, &sqto, NULL,              /* seq position info  */
                            &hmmfrom, &hmmto, &hmmlen,         /* HMM position info  */
                            &domidx, &ndom,                    /* domain info        */
                            &ali);	                      /* alignment info     */
                    evalue = pvalue * (double) thresh->Z;

                    if (motherp * (double) thresh->Z > thresh->globE || mothersc < thresh->globT)
                        continue;
                    else if (evalue <= thresh->domE && sc >= thresh->domT)
                    {
                        printf("%s: domain %d of %d, from %d to %d: score %.1f, E = %.2g\n",
                                (show_acc && acc != NULL) ?  acc : name,
                                domidx, ndom, sqfrom, sqto, sc, evalue);
                        PrintFancyAli(stdout, ali);
                        nreported++;
                    }
                }
                if (nreported == 0)      printf("\t[no hits above thresholds]\n");
                if (nreported == Alimit) printf("\t[output cut off at A = %d top alignments]\n", Alimit);
            }


            printf("//\n");
            FreeSequence(seq, sqinfo);
            FreeTophits(ghit);
            FreeTophits(dhit);
        }
        j=-1;
        for(i=1; i<=mastertemp; i++)
        {
            MPI_Send(&j, 1, MPI_INT, i, 500, mastercomm);
        }
    }

}

//master
static void main_loop_mpi_master(char *hmmfile, HMMFILE *hmmfp, char *seq,
        SQINFO *sqinfo, struct threshold_s *thresh,
        int do_xnu, int do_forward, int do_null2, struct transporter pack,
        struct tophit_s *ghit,
        struct tophit_s *dhit, int *ret_nhmm, int num_nodes,
        MPI_Comm subcomm, MPI_Comm mastercomm, int real_rank)
{
    struct plan7_s   *hmm;        /* HMM that was searched with */
    struct p7trace_s *tr;         /* a traceback structure */
    unsigned char    *dsq;        /* digitized sequence */
    float  sc;			/* score of an HMM match */
    int   *hmmlist = MallocOrDie(sizeof(int) * (num_nodes-1));               /* array of hmm indexes being worked on by slaves */
    int    nslaves;	       	/* number of slaves in virtual machine   */
    int    nhmm;			/* number of HMMs searched */
    int    slaveidx;		/* index of a slave wanting work         */
    int    slave, msg;
    int    sent_trace;		/* TRUE if slave sent us a trace */
    char   slavename[32];		/* name of HMM that slave actually did */
    double pvalue;		/* pvalue of HMM score */
    int    count = 0;
    int    i, j, state;
    int    offset, length;
    int    ranks[][3] = {{real_rank-1, real_rank-1, 1}, {0, real_rank, cluster}};
    int    singlemode = 0;
    int    chosen = 0;
    int    alphatype;
    int     ndom;			/* total # of domains in this seq          */
    int     namewidth;		/* max width of printed HMM name           */
    int     descwidth;		/* max width of printed description        */
    double  evalue;		/* evalue of an HMM score                  */
    double  motherp;		/* pvalue of a whole seq HMM score         */
    float   mothersc;		/* score of a whole seq parent of domain   */
    int     sqfrom, sqto;		/* coordinates in sequence                 */
    int     hmmfrom, hmmto;	/* coordinate in HMM                       */
    char   *name, *acc, *desc;    /* hit HMM name, accession, description            */
    int     hmmlen;		/* length of HMM hit                       */
    int     domidx;		/* number of this domain                   */
    int   nreported;
    int    Alimit = INT_MAX;		/* A parameter limiting output alignments   */
    int    arglen;
    int    length2;
    int    dhitc=0, ghitc=0;
    SQFILE *sqfp = NULL;
    char *seqfile = pack.seqfile;
    MPI_Group wgroup, mastergroup;
    MPI_Status temp;
    MPI_Status * useless = &temp;
    int    format = pack.format;
    int    be_backwards = pack.be_backwards;
    int    show_acc = pack.show_acc;
    struct fancyali_s *ali;	/* an alignment for display                */
    MPI_Comm tempcomm = mastercomm;

    //database in memory
    struct plan7_s *list[10000];

    //index of which hmms have been processed 
    char switches[10000];

    int index[cluster];

    MPI_Comm_size(subcomm, &nslaves);
    nslaves--;

    for(int i=0; i<10000; i++)
    	list[i] = NULL;

    /* Sanity checks.
     */
    if (hmmfp->ssi == NULL)
        Die("HMM file %s needs an SSI index to use PVM. See: hmmindex.", hmmfile);

    /* Prepare sequence.
     */

    MPI_Bcast(&do_xnu, 1, MPI_INT, 0, MPI_COMM_WORLD);
    MPI_Bcast(&do_forward, 1, MPI_INT, 0, MPI_COMM_WORLD);
    MPI_Bcast(&do_null2, 1, MPI_INT, 0, MPI_COMM_WORLD);
    MPI_Bcast(&alphatype, 1, MPI_INT, 0, MPI_COMM_WORLD);

    SetAlphabet(alphatype);

    while(1)
    {
        int thresh2;
        if(!singlemode || length2==0)
        {
            MPI_Recv(&state, 1, MPI_INT, 0, 500, tempcomm, useless);

            //After writing this code I came across MPI_Probe and realized the
            //potential of MPI_TAG.  The next section is composed of control
            //messages to tell the master how to set up, but this could be much
            //more elegantly implemented using Probe and tags.
            
            if(state==-5)
            {
                tempcomm = mastercomm;
                continue;
            }
            if(state==-3)
            {
                nslaves++;
            }
            if(state==-1)
            {
                for (slave = 2; slave <= nslaves+1; slave++)
                {
                    msg = -1;			/* the special "shut down now" flag */
                    MPI_Send(&msg, 1, MPI_INT, (slave)%cluster, 0, subcomm);
                }
                break;
            }

            if(state==-4)
            {
                submasterCommCreate(useless, &mastercomm, mastercomm, subcomm);
                continue;
            }

            MPI_Recv(&offset, 1, MPI_INT, 0, 501, tempcomm, useless);
            MPI_Recv(&length, 1, MPI_INT, 0, 502, tempcomm, useless);

            if(state>1)
            {
                j=-2;
                MPI_Send(&j, 1, MPI_INT, 0, 0, subcomm);
                MPI_Send(&offset, 1, MPI_INT, 0, 881, subcomm);
                MPI_Send(&length, 1, MPI_INT, 0, 882, subcomm);
                MPI_Send(&state, 1, MPI_INT, 0, 883, subcomm);
                tempcomm = mastercomm;
                chosen = 1;
                nslaves--;
                singlemode = 0;
                continue;
            }
            if(state==-2)
            {   singlemode = 1;
                chosen = 0;
                if(real_rank==1)
                    nslaves--;
                length2=length;
                length = hmmfp->ssi->nprimary;
            }
            if(state==1)
            {
                singlemode = 1;
                chosen = 0;
                if(real_rank==1)
                    nslaves--;
                length2=length;
                length = hmmfp->ssi->nprimary;
            }

            if(singlemode) //singlemode means the masters are not all tied together to the supermaster,
                           //this is most likely the case.
            {
                if(sqfp==NULL)
                {   

                    if ((sqfp = SeqfileOpen(seqfile, format, NULL)) == NULL)
                        Die("Failed to open sequence file %s\n%s\n", seqfile, usage);

                }
                if(offset!=0)
                {
                    sqinfo = malloc(sizeof(SQINFO));
                    fseek(sqfp->f, offset, SEEK_SET);
                    ReadSeq(sqfp, sqfp->format, &seq, sqinfo);
                    free(sqinfo);
                }
                offset=0;
            }
                ghit = AllocTophits(3000);
                dhit = AllocTophits(3000);

        }


        const double perc = .97;  //load balancing default = .97 (3% load balancing)
                                  //i.e. 97% of the database is loaded across the slaves,
                                  //remaining 3% all slaves load.

        for(int i=2; i<nslaves+2; i++)
      	{
            index[i%cluster] = offset + (i-2)*perc*length/nslaves;
            if(i==nslaves+1) thresh2 = index[i%cluster];
        }

      	for(int i=0; i<10000; i++)
      		switches[i] = 0;

        if(singlemode)
        {
            sqinfo = malloc(sizeof(SQINFO));
            ReadSeq(sqfp, sqfp->format, &seq, sqinfo);
            length2--;
            offset = 0;
        }
        else
            mpiRecvSeq(&seq, sqinfo, 0, mastercomm);
        dsq = DigitizeSequence(seq, sqinfo->len);
        if (do_xnu && Alphabet_type == hmmAMINO) XNU(dsq, sqinfo->len);

        /* Initialize the slaves
         */
        arglen = strlen(hmmfile);

        SQD_DPRINTF1(("Broadcasting to %d slaves...\n", nslaves));
        j=-3;
        for(int i=2; i<=nslaves+1; i++)
        {
            MPI_Send(&j, 1, MPI_INT, i%cluster, 0, subcomm);
            MPI_Send(&(sqinfo->len), 1, MPI_INT, i%cluster, 770, subcomm);
            MPI_Send(seq, sqinfo->len, MPI_CHAR, i%cluster, 771, subcomm);
        }

        /* Load the slaves.
         * For efficiency reasons, we don't want the master to
         * load HMMs from disk until she absolutely needs them.
         */
        length += offset;
    	int last = length-1;


        for (i = 2; i < nslaves+2; i++)
        {
            MPI_Send(index+i%cluster, 1, MPI_INT, i%cluster, 0, subcomm);	
            switches[index[i%cluster]] = 1;
            hmmlist[i%cluster] = index[i%cluster];
            if(index[i%cluster]+1<length && switches[index[i%cluster]+1]==0)
            	index[i%cluster]++;
            else index[i%cluster] = -1;
        }

        nhmm = nslaves + offset;
        /* Receive/send loop
         */
        for (; nhmm < length; nhmm++)
        {
            /* check slaves before blocking */
            /* receive output */
            SQD_DPRINTF1(("Waiting for a slave to give me output...\n"));
            MPI_Recv(&slaveidx, 1, MPI_INT, MPI_ANY_SOURCE, 1, subcomm, useless);     /* # of slave who's sending us stuff */
            MPI_Recv(&arglen, 1, MPI_INT, slaveidx, 2, subcomm, useless);
            MPI_Recv(slavename, arglen, MPI_CHAR, slaveidx, 3, subcomm, useless);           /* name of HMM that slave did */
            slavename[arglen]=0;
            MPI_Recv(&sc, 1, MPI_FLOAT, slaveidx, 4, subcomm, useless);         /* score   */
            MPI_Recv(&pvalue, 1, MPI_DOUBLE, slaveidx, 5, subcomm, useless);    /* P-value */
            MPI_Recv(&sent_trace, 1, MPI_INT, slaveidx, 6, subcomm, useless);   /* TRUE if trace is coming */
            tr = (sent_trace) ? mpiRecvTrace(slaveidx, subcomm) : NULL;
            SQD_DPRINTF1(("Slave %d finished %s for me...\n", slaveidx, slavename));
            /* send new work */
            int queue = index[slaveidx];
            if(queue==-1 || switches[queue]>0)  //load balancing section
            {
                index[slaveidx] = -1;
                for( ; last>=offset; last--)
                {
                    if(switches[last]==0)
                    {
                            queue = last;
                            break;
                    }
                }
            }

            MPI_Send(&queue, 1, MPI_INT, slaveidx, 0, subcomm);

            switches[queue] = 1;

            if(index[slaveidx]!=-1)
            {
                if(queue+1<length && switches[queue+1]==0)
                        index[slaveidx]++;
                else index[slaveidx] = -1;
            }

            SQD_DPRINTF1(("Assigned %d -> slave %d\n", nhmm, slaveidx));
            /* process output */
            /* 1b. Store scores/pvalue for each HMM aligned to this sequence, overall
             */
            count++;
            SQD_DPRINTF1(("%15s : %2d : %f\n", slavename, slaveidx, sc));
            if (sent_trace)
            {
                /* now load the HMM, because the hit is significant */
                if(list[hmmlist[slaveidx]]==NULL)
                {
                    HMMFilePositionByIndex(hmmfp, hmmlist[slaveidx]);
                    if (!HMMFileRead(hmmfp, &hmm))
                    { Die("Unexpected failure to read HMM file %s", hmmfile); }
                    if (hmm == NULL)
                    { Die("HMM file %s may be corrupt; parse failed", hmmfile); }
                    list[hmmlist[slaveidx]] = hmm;
                }
                else
                {
                    hmm = list[hmmlist[slaveidx]];
                }

                P7Logoddsify(hmm, TRUE);
                if (! SetAutocuts(thresh, hmm))
                    Die("HMM %s did not contain your GA, NC, or TC cutoffs", hmm->name);
                sc = PostprocessSignificantHit(ghit, dhit,
                    tr, hmm, dsq, sqinfo->len,
                    sqinfo->name,
                    sqinfo->flags & SQINFO_ACC ? sqinfo->acc : NULL,
                    sqinfo->flags & SQINFO_DESC ? sqinfo->desc : NULL,
                    do_forward, sc,
                    do_null2,
                    thresh,
                    TRUE); /* TRUE -> hmmpfam mode */
                P7FreeTrace(tr);
            }
            if(list[hmmlist[slaveidx]]!=NULL && hmmlist[slaveidx] <= thresh2 )
            {
                FreePlan7(list[hmmlist[slaveidx]]);
                list[hmmlist[slaveidx]] = NULL;
            }
            if(count==countol)
            {   if((dhit->num - dhitc) > hitol && singlemode==0)
                {
                    i=1;
                    MPI_Send(&i, 1, MPI_INT, 0, 151, mastercomm);
                    MPI_Send(dhit->unsrt + dhitc, (dhit->num - dhitc)*sizeof(struct hit_s),
                        MPI_BYTE, 0, 152, mastercomm);
                    mpiPackHits(dhit->unsrt+dhitc, dhit->num-dhitc, 0, mastercomm);
                    dhitc = dhit->num;
                }
                else if((ghit->num - ghitc) > hitol && singlemode==0)
                {
                    i=2;
                    MPI_Send(&i, 1, MPI_INT, 0, 151, mastercomm);
                    MPI_Send(ghit->unsrt + ghitc, (ghit->num - ghitc)*sizeof(struct hit_s),
                        MPI_BYTE, 0, 152, mastercomm);
                    mpiPackHits(ghit->unsrt+ghitc, ghit->num-ghitc, 0, mastercomm);
                    ghitc = ghit->num;
                }
                count=0;
            }
            hmmlist[slaveidx] = queue;
        }

        /* Collect the output. all n slaves are still working, so wait for them.
         */

        for (slave = 0; slave < nslaves; slave++)
        {
            /* don't check slaves (they're exiting normally);
               window of vulnerability here to slave crashes */
            /* receive output */
            MPI_Recv(&slaveidx, 1, MPI_INT, MPI_ANY_SOURCE, 1, subcomm, useless);     /* # of slave who's sending us stuff */
            MPI_Recv(&arglen, 1, MPI_INT, slaveidx, 2, subcomm, useless);
            MPI_Recv(slavename, arglen, MPI_CHAR, slaveidx, 3, subcomm, useless);           /* name of HMM that slave did */
            slavename[arglen] = 0;
            MPI_Recv(&sc, 1, MPI_FLOAT, slaveidx, 4, subcomm, useless);         /* score   */
            MPI_Recv(&pvalue, 1, MPI_DOUBLE, slaveidx, 5, subcomm, useless);    /* P-value */
            MPI_Recv(&sent_trace, 1, MPI_INT, slaveidx, 6, subcomm, useless);   /* TRUE if trace is coming */
            tr = (sent_trace) ? mpiRecvTrace(slaveidx, subcomm) : NULL;

            /* process output */

            SQD_DPRINTF1(("%15s : %2d : %f\n", slavename, slaveidx, sc));
            if (sent_trace)
            {
                /* now load the HMM, because the hit is significant */
                if(list[hmmlist[slaveidx]]==NULL)
                {
                    HMMFilePositionByIndex(hmmfp, hmmlist[slaveidx]);
                    if (!HMMFileRead(hmmfp, &hmm))
                    { Die("Unexpected failure to read HMM file %s", hmmfile); }
                    if (hmm == NULL)
                    { Die("HMM file %s may be corrupt; parse failed", hmmfile); }
                    list[hmmlist[slaveidx]] = hmm;
                }
                else hmm = list[hmmlist[slaveidx]];

                P7Logoddsify(hmm, TRUE);
                if (! SetAutocuts(thresh, hmm))
                    Die("HMM %s did not contain your GA, NC, or TC cutoffs", hmm->name);

                sc = PostprocessSignificantHit(ghit, dhit,
                    tr, hmm, dsq, sqinfo->len,
                    sqinfo->name, NULL, NULL, /* won't need acc or desc even if we have 'em */
                    do_forward, sc,
                    do_null2,
                    thresh,
                    TRUE); /* TRUE -> hmmpfam mode */

                P7FreeTrace(tr);
            }
            if(list[hmmlist[slaveidx]]!=NULL && hmmlist[slaveidx] <=thresh2)
			{
				FreePlan7(list[hmmlist[slaveidx]]);
				list[hmmlist[slaveidx]] = NULL;
			}
        }
        if(singlemode==0)
        {   
            i=3;
            MPI_Send(&i, 1, MPI_INT, 0, 151, mastercomm);
            MPI_Send(dhit->unsrt + dhitc, (dhit->num - dhitc)*sizeof(struct hit_s),
                MPI_BYTE, 0, 152, mastercomm);
            mpiPackHits(dhit->unsrt+dhitc, dhit->num-dhitc, 0, mastercomm);
        dhitc = dhit->num;
            MPI_Send(ghit->unsrt + ghitc, (ghit->num - ghitc)*sizeof(struct hit_s),
                 MPI_BYTE, 0, 152, mastercomm);
            mpiPackHits(ghit->unsrt+ghitc, ghit->num-ghitc, 0, mastercomm);
            ghitc = ghit->num;
        }

        ghitc = 0;
        dhitc = 0;
        if(singlemode)
        {
            if (!thresh->Z) thresh->Z = nhmm;

            /* 2. (Done searching all HMMs for this query seq; start output)
             *    Report the overall sequence hits, sorted by significance.
             */
             int hold;
             if(length2<=0)
                 hold = 1;
             else hold = 0;
             char buffer[700000];
		int count = 0;

             MPI_Send(&hold, 1, MPI_INT, 0, 987, mastercomm);

            if (be_backwards)
            {
                count+=sprintf(buffer+count, "Query:  %s  %s\n", sqinfo->name,
                        sqinfo->flags & SQINFO_DESC ? sqinfo->desc : "");
            }
            else
            {
                count+=sprintf(buffer+count, "\nQuery sequence: %s\n", sqinfo->name);
                count+=sprintf(buffer+count, "Accession:      %s\n", sqinfo->flags &SQINFO_ACC ? sqinfo->acc  : "[none]");
                count+=sprintf(buffer+count, "Description:    %s\n", sqinfo->flags &SQINFO_DESC? sqinfo->desc : "[none]");
            }
            /* We'll now sort the global hit list by evalue...
             * (not score! that was bug #12. in hmmpfam, score and evalue are not
             *  monotonic.)
             */
            FullSortTophits(ghit);
            namewidth = MAX(8, TophitsMaxName(ghit)); /* must print whole name, no truncation  */
            descwidth = MAX(52-namewidth, 11);        /* may truncate desc, but avoid neg len! */

            count+=sprintf(buffer+count, "\nScores for sequence family classification (score includes all domains):\n");
            count+=sprintf(buffer+count, "%-*s %-*s %7s %10s %3s\n", namewidth, "Model", descwidth, "Description", "Score", "E-value", " N ");
            count+=sprintf(buffer+count, "%-*s %-*s %7s %10s %3s\n", namewidth, "--------", descwidth, "-----------", "-----", "-------", "---");
            for (i = 0, nreported = 0; i < ghit->num; i++)
            {
                char *safedesc;
                GetRankedHit(ghit, i,
                        &pvalue, &sc, NULL, NULL,
                        &name, &acc, &desc,
                        NULL, NULL, NULL,           /* seq positions */
                        NULL, NULL, NULL,           /* HMM positions */
                        NULL, &ndom,                /* domain info   */
                        NULL);	                   /* alignment info*/

                evalue = pvalue * (double) thresh->Z;

                /* safedesc is a workaround for an apparent Linux printf()
                 * bug with the *.*s format. dbmalloc crashes with a memchr() ptr out of bounds
                 * flaw if the malloc'ed space for desc is short. The workaround
                 * is to make sure the ptr for *.* has a big malloc space.
                 */
                if (desc != NULL && strlen(desc) < 80)
                {
                    safedesc = MallocOrDie(sizeof(char) * 80);
                    strcpy(safedesc, desc);
                }
                else safedesc = Strdup(desc);

                /* sneaky trick warning:
                 * if we're using dynamic Pfam score cutoffs (GA, TC, NC),
                 * then the list of hits is already correct and does not
                 * need any score cutoffs. Unset the thresholds. They'll
                 * be reset in the main_loop if we still have sequences
                 * to process.
                 */
                if (thresh->autocut != CUT_NONE) {
                    thresh->globE = thresh->domE = FLT_MAX;
                    thresh->globT = thresh->domT = -FLT_MAX;
                }

                if (evalue <= thresh->globE && sc >= thresh->globT)
                {
                    count+=sprintf(buffer+count, "%-*s %-*.*s %7.1f %10.2g %3d\n",
                            namewidth,
                            (show_acc && acc != NULL) ?  acc : name,
                            descwidth, descwidth, safedesc != NULL ? safedesc : "",
                            sc, evalue, ndom);
                    nreported++;
                }
                free(safedesc);
            }
            if (nreported == 0) count+=sprintf(buffer+count, "\t[no hits above thresholds]\n");

            /* 3. Report domain hits (sorted on sqto coordinate)
             */
            FullSortTophits(dhit);
            namewidth = MAX(8, TophitsMaxName(dhit)); /* must print whole name, no truncation  */

            count+=sprintf(buffer+count, "\nParsed for domains:\n");
            count+=sprintf(buffer+count, "%-*s %7s %5s %5s    %5s %5s    %7s %8s\n",
                    namewidth, "Model", "Domain ", "seq-f", "seq-t", "hmm-f", "hmm-t", "score", "E-value");
            count+=sprintf(buffer+count, "%-*s %7s %5s %5s    %5s %5s    %7s %8s\n",
                    namewidth, "--------", "-------", "-----", "-----", "-----", "-----", "-----", "-------");

            for (i = 0, nreported = 0; i < dhit->num; i++)
            {
                GetRankedHit(dhit, i,
                        &pvalue, &sc, &motherp, &mothersc,
                        &name, &acc, NULL,
                        &sqfrom, &sqto, NULL,
                        &hmmfrom, &hmmto, &hmmlen,
                        &domidx, &ndom,
                        NULL);
                evalue = pvalue * (double) thresh->Z;

                /* Does the "mother" (complete) sequence satisfy global thresholds? */
                if (motherp * (double)thresh->Z > thresh->globE || mothersc < thresh->globT)
                    continue;
                else if (evalue <= thresh->domE && sc >= thresh->domT) {
                    count+=sprintf(buffer+count, "%-*s %3d/%-3d %5d %5d %c%c %5d %5d %c%c %7.1f %8.2g\n",
                            namewidth,
                            (show_acc && acc != NULL) ?  acc : name,
                            domidx, ndom,
                            sqfrom, sqto,
                            sqfrom == 1 ? '[' : '.', sqto == sqinfo->len ? ']' : '.',
                            hmmfrom, hmmto,
                            hmmfrom == 1 ? '[':'.',  hmmto == hmmlen ? ']' : '.',
                            sc, evalue);
                    nreported++;
                }
            }
            if (nreported == 0) count+=sprintf(buffer+count, "\t[no hits above thresholds]\n");


            /* 3. Alignment output, also by domain.
             *    dhits is already sorted and namewidth is set, from above code.
             *    Number of displayed alignments is limited by Alimit parameter;
             *    also by domE (evalue threshold), domT (score theshold).
             */
            if (Alimit != 0)
            {
                count+=sprintf(buffer+count, "\nAlignments of top-scoring domains:\n");
                printf("");
                for (i = 0, nreported = 0; i < dhit->num; i++)
                {
                    if (nreported == Alimit) break; /* limit to Alimit output alignments */
                    GetRankedHit(dhit, i,
                            &pvalue, &sc, &motherp, &mothersc,
                            &name, &acc, NULL,
                            &sqfrom, &sqto, NULL,              /* seq position info  */
                            &hmmfrom, &hmmto, &hmmlen,         /* HMM position info  */
                            &domidx, &ndom,                    /* domain info        */
                            &ali);	                      /* alignment info     */
                    evalue = pvalue * (double) thresh->Z;

                    if (motherp * (double) thresh->Z > thresh->globE || mothersc < thresh->globT)
                        continue;
                    else if (evalue <= thresh->domE && sc >= thresh->domT)
                    {
                        count+=sprintf(buffer+count, "%s: domain %d of %d, from %d to %d: score %.1f, E = %.2g\n",
                                (show_acc && acc != NULL) ?  acc : name,
                                domidx, ndom, sqfrom, sqto, sc, evalue);
                        count+=sPrintFancyAli(buffer+count, ali);
                        nreported++;
                    }
                }
                if (nreported == 0)      count+=sprintf(buffer+count, "\t[no hits above thresholds]\n");
                if (nreported == Alimit) count+=sprintf(buffer+count, "\t[output cut off at A = %d top alignments]\n", Alimit);
            }
            count+=sprintf(buffer+count, "--------------//\n");
            
            MPI_Send(buffer, count, MPI_BYTE, 0, 985, mastercomm);
            int offset;
            FreeSequence(seq, sqinfo);
            if(hold==1)
            {
                MPI_Recv(&offset, 1, MPI_INT, 0, 984, mastercomm, MPI_STATUS_IGNORE);
                if(offset!=-1)
                {
                    sqinfo = malloc(sizeof(SQINFO));
                    fseek(sqfp->f, offset, SEEK_SET);
                    ReadSeq(sqfp, sqfp->format, &seq, sqinfo);
                    free(sqinfo);
                    length2=1;
                }
            }
            FreeTophits(ghit);
            FreeTophits(dhit);
            if(singlemode)
            {

                ghit = AllocTophits(4000);
                dhit = AllocTophits(4000);
            }
            fclose(hmmfp->f);
            hmmfp->f = fopen(hmmfile, "r");
        }
        if(length2<=0 && singlemode ==1)
        {       
            for (slave = 0; slave < nslaves; slave++)
            {
                msg = -1;			/* the special "shut down now" flag */
                MPI_Send(&msg, 1, MPI_INT, (slave+2)%cluster, 0, subcomm);
            }
            break;
        }
    }
    free(hmmlist);
    *ret_nhmm = nhmm;
    return;
}

static void main_loop_mpi_slave(char *hmmfile, HMMFILE *hmmfp, char *seq,
        struct threshold_s thresh, int do_xnu,
        int do_forward, int do_null2, struct transporter pack, int nhmm, int rank, char * index,
        MPI_Comm subcomm, MPI_Comm mastercomm, char * seqfile, int format)
{
    struct p7trace_s *tr;
    struct plan7_s *hmm = NULL;
    struct dpmatrix_s *mx = NULL;
    unsigned char *dsq = NULL;
    int len, len2;
    float sc;
    int my_idx = rank%cluster;
    double pvalue;
    double evalue;
    int send_trace;
    int alphatype;
    int code, i;
    int ret_nhmm;
    int offset, length;
    int state;
    struct dpmatrix_s *mx2 = NULL;
    struct threshold_s thresh2;
    SQINFO sqinfo;
    MPI_Status temp;
    MPI_Status *useless = &temp;
    seq = NULL;
    char asdf[10000];
    struct plan7_s **list;


    if(hmmfp==NULL)
    hmmfp = HMMFileOpen(hmmfile, "HMMERDB");
    if(rank!=0)
    {
        MPI_Bcast(&do_xnu, 1, MPI_INT, 0, MPI_COMM_WORLD);
        MPI_Bcast(&do_forward, 1, MPI_INT, 0, MPI_COMM_WORLD);
        MPI_Bcast(&do_null2, 1, MPI_INT, 0, MPI_COMM_WORLD);
        MPI_Bcast(&alphatype, 1, MPI_INT, 0, MPI_COMM_WORLD);
        SQD_DPRINTF1(("My master has told me how to initialize, and I am happy.\n"));

        SetAlphabet(alphatype);
    }

    /*
     * We'll create for at least N=300xM=300, and thus consume at least 1 MB,
     * regardless of RAMLIMIT -- this helps us avoid reallocating some weird
     * asymmetric matrices.
     *
     * We're growable in both M and N, because inside of P7SmallViterbi,
     * we're going to be calling P7Viterbi on subpieces that vary in size,
     * and for different models.
     */

    /* report our status.
     */
    SQD_DPRINTF1(("I have told my master my initialization status and I await his command.\n"));
    list = (struct plan7_s **) malloc(sizeof(int)*10000);
    for(int i=0; i<10000; i++)
    {
        list[i] = NULL;
        asdf[i] = 0;
    }

    for (;;)
    {
        MPI_Recv(&nhmm, 1, MPI_INT, 1, 0, subcomm, useless);
        if (my_idx < 0) my_idx = nhmm; /* first time thru, remember what index we are. */

        if (nhmm == -1) { /* shutdown signal */
            SQD_DPRINTF1(("I've been told to shut down."));
            break;
        }
        if (nhmm == -4)
        {
            slaveCommCreate(&mastercomm, subcomm);
            continue;
        }

        if (nhmm == -3)
        {   if(seq!=NULL) free(seq);
            if(mx!=NULL) FreePlan7Matrix(mx);
            if(dsq!=NULL) free(dsq);
            MPI_Recv(&len, 1, MPI_INT, 1, 770, subcomm, useless);
            seq = MallocOrDie(sizeof(char *) * (len+1));
            MPI_Recv(seq, len, MPI_CHAR, 1, 771, subcomm, useless);
            const int three = 300;
            mx = CreatePlan7Matrix(three, three, 25, 25);
            dsq = DigitizeSequence(seq, len);
            if (do_xnu) XNU(dsq, len);
            continue;
        }

        if (nhmm == -2)
        {
            MPI_Recv(&offset, 1, MPI_INT, 1, 881, subcomm, useless);
            MPI_Recv(&length, 1, MPI_INT, 1, 882, subcomm, useless);
            MPI_Recv(&state, 1, MPI_INT, 1, 883, subcomm, useless);
            main_loop_mpi(hmmfile, hmmfp, seq, &sqinfo, &thresh, do_xnu,
                do_forward, do_null2, pack, NULL, NULL, &ret_nhmm, 0,
                mastercomm, subcomm, NULL, state, offset, length, index, seqfile, format);
            return;
        }
        thresh2 = thresh;
        if(list[nhmm]==NULL)
        {    
            if (! HMMFilePositionByIndex(hmmfp, nhmm)) Die("didn't position the HMM file");
            if (! HMMFileRead(hmmfp, &hmm))            Die("unexpected end of HMM file");
            if (hmm == NULL)                           Die("unexpected failure to parse HMM file");
            list[nhmm] = hmm;
            P7Logoddsify(hmm, TRUE);
        }
       
        else 
            hmm = list[nhmm];  //Database caching: if we have it, don't read it
                                //again.

      /* move to our assigned HMM in the HMM file, and read it
       */
       SQD_DPRINTF1(("The master says to do HMM #%d - I hear and obey\n", nhmm));

                        /* set Pfam specific score thresholds if needed */
        if (! SetAutocuts(&thresh, hmm))
            Die("HMM %s doesn't have the score cutoffs you wanted", hmm->name);
        thresh = thresh2;

        SQD_DPRINTF1(("The master says to do HMM #%d - I hear and obey\n", nhmm));
        if (P7ViterbiSpaceOK(len, hmm->M, mx))
	{
	  SQD_DPRINTF1(("P7Viterbi(): Estimated size %d Mb\n", P7ViterbiSize(len, hmm->M)));
	  sc = P7Viterbi(dsq, len, hmm, mx, &tr);
	}
        else
	{
	  SQD_DPRINTF1(("P7SmallViterbi() called; %d Mb > %d\n", P7ViterbiSize(len, hmm->M), RAMLIMIT));
	  sc = P7SmallViterbi(dsq, len, hmm, mx, &tr);
	}

      /* The Forward score override.
       * See comments in hmmpfam.c in serial version.
       */
      if (do_forward) {
	sc  = P7Forward(dsq, len, hmm, NULL);
	if (do_null2)   sc -= TraceScoreCorrection(hmm, tr, dsq);
      }

      pvalue = PValue(hmm, sc);
      evalue = thresh.Z ? (double) thresh.Z * pvalue : (double) nhmm * pvalue;
        send_trace = (tr != NULL && sc >= thresh.globT && evalue <= thresh.globE) ? 1 : 0;

        /* return output
         */

        MPI_Send(&my_idx, 1, MPI_INT, 1, 1, subcomm);	/* tell master who we are */
        len2 = strlen(hmm->name);
        MPI_Send(&len2, 1, MPI_INT, 1, 2, subcomm);
        MPI_Send(hmm->name, len2, MPI_CHAR, 1, 3, subcomm);	/* double check that we did the right thing */
        MPI_Send(&sc, 1, MPI_FLOAT, 1, 4, subcomm);
        MPI_Send(&pvalue, 1, MPI_DOUBLE, 1, 5, subcomm);
        MPI_Send(&send_trace, 1, MPI_INT, 1, 6, subcomm); /* flag for whether a trace structure is coming */
        if (send_trace)
            mpiSendTrace(tr, 1, subcomm);

        P7FreeTrace(tr);
    }

    /***********************************************
     * Cleanup, return.
     ***********************************************/
}
#else

static void main_loop_mpi(char * hmmfile, HMMFILE *hmmfp, char *seq,
        SQINFO *sqinfo, struct threshold_s *thresh,
        int do_xnu, int do_forward, int do_null2, struct transporter pack,
        struct tophit_s *ghit,
        struct tophit_s *dhit, int *ret_nhmm, int num_nodes,
        MPI_Comm mastercomm, MPI_Comm subcomm, FILE * indexfile, int num_mast, int off,
        int num_seq, char * index, char * seqfile, int format)
{
    Die("No MPI support");
}

static void main_loop_mpi_master(char *hmmfile, HMMFILE *hmmfp, char *seq,
        SQINFO *sqinfo, struct threshold_s *thresh,
        int do_xnu, int do_forward, int do_null2, struct transporter pack,
        struct tophit_s *ghit,
        struct tophit_s *dhit, int *nhmm, int num_nodes, MPI_Comm subcomm,
        MPI_Comm mastercomm, int real_rank)
{
    Die("No MPI support");
}

static void main_loop_mpi_slave(char *hmmfile, HMMFILE *hmmfp, char *seq,
        struct threshold_s thresh, int do_xnu,
        int do_forward, int do_null2, struct transporter pack, int nhmm, int rank, char * index,
        MPI_Comm subcomm, MPI_Comm mastercomm, char * seqfile, int format)
{
    Die("No MPI support");
}
#endif /* HMMER_MPI */

