/************************************************************
 * HMMER - Biological sequence analysis with profile HMMs
 * Copyright (C) 1992-2003 Washington University School of Medicine
 * All Rights Reserved
 *
 *     This source code is distributed under the terms of the
 *     GNU General Public License. See the files COPYING and LICENSE
 *     for details.
 ************************************************************/

/* hmmsearch.c
 * SRE, Tue Jan  7 17:19:20 1997 [St. Louis]
 *
 * Search a sequence database with a profile HMM.
 * Conditionally includes PVM parallelization when HMMER_PVM is defined
 *    at compile time; hmmsearch --pvm runs the PVM version.
 *
 * CVS $Id: hmmsearch.c,v 1.2 2008/03/16 14:52:28 cpsosa Exp $
 */

#define HMMER_MPI

#include "config.h"		/* compile-time configuration constants */
#include "squidconf.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <float.h>
#ifdef HMMER_THREADS
#include <pthread.h>
#endif
#ifdef HMMER_PVM
#include <pvm3.h>
#endif
#ifdef HMMER_MPI
#include <mpi.h>
#endif

#define newMallocOrDie(x)     newsre_malloc(__FILE__, __LINE__, (x))

#include "squid.h"		/* general sequence analysis library    */
#include "structs.h"		/* data structures, macros, #define's   */
#include "funcs.h"		/* function declarations                */
#include "globals.h"		/* alphabet global variables            */

#define sctol 4096  //packet size for master-supermaster communications
#define hitol 32    //group size for single master



static char usage[]  = "\
	Usage: hmmsearch [-options] <hmmfile> <sequence file or database>\n\
	Available options are:\n\
	-h        : help; print brief help on version and usage\n\
	-A <n>    : sets alignment output limit to <n> best domain alignments\n\
	-E <x>    : sets E value cutoff (globE) to <= x\n\
	-T <x>    : sets T bit threshold (globT) to >= x\n\
	-Z <n>    : sets Z (# seqs) for E-value calculation\n\
	";
#ifdef HMMER_MPI
const int HMMMPIWORK1 = 1;
const int HMMMPIWORK2 = 2;
const int HMMMPIWORK3 = 3;
const int HMMMPIRESULTS = 4;
const int TRACECOM = 8;
const int TRACECOM1 = 5;
const int TRACECOM2 = 6;
const int TRACECOM3 = 7;

struct resultpkt
{	
    int sid;
    float sc;
    double pvalue;
    int send_trace;
};

int mpiStringBcast(char * string, int rank)
{	
    int len;
    if(rank==0)
    {       if(string==NULL) return 0;
            len = strlen(string);
    }
    MPI_Bcast(&len, 1, MPI_INT, 0, MPI_COMM_WORLD);
    if(rank!=0) string = malloc(len * sizeof(char));
    if(MPI_Bcast(string, len, MPI_CHAR, 0, MPI_COMM_WORLD) != MPI_SUCCESS) return 0;
    return 1;
}

char * mpiRecvString(int rank, MPI_Comm comm, int tag1, int tag2)
{       
    int len;
    char * buffer;
    MPI_Status temp;
    MPI_Status *useless;

    MPI_Recv(&len, 1, MPI_INT, rank, tag1, comm, useless);
    if(len==0) return NULL;
    buffer = malloc((len+4)*sizeof(char));
    MPI_Recv(buffer, len+1, MPI_CHAR, rank, tag2, comm, useless);
    buffer[len] = '\0';
    return buffer;
}

void mpiSendString(char* sendee, int rank, MPI_Comm comm, int tag1, int tag2)
{       
    int len = (sendee==NULL ? 0 : strlen(sendee));
    MPI_Send(&len, 1, MPI_INT, rank, tag1, comm);
    if(sendee==NULL) return;
    MPI_Send(sendee, len, MPI_CHAR, rank, tag2, comm);
}

int mpiSendTrace(struct p7trace_s *tr, int target, MPI_Comm comm)
{
    if (MPI_Send(&(tr->tlen), 1, MPI_INT, target, 100, comm) != MPI_SUCCESS) return 0;
    if (MPI_Send(tr->statetype, tr->tlen, MPI_BYTE, target, TRACECOM1, comm) != MPI_SUCCESS) return 0;
    if (MPI_Send(tr->nodeidx, tr->tlen, MPI_INT, target, TRACECOM2, comm) !=MPI_SUCCESS) return 0;
    if (MPI_Send(tr->pos, tr->tlen, MPI_INT, target, TRACECOM3, comm) !=MPI_SUCCESS) return 0;
    return 1;
}

struct p7trace_s * mpiRecvTrace(int target, MPI_Comm comm)
{ 	
    struct p7trace_s * tr;
    int tlen;
    MPI_Status temp;
    MPI_Status * useless = &temp;
    MPI_Recv(&tlen, 1, MPI_INT, target, 100, comm, useless);

    P7AllocTrace(tlen, &tr);
    if (MPI_Recv(tr->statetype, tlen, MPI_BYTE, target, TRACECOM1, comm, useless) != MPI_SUCCESS) { P7FreeTrace(tr); return NULL;}
    if (MPI_Recv(tr->nodeidx, tlen, MPI_INT, target, TRACECOM2, comm, useless) != MPI_SUCCESS) { P7FreeTrace(tr); return NULL;}
    if (MPI_Recv(tr->pos, tlen, MPI_INT, target, TRACECOM3, comm, useless) != MPI_SUCCESS) { P7FreeTrace(tr); return NULL;}
    tr->tlen = tlen;
    return tr;
}


struct fancyali_s * mpiRecvAli(int rank, MPI_Comm comm)
{       
    struct fancyali_s *ans;
    MPI_Status temp;
    ans = malloc(sizeof(struct fancyali_s));
    MPI_Recv(ans, sizeof(struct fancyali_s), MPI_BYTE, rank, 119, comm, &temp);
    ans->rfline = mpiRecvString(rank, comm, 120, 121);
    ans->csline = mpiRecvString(rank, comm, 122, 123);
    ans->model = mpiRecvString(rank, comm, 124, 125);
    ans->mline = mpiRecvString(rank, comm, 126, 127);
    ans->query = mpiRecvString(rank, comm, 128, 129);
    ans->target = mpiRecvString(rank, comm, 130, 131);
    ans->aseq = mpiRecvString(rank, comm, 200, 201);
    return ans;
}

void mpiSendAli(struct fancyali_s * ans, int rank, MPI_Comm comm)
{       
    MPI_Send(ans, sizeof(struct fancyali_s), MPI_BYTE, rank, 119, comm);
    mpiSendString(ans->rfline, rank, comm, 120, 121);
    mpiSendString(ans->csline, rank, comm, 122, 123);
    mpiSendString(ans->model, rank, comm, 124, 125);
    mpiSendString(ans->mline, rank, comm, 126, 127);
    mpiSendString(ans->query, rank, comm, 128, 129);
    mpiSendString(ans->target, rank, comm, 130, 131);
    mpiSendString(ans->aseq, rank, comm, 200, 201);
}

/* Obselete, replaced by mpiPackHits and mpiUnpackHits
void mpiRecvHit(struct hit_s * hit, int rank, MPI_Comm comm)
{       
    fprintf(stderr, "");
    hit->name = mpiRecvString(rank, comm, 113, 114);
    hit->acc = mpiRecvString(rank, comm, 115, 116);
    hit->desc = mpiRecvString(rank, comm, 117, 118);
    if((hit->ali)!=NULL){
        fprintf(stderr, "");
        hit->ali = mpiRecvAli(rank, comm);
    }
}

void mpiSendHit(struct hit_s * hit, int rank, MPI_Comm comm)
{       
    mpiSendString(hit->name, rank, comm, 113, 114);
    mpiSendString(hit->acc, rank, comm, 115, 116);
    mpiSendString(hit->desc, rank, comm, 117, 118);
    if((hit->ali)!=NULL)    mpiSendAli(hit->ali, rank, comm);
}
*/

//these functions (pack, unpack) should probably not be regular functions
//because of the function call overhead.  Instead consider macros or inline
//functions.

int packInt(unsigned char * buffer, int num)
{
    memcpy(buffer, &num, sizeof(int));
    return sizeof(int);
}

int unpackInt(unsigned char * buffer, int *num)
{
    memcpy( num, buffer, sizeof(int));
    return sizeof(int);
}

int mpiPackString(char* buffer, char* string)
{
    if(string == NULL)
    {
        return packInt(buffer, 0);
    }
    else
    {
        int len = strlen(string);
        memcpy(buffer + packInt(buffer, len), string, len);
        return len+sizeof(int);
    }
}

int mpiUnpackString(char* buffer, char**string)
{
    int len;
    unpackInt(buffer, &len);
    if(len==0)
    {
        *string = NULL;
        return sizeof(int);
    }
    else
    {
        *string = malloc(sizeof(char)*(len+1));
        memcpy(*string, buffer+sizeof(int), len);
        (*string)[len]='\0';
        return len+sizeof(int);
    }
}

// mpiPackHits()
//
// Used by lower masters to pack ghit and dhit into a buffer and send it
// to the supermaster (rank).  num = number of hits, comm = communicator
    
void mpiPackHits(struct hit_s * hit, int num, int rank, MPI_Comm comm)
{
    char * buffer = (char *) malloc(1024*1024*10);
    int count = 0;
    
    int temp;
    MPI_Comm_rank(MPI_COMM_WORLD, &temp);
    for(int i=0; i<num; i++)
    {
        int len;
        count+=mpiPackString(buffer+count, (hit+i)->name);
        count+=mpiPackString(buffer+count, (hit+i)->acc);
        count+=mpiPackString(buffer+count, (hit+i)->desc);
        if((hit+i)->ali == NULL)
        {
            count+=packInt(buffer+count, 0);
        }
        else
        {
            if((hit+i)->ali->rfline==NULL)
            {
                count+=packInt(buffer+count, -1);
            }
            else
            {
                len = strlen((hit+i)->ali->rfline);
                count+=packInt(buffer+count, len);
                memcpy(buffer+count, (hit+i)->ali->rfline, len);
                count+=len;
            }
            count+=mpiPackString(buffer+count, (hit+i)->ali->csline);
            count+=mpiPackString(buffer+count, (hit+i)->ali->model);
            count+=mpiPackString(buffer+count, (hit+i)->ali->mline);
            count+=mpiPackString(buffer+count, (hit+i)->ali->aseq);
            count+=mpiPackString(buffer+count, (hit+i)->ali->query);
            count+=mpiPackString(buffer+count, (hit+i)->ali->target);
            count+=packInt(buffer+count, (hit+i)->ali->len);
            count+=packInt(buffer+count, (hit+i)->ali->sqfrom);
            count+=packInt(buffer+count, (hit+i)->ali->sqto);
        }
    }

    MPI_Send(&count, 1, MPI_INT, rank, 113, comm);
    MPI_Send(buffer, count, MPI_BYTE, rank, 114, comm);

    free(buffer);
}

// mpiUnpackHits()
//
// Supermaster uses this to unpack the hits (ghit, dhit) from a buffer
// sent by lower master (rank) of communicator (comm).

void mpiUnpackHits(struct hit_s * hit, int num, int rank, MPI_Comm comm)
{
    int count;

    MPI_Recv(&count, 1, MPI_INT, rank, 113, comm, MPI_STATUS_IGNORE);

    char * buffer = (char *) malloc(count);
    MPI_Status temp;
    MPI_Recv(buffer, count, MPI_BYTE, rank, 114, comm, &temp);
    count = 0;

    for(int i=0; i<num; i++)
    {
        int len;
        count+=mpiUnpackString(buffer+count, &((hit+i)->name));
        count+=mpiUnpackString(buffer+count, &((hit+i)->acc));
        count+=mpiUnpackString(buffer+count, &((hit+i)->desc));
        count+=unpackInt(buffer+count, &len);
        if(len==0)
            (hit+i)->ali = NULL;
        else
        {
            (hit+i)->ali = (struct fancyali_s *) malloc(sizeof(struct fancyali_s));
            if(len == -1)
            {
                (hit+i)->ali->rfline = NULL;
            }
            else
            {
                (hit+i)->ali->rfline = (char *) malloc(sizeof(char)*(len+1));
                memcpy((hit+i)->ali->rfline, buffer+count, len);
                ((hit+i)->ali->rfline)[len] = '\0';
                count+=len;
            }
            count+=mpiUnpackString(buffer+count, &((hit+i)->ali->csline));
            count+=mpiUnpackString(buffer+count, &((hit+i)->ali->model));
            count+=mpiUnpackString(buffer+count, &((hit+i)->ali->mline));
            count+=mpiUnpackString(buffer+count, &((hit+i)->ali->aseq));
            count+=mpiUnpackString(buffer+count, &((hit+i)->ali->query));
            count+=mpiUnpackString(buffer+count, &((hit+i)->ali->target));
            count+=unpackInt(buffer+count, &((hit+i)->ali->len));
            count+=unpackInt(buffer+count, &((hit+i)->ali->sqfrom));
            count+=unpackInt(buffer+count, &((hit+i)->ali->sqto));
        }
    }
    free(buffer);
}

// mpiHMMBcast()
//
// This broadcasts hmm from rank (rank).  It does not pack,
// in contrast to the mpiSendHMM function found in hmmpfam

struct plan7_s * mpiHMMBcast(struct plan7_s *hmm, int rank)
{
	int k, M;
	int sendflags;		/* HMM flags to send */
	
	if(rank==0)
	{	
            sendflags = hmm->flags;
	    sendflags &= ~PLAN7_HASBITS;	/* no log odds scores sent */
	    sendflags &= ~PLAN7_HASDNA;	/* no DNA scores sent */
	}
	if(rank==0)
        {    
            if (MPI_Bcast(&(hmm->M), 1, MPI_INT, 0, MPI_COMM_WORLD) !=MPI_SUCCESS)  return NULL;
        }
	else
	{	
            MPI_Bcast(&M, 1, MPI_INT, 0, MPI_COMM_WORLD);
	    hmm = AllocPlan7(M);
	}	

	if (MPI_Bcast(&sendflags, 1, MPI_INT, 0, MPI_COMM_WORLD) != MPI_SUCCESS)
            return NULL;
	if (rank != 0) hmm->flags = sendflags;
	if (!mpiStringBcast(hmm->name, rank)) return NULL;
	if (hmm->flags & PLAN7_DESC) { if (!mpiStringBcast(hmm->desc, rank))
            return NULL; }
	if (hmm->flags & PLAN7_RF)   { if (!mpiStringBcast(hmm->rf, rank))
            return NULL; }
	if (hmm->flags & PLAN7_CS)   { if (!mpiStringBcast(hmm->cs, rank))
            return NULL; }
	if (! mpiStringBcast(hmm->comlog, rank))  return NULL;
	if (MPI_Bcast(&(hmm->nseq), 1, MPI_INT, 0, MPI_COMM_WORLD) !=
            MPI_SUCCESS) return NULL;
	if (!mpiStringBcast(hmm->ctime, rank)) return NULL;
	if (hmm->flags & PLAN7_MAP) { if (MPI_Bcast(hmm->map, hmm->M+1, MPI_INT,
            0, MPI_COMM_WORLD) != MPI_SUCCESS) return NULL; }
	if (MPI_Bcast(&(hmm->checksum), 1, MPI_INT, 0, MPI_COMM_WORLD) !=
            MPI_SUCCESS) return NULL;

	for (k = 1; k < hmm->M; k++)
		if (MPI_Bcast(hmm->t[k], 7, MPI_FLOAT, 0, MPI_COMM_WORLD) !=
                    MPI_SUCCESS) return NULL;
	for (k = 1; k <= hmm->M; k++)
		if (MPI_Bcast(hmm->mat[k], Alphabet_size, MPI_FLOAT, 0,
                    MPI_COMM_WORLD) != MPI_SUCCESS) return NULL;
	for (k = 1; k < hmm->M; k++)
		if (MPI_Bcast(hmm->ins[k], Alphabet_size, MPI_FLOAT, 0,
                    MPI_COMM_WORLD) != MPI_SUCCESS) return NULL;
	if (MPI_Bcast(&(hmm->tbd1), 1, MPI_FLOAT, 0, MPI_COMM_WORLD) !=
            MPI_SUCCESS) return NULL;
	for (k = 0; k < 4; k++)
		if (MPI_Bcast(hmm->xt[k], 2, MPI_FLOAT, 0, MPI_COMM_WORLD) !=
                    MPI_SUCCESS) return NULL;
	if (MPI_Bcast(hmm->begin, hmm->M+1, MPI_FLOAT, 0, MPI_COMM_WORLD) !=
            MPI_SUCCESS) return NULL;
	if (MPI_Bcast(hmm->end,   hmm->M+1, MPI_FLOAT, 0, MPI_COMM_WORLD) !=
            MPI_SUCCESS) return NULL;
	if (MPI_Bcast(hmm->null,  Alphabet_size, MPI_FLOAT, 0, MPI_COMM_WORLD)
            != MPI_SUCCESS) return NULL;
	if (MPI_Bcast(&(hmm->p1), 1, MPI_FLOAT, 0, MPI_COMM_WORLD) !=
            MPI_SUCCESS) return NULL;
	if (hmm->flags & PLAN7_STATS)
	{
		if (MPI_Bcast(&(hmm->mu), 1, MPI_FLOAT, 0, MPI_COMM_WORLD) !=
                    MPI_SUCCESS) return NULL;
		if (MPI_Bcast(&(hmm->lambda), 1, MPI_FLOAT, 0, MPI_COMM_WORLD)
                    != MPI_SUCCESS) return NULL;
	}
	return hmm;
}


	void *
newsre_malloc(char *file, int line, size_t size)
{
	void *ptr;

	SQD_DPRINTF3(("MALLOC: %d bytes (file %s line %d)\n", size, file, line));
	if ((ptr = malloc (size)) == NULL)
		Die("malloc of %ld bytes failed: file %s line %d", size, file, line);
	return ptr;
}

// INITIAL is the percentage of the database initially distributed.
// TIMES is the multiple of SCTOL number of sequences given to a master at a time
// during load balancing.  It is roughly equivalent to the number of times the master
// will flush its buffer back to the supermaster before asking for more work.

#define INITIAL  0.7
#define TIMES 5

/*****************************************************************
 * MPI specific functions
 ****************************************************************/

/* Function: main_loop_mpi_master()
 * 
 *
 * Purpose:  Coordinate the main_loop_mpi()s 
 *
 *           In:   HMM and index file, plus options
 *                 Communicators, and sequence file
 *           Out:  histogram, global hits list, domain hits list, nseq.
 *
 * Args:     hmm        - the HMM to search with. scoring form.
 *           index      - name of sequence index file
 *           thresh     - score/evalue threshold information
 *           do_forward - TRUE to score using Forward()
 *           do_null2   - TRUE to use ad hoc null2 score correction
 *           do_xnu     - TRUE to apply XNU mask
 *           histogram  - RETURN: score histogram
 *           ghit       - RETURN: ranked global scores
 *           dhit       - RETURN: ranked domain scores
 *           ret_nseq   - RETURN: actual number of seqs searched
 *           subcomm    - Communicator shared with slaves
 *           mastercomm - Communicator shared with supermaster
 *
 * Returns:  (void)
 */

//supermaster function

static void
main_loop_master_mpi(struct plan7_s *hmm, FILE *index, struct threshold_s *thresh, int do_forward, 
     int do_null2, int do_xnu, struct histogram_s *histogram, struct tophit_s *ghit, 
    struct tophit_s *dhit, int *ret_nseq, int mpi_threads, MPI_Comm mastercomm)
{       
        int offset, totals, total, check, next, sofar;
        int masters, mastcount;
        int sqpermaster;
        int diff, i, temp, sum, j, thresh1;
        float *scores;
        int rank;
        float *scarray;
        
        MPI_Status stat;
        MPI_Status *useless = &stat;
        
	MPI_Bcast(&(thresh->globT), 1, MPI_FLOAT, 0, MPI_COMM_WORLD);
	MPI_Bcast(&(thresh->globE), 1, MPI_DOUBLE, 0, MPI_COMM_WORLD);
	MPI_Bcast(&(thresh->Z), 1, MPI_INT, 0, MPI_COMM_WORLD);
	MPI_Bcast(&do_forward, 1, MPI_INT, 0, MPI_COMM_WORLD);
	MPI_Bcast(&do_null2,   1, MPI_INT, 0, MPI_COMM_WORLD);
	MPI_Bcast(&Alphabet_type, 1, MPI_INT, 0, MPI_COMM_WORLD);
        
        mpiHMMBcast(hmm, 0);
        fscanf(index, "%d", &totals);
        total = totals;
        sofar = INITIAL*totals;
        masters = (mpi_threads - 1)/32 + 1;
        sqpermaster = sofar/masters;
        scarray = malloc(sctol*sizeof(float));
        diff = sofar - masters*sqpermaster;
        mastcount = masters;
        temp = 0;
        thresh1 = TIMES * sctol;
        for(i=0; i<masters; i++)
        {       
            if(i==0)
            {      
                MPI_Send(&temp, 1, MPI_INT, 1, 101, MPI_COMM_WORLD);
               diff = diff + sqpermaster;
               if(diff!=sqpermaster)
                   diff = -diff;
               MPI_Send(&diff, 1, MPI_INT, i*32+1, 102, MPI_COMM_WORLD);
               if(diff < 0)
                   MPI_Send(&sqpermaster, 1, MPI_INT, i*32+1, 107, MPI_COMM_WORLD);
            }
            else
            {      
                fseek(index, (sqpermaster*i+1)*12, SEEK_SET);
                fscanf(index, "%d", &offset);
                MPI_Send(&offset, 1, MPI_INT, i*32+1, 101, MPI_COMM_WORLD);
                MPI_Send(&sqpermaster, 1, MPI_INT, i*32+1, 102, MPI_COMM_WORLD);
            }
        }
        MPI_Comm_rank(mastercomm, &rank);
        while(mastcount>0)
        {       
            MPI_Recv(scarray, sctol, MPI_FLOAT, MPI_ANY_SOURCE, 150, mastercomm, useless);
            
            for(i=useless->count/4-1; i>=0; i--)
                AddToHistogram(histogram, scarray[i]);
            total-=useless->count/4;
            
            
            MPI_Recv(&check, 1, MPI_INT, useless->MPI_SOURCE, 151, mastercomm, useless);
            
            if(check==1 || check==3) // master sending dhit
            {   
                MPI_Recv(dhit->unsrt + dhit->num, 500*sizeof(struct hit_s),
                    MPI_BYTE, useless->MPI_SOURCE, 152, mastercomm,
                    useless);
                next = useless->count/sizeof(struct hit_s);
                mpiUnpackHits(dhit->unsrt+dhit->num, next, useless->MPI_SOURCE, mastercomm);
                dhit->num += next;
            }
            if(check==2 || check==3) // master sending ghit
            {   
                MPI_Recv(ghit->unsrt + ghit->num, 2500*sizeof(struct hit_s),
                    MPI_BYTE, useless->MPI_SOURCE, 999, mastercomm, useless);
                next = useless->count/sizeof(struct hit_s);
                mpiUnpackHits(ghit->unsrt+ghit->num, next, useless->MPI_SOURCE, mastercomm);
                ghit->num += next;
            }
            if(check==3) // master quiting
                mastcount--;
            if(check==4) //master wants more work
            {  
                fseek(index, (sofar+1)*12, SEEK_SET);
                fscanf(index, "%d", &check);
                if(sofar==0)
                {   
                    check=0;
                    MPI_Send(&check, 1, MPI_INT, useless->MPI_SOURCE, 330,
                    mastercomm);
                }
                else if(totals - sofar < TIMES*sctol)
                {   
                    sofar = totals - sofar;
                    if(sofar < 0) check = 0;
                    MPI_Send(&check, 1, MPI_INT, useless->MPI_SOURCE, 330,
                    mastercomm);
                    if(sofar > 0)
                        MPI_Send(&sofar, 1, MPI_INT, useless->MPI_SOURCE, 331,
                        mastercomm);
                    sofar = 0;
                }
                else
                {   
                    MPI_Send(&check, 1, MPI_INT, useless->MPI_SOURCE,
                    330, mastercomm);
                    MPI_Send(&thresh1, 1, MPI_INT, useless->MPI_SOURCE, 331,
                    mastercomm);
                    sofar+=thresh1;
                }
            }
        }
        free(scarray);
       *ret_nseq = totals; 
}

/*****************************************************************
 * MPI specific functions
 ****************************************************************/

/* Function: main_loop_mpi()
 * 
 *
 * Purpose:  Search an HMM against a sequence database.
 *           main loop for the MPI version.
 *
 *           In:   HMM and sequence file, plus options
 *                 Communicators, and sequence file
 *           Out:  histogram, global hits list, domain hits list, nseq.
 *
 * Args:     hmm        - the HMM to search with. scoring form.
 *           seqfile    - name of sequence file
 *           thresh     - score/evalue threshold information
 *           do_forward - TRUE to score using Forward()
 *           do_null2   - TRUE to use ad hoc null2 score correction
 *           do_xnu     - TRUE to apply XNU mask
 *           histogram  - RETURN: score histogram
 *           ghit       - RETURN: ranked global scores
 *           dhit       - RETURN: ranked domain scores
 *           ret_nseq   - RETURN: actual number of seqs searched
 *           subcomm    - Communicator shared with slaves
 *           mastercomm - Communicator shared with supermaster
 *
 * Returns:  (void)
 */

 // master function
 
static void
main_loop_mpi(struct plan7_s *hmm, char *seqfile, int format, struct threshold_s *thresh, int do_forward,
		int do_null2, int do_xnu, struct histogram_s *histogram,
		struct tophit_s *ghit, struct tophit_s *dhit, int *ret_nseq, int
                mpi_threads, MPI_Comm subcomm, int real_rank, MPI_Comm mastercomm)
{
	char *seq;                    /* target sequence */
	unsigned char *dsq;           /* digitized target seq */
	SQINFO sqinfo;                /* optional info about target seq */
	int   nslaves = mpi_threads-(real_rank==1 ? 2 :1);		/* number of slaves      */
	int   code;			/* status code rec'd from a slave */
	int   nseq;			/* number of sequences searched */
	int   sent_trace;		/* TRUE if slave gave us a trace */
	unsigned char **dsqlist;      /* remember what seqs slaves are doing */
	char **namelist;              /* remember what seq names slaves are doing */
	char **acclist ;              /* remember what seq accessions slaves are doing */
	char **desclist;              /* remember what seq desc's slaves are doing */
	int   *lenlist;               /* remember lengths of seqs slaves are doing */
	int    slaveidx;		/* counter for slaves */
	float  sc;			/* score of an alignment */
	double pvalue;		/* P-value of a score of an alignment */
	struct p7trace_s *tr;         /* Viterbi traceback of an alignment */
	int    i, offset, number, number2;			/* generic counter */
        int   alpha_type;
        SQFILE *sqfp;
        float * scarray;
        float dummy;
        double dummy2;
        int dummy3;
        int count = 0, ghitc = 0, dhitc = 0;
	struct resultpkt results;
        MPI_Status temp;
	MPI_Status * useless = &temp;
        struct plan7_s *dummyhmm;
        int addon=0, seqper, mess;
        int cont= 1;
        int nextlen, nextoff;
	
	/* Alloc arrays for remembering what seq each
	 * slave was working on.
	 */
         
	if ((sqfp = SeqfileOpen(seqfile, format, "BLASTDB")) == NULL)
	{

	    Die("Failed to open sequence database file %s\n%s\n", seqfile, usage);

	}

	MPI_Bcast(&dummy, 1, MPI_FLOAT, 0, MPI_COMM_WORLD);
	MPI_Bcast(&dummy2, 1, MPI_DOUBLE, 0, MPI_COMM_WORLD);
	MPI_Bcast(&dummy3, 1, MPI_INT, 0, MPI_COMM_WORLD);
	MPI_Bcast(&do_forward, 1, MPI_INT, 0, MPI_COMM_WORLD);
	MPI_Bcast(&do_null2,   1, MPI_INT, 0, MPI_COMM_WORLD);
	MPI_Bcast(&alpha_type, 1, MPI_INT, 0, MPI_COMM_WORLD);
        SetAlphabet(alpha_type);
        mpiHMMBcast(dummyhmm, 1);
	namelist = MallocOrDie(sizeof(char *) * nslaves + 1);
	acclist  = MallocOrDie(sizeof(char *) * nslaves + 1);
	desclist = MallocOrDie(sizeof(char *) * nslaves + 1);
	dsqlist  = MallocOrDie(sizeof(unsigned char *) * nslaves + 1);
	lenlist  = MallocOrDie(sizeof(int) * nslaves + 1);
	/* Load the slaves.
	 * Give them all a sequence number and a digitized sequence
	 * to work on.
	 * A side effect of the seq number is that we assign each slave
	 * a number from 0..nslaves-1.
	 */

        MPI_Recv(&offset, 1, MPI_INT, 0, 101, MPI_COMM_WORLD, useless);
        MPI_Recv(&number, 1, MPI_INT, 0, 102, MPI_COMM_WORLD, useless);
        if(number < 0)
        {       number = -number;
                addon = 1;
                MPI_Recv(&seqper, 1, MPI_INT, 0, 107, MPI_COMM_WORLD, useless);
        }
        else seqper = number;
        number2= number;
        scarray = malloc(sctol*sizeof(float));
        
        if(offset!=0)
        {    
             fseek(sqfp->f, offset, SEEK_SET);
             ReadSeq(sqfp, sqfp->format, &seq, &sqinfo);
        }
        
	for (nseq =  0; nseq < nslaves; nseq++)
	{
		if (! ReadSeq(sqfp, sqfp->format, &seq, &sqinfo)) break;
		number--;
                if(number<0) break;
                if (sqinfo.len == 0) { nseq--; continue; }
                
		dsq = DigitizeSequence(seq, sqinfo.len);
		if (do_xnu && Alphabet_type == hmmAMINO) XNU(dsq, sqinfo.len);
                i = (nseq+2)%32;
		MPI_Send(&nseq, 1, MPI_INT, i, HMMMPIWORK1, subcomm);
		MPI_Send(&(sqinfo.len), 1, MPI_INT, i, HMMMPIWORK2, subcomm);
		MPI_Send((char *) dsq, sqinfo.len+2, MPI_BYTE, i, HMMMPIWORK3, subcomm); /* is this char * casting safe?  */
		SQD_DPRINTF1(("sent a dsq : %d bytes\n", sqinfo.len+2));

		namelist[i] = Strdup(sqinfo.name);
		acclist[i]  = (sqinfo.flags & SQINFO_ACC)  ? Strdup(sqinfo.acc)  : NULL;
		desclist[i] = (sqinfo.flags & SQINFO_DESC) ? Strdup(sqinfo.desc) : NULL;
		lenlist[i]  = sqinfo.len;
		dsqlist[i]  = dsq;

		FreeSequence(seq, &sqinfo);
	}
	SQD_DPRINTF1(("%d slaves are loaded\n", nseq));
	/* main receive/send loop
	 */
	while (ReadSeq(sqfp, sqfp->format, &seq, &sqinfo))
	{        number--;
                if(number<0) break;
		if (sqinfo.len == 0) { continue; }
		nseq++;
		/* receive output */
		SQD_DPRINTF1(("Waiting for a slave to give me output...\n"));
		MPI_Recv(&results, sizeof(results), MPI_BYTE, MPI_ANY_SOURCE, HMMMPIRESULTS, subcomm, useless);
		slaveidx = results.sid;
		sc = results.sc;
		pvalue = results.pvalue;
		sent_trace = results.send_trace;
		tr = (sent_trace) ? mpiRecvTrace(slaveidx, subcomm) : NULL;

		/* send new work */
		dsq = DigitizeSequence(seq, sqinfo.len);
		if (do_xnu) XNU(dsq, sqinfo.len);
		MPI_Send(&nseq, 1, MPI_INT, slaveidx,  HMMMPIWORK1, subcomm);
		MPI_Send(&(sqinfo.len), 1, MPI_INT, slaveidx, HMMMPIWORK2, subcomm);
		MPI_Send((char *) dsq, sqinfo.len+2, MPI_BYTE, slaveidx, HMMMPIWORK3, subcomm); /* is this char * casting safe?  */

		/* process output */
		if (sent_trace)
		{
                        sc = PostprocessSignificantHit(ghit, dhit,
					tr, hmm, dsqlist[slaveidx], lenlist[slaveidx],
					namelist[slaveidx], acclist[slaveidx], desclist[slaveidx],
					do_forward, sc,
					do_null2,
					thresh,
					FALSE); /* FALSE-> not hmmpfam mode, hmmsearch mode */
			P7FreeTrace(tr);
		}
                scarray[count] = sc;
                
                count++;
                if(count==sctol)
                {       
                        MPI_Send(scarray, sctol, MPI_FLOAT, 0, 150, mastercomm);
                        count=0;
                        if(number <= 3*sctol && cont==1)
                        {    
                            mess = 4;
                            MPI_Send(&mess, 1, MPI_INT, 0, 151, mastercomm);
                            MPI_Recv(&nextoff, 1, MPI_INT, 0, 330, mastercomm,
                            useless);
                            if(nextoff==0)
                                cont = 0;
                            else 
                            {    
                                MPI_Recv(&nextlen, 1, MPI_INT, 0, 331,
                                mastercomm, useless);
                                cont = 2;
                            }
                        }
                        else if(dhit->num - dhitc >hitol)
                        {   mess=1;   
                            MPI_Send(&mess, 1, MPI_INT, 0, 151, mastercomm);
                            MPI_Send(dhit->unsrt + dhitc, (dhit->num - dhitc)*sizeof(struct hit_s), MPI_BYTE, 0, 152, mastercomm);
                            mpiPackHits(dhit->unsrt+dhitc, dhit->num - dhitc, 0, mastercomm);
                            dhitc = dhit->num;
                        }
                        else if(ghit->num - ghitc >hitol)
                        {   mess=2;
                            MPI_Send(&mess, 1, MPI_INT, 0, 151, mastercomm);
                            MPI_Send(ghit->unsrt + ghitc, (ghit->num -
                            ghitc)*sizeof(struct hit_s), MPI_BYTE, 0, 999, mastercomm);
                            
                            mpiPackHits(ghit->unsrt+ghitc, ghit->num - ghitc, 0, mastercomm);
                            ghitc = ghit->num;
                        }
                        else
                        {   mess = 0;
                            MPI_Send(&mess, 1, MPI_INT, 0, 151, mastercomm);
                        }
                }
		/* record seq info for seq we just sent */
	        if(slaveidx!=7)	
                free(namelist[slaveidx]);
		if (acclist[slaveidx]  != NULL) free(acclist[slaveidx]);
		if (desclist[slaveidx] != NULL) free(desclist[slaveidx]);
		free(dsqlist[slaveidx]);

		dsqlist[slaveidx]  = dsq;
		namelist[slaveidx] = Strdup(sqinfo.name);
		acclist[slaveidx]  = (sqinfo.flags & SQINFO_ACC)  ? Strdup(sqinfo.acc)  : NULL;
		desclist[slaveidx] = (sqinfo.flags & SQINFO_DESC) ? Strdup(sqinfo.desc) : NULL;
		lenlist[slaveidx]  = sqinfo.len;
		FreeSequence(seq, &sqinfo);
                if(number<=0 && cont>0)
                {
                     fseek(sqfp->f, nextoff, SEEK_SET);
                     ReadSeq(sqfp, sqfp->format, &seq, &sqinfo);
                     number = nextlen;
                     cont=1;
                }
	}
	SQD_DPRINTF1(("End of receive/send loop\n"));
	/* Collect the output. All n slaves are still working.
	 */
	for (i =  0; i < nslaves && i < nseq; i++)
	{
		/* don't check slaves (they're exiting normally);
		   window of vulnerability here to slave crashes */
		/* receive output */
		MPI_Recv(&results, sizeof(results), MPI_BYTE, MPI_ANY_SOURCE, HMMMPIRESULTS, subcomm, useless);
                slaveidx = results.sid;
		sc = results.sc;
		pvalue = results.pvalue;
		sent_trace = results.send_trace;

		tr = (sent_trace) ? mpiRecvTrace(slaveidx, subcomm) : NULL;

		SQD_DPRINTF1(("Slave %d finished %s for me...\n", slaveidx, namelist[slaveidx]));
		/* process output */
		if (sent_trace)
		{       
			sc = PostprocessSignificantHit(ghit, dhit,
					tr, hmm, dsqlist[slaveidx], lenlist[slaveidx],
					namelist[slaveidx], acclist[slaveidx], desclist[slaveidx],
					do_forward, sc,
					do_null2,
					thresh,
					FALSE); /* FALSE-> not hmmpfam mode, hmmsearch mode */
			P7FreeTrace(tr);
		}
                scarray[count] = sc;
                count++;
                if(count==sctol)
                {       
                        MPI_Send(scarray, sctol, MPI_FLOAT, 0, 150, mastercomm);
                        count=0;
                        if(dhit->num - dhitc >hitol)
                        {   mess=1;   
                            MPI_Send(&mess, 1, MPI_INT, 0, 151, mastercomm);
                            MPI_Send(dhit->unsrt + dhitc, (dhit->num - dhitc)*sizeof(struct hit_s), MPI_BYTE, 0, 152, mastercomm);
                            mpiPackHits(dhit->unsrt+dhitc, dhit->num-dhitc, 0, mastercomm);
                            dhitc = dhit->num;
                        }
                        else if(ghit->num - ghitc >hitol)
                        {   mess=2;
                            MPI_Send(&mess, 1, MPI_INT, 0, 151, mastercomm);
                            MPI_Send(ghit->unsrt + ghitc, (ghit->num -
                                ghitc)*sizeof(struct hit_s), MPI_BYTE, 0, 999, mastercomm);
                            
                            mpiPackHits(ghit->unsrt+ghitc, ghit->num-ghitc, 0, mastercomm);
                            ghitc = ghit->num;
                        }
                        else 
                        {   mess = 0;
                            MPI_Send(&mess, 1, MPI_INT, 0, 151, mastercomm);
                        }
                }
		free(namelist[slaveidx]);
		if (acclist[slaveidx]  != NULL) free(acclist[slaveidx]);
		if (desclist[slaveidx] != NULL) free(desclist[slaveidx]);
                    free(dsqlist[slaveidx]);
		/* send cleanup/shutdown flag to slave */

		code = -1;
		MPI_Send(&code, 1, MPI_INT, slaveidx, HMMMPIWORK1, subcomm);
	}
        
        
        MPI_Send(scarray, count, MPI_FLOAT, 0, 150, mastercomm);
        mess = 3;
        
        MPI_Send(&mess, 1, MPI_INT, 0, 151, mastercomm);
        MPI_Send(dhit->unsrt + dhitc, (dhit->num - dhitc)*sizeof(struct hit_s), MPI_BYTE, 0, 152, mastercomm);
        mpiPackHits(dhit->unsrt+dhitc, dhit->num-dhitc, 0, mastercomm);
        MPI_Send(ghit->unsrt + ghitc, (ghit->num - ghitc)*sizeof(struct hit_s), MPI_BYTE, 0, 999, mastercomm);
        mpiPackHits(ghit->unsrt+ghitc, ghit->num-ghitc, 0, mastercomm);
        for(i = nseq; i < nslaves; i++)
        {       
                code = -1;
                MPI_Send(&code, 1, MPI_INT, i+1, HMMMPIWORK1, subcomm);
        }
    
	/* Cleanup; quit the VM; and return
	 */
        *ret_nseq = nseq;
	return;
}

void main_loop_mpi_slave(int real_rank, int rank, int num_threads, int do_mpi,
	int   do_null2,		/* TRUE to adjust scores with null model #2 */
	int   do_forward,		/* TRUE to use Forward() not Viterbi()      */
	int   do_xnu,		/* TRUE to filter sequences thru XNU        */
	int   be_backwards,		/* TRUE to be backwards-compatible inoutput*/
        MPI_Comm subcomm)
{       char    *hmmfile;             /* file to read HMM(s) from                */
	HMMFILE *hmmfp;               /* opened hmmfile for reading              */
	char    *seqfile;             /* file to read target sequence(s) from    */
	SQFILE   *sqfp;               /* opened seqfile for reading              */
	int       format;             /* format of seqfile                       */
	int       i;
	struct plan7_s  *hmm;         /* HMM to search with                      */
	struct histogram_s *histogram;/* histogram of all scores                 */
	struct fancyali_s *ali;       /* displayed alignment info                */
	struct tophit_s   *ghit;      /* list of top hits for whole sequences    */
	struct tophit_s   *dhit;      /* list of top hits for domains            */

	float  globT;			/* T parameter: keep only hits > globT bits */
	double globE;			/* E parameter: keep hits < globE E-value   */
	int my_idx;			/* my slave index: 0..nslaves-1, master assigns */
	int L;			/* length of sequence */
	unsigned char *dsq;           /* digitized sequence 1..L */
	struct dpmatrix_s *mx;        /* growable DP matrix */
	struct p7trace_s *tr;         /* trace structure for a Viterbi alignment */

	int alphatype;		/* alphabet type */

	float   sc;	        /* score of an HMM search                  */
	double  pvalue;		/* pvalue of an HMM score                  */
	double  evalue;		/* evalue of an HMM score                  */
	double  motherp;		/* pvalue of a whole seq HMM score         */
	float   mothersc;		/* score of a whole seq parent of domain   */
	int     sqfrom, sqto;	/* coordinates in sequence                 */
	int     hmmfrom, hmmto;	/* coordinate in HMM                       */
	char   *name, *desc;        /* hit sequence name and description       */
	int     sqlen;		/* length of seq that was hit              */
	int     nseq;		/* number of sequences searched            */
	int     Z;			/* # of seqs for purposes of E-val calc    */
	int     domidx;		/* number of this domain                   */
	int     ndom;		/* total # of domains in this seq          */
	int     namewidth;		/* max width of sequence name              */
	int     descwidth;		/* max width of description */
	int     nreported;		/* # of hits reported in a list            */

	int    Alimit;		/* A parameter limiting output alignments   */
	struct threshold_s thresh;  /* contains all threshold (cutoff) info     */

	char *optname;              /* name of option found by Getopt()         */
	char *optarg;               /* argument found by Getopt()               */
	int   optind;               /* index in argv[]                          */
	MPI_Status temp;
        MPI_Status *useless = &temp;
	int   send_trace;
	struct resultpkt newresult;

        if(!do_mpi)
                return;

        /*****************************************************************
         * Initialization.
         * Master broadcasts the problem to us:
         * globT, globE, Z, do_forward, do_null2, alphabet type, HMM,
         ******************************************************************/

        my_idx     = real_rank;

        /* wait for a HMMPVM_INIT message, and unpack it;
         * get options, set alphabet type, get HMM.
         */

        
        MPI_Bcast(&globT, 1, MPI_FLOAT, 0, MPI_COMM_WORLD);
        MPI_Bcast(&globE, 1, MPI_DOUBLE, 0, MPI_COMM_WORLD);
        MPI_Bcast(&Z,          1, MPI_INT, 0, MPI_COMM_WORLD);
        MPI_Bcast(&do_forward, 1, MPI_INT, 0, MPI_COMM_WORLD);
        MPI_Bcast(&do_null2,   1, MPI_INT, 0, MPI_COMM_WORLD);
        MPI_Bcast(&alphatype,  1, MPI_INT, 0, MPI_COMM_WORLD);
        SetAlphabet(alphatype);
        hmm = mpiHMMBcast(hmm, rank);
        mx = CreatePlan7Matrix(1, hmm->M, 25, 0);

       P7Logoddsify(hmm, TRUE);
        
        /* tell the master we're OK and ready to go (or not)
         */

        /*****************************************************************
         * Main loop.
         * Receive a digitized sequence to search against.
         *****************************************************************/

        for (;;)
        {
                SQD_DPRINTF1(("Slave about to do a blocking receive, waiting for input.\n"));
                MPI_Recv(&nseq, 1, MPI_INT, 1, HMMMPIWORK1, subcomm, useless);
                if (nseq == -1) break;	/* shutdown signal */
                MPI_Recv(&L,    1, MPI_INT, 1, HMMMPIWORK2, subcomm, useless);
                SQD_DPRINTF1(("Slave received nseq=%d L=%d my_idx=%d\n", nseq, L, my_idx));
                dsq = MallocOrDie(sizeof(unsigned char) * (L + 2));
                MPI_Recv((char *) dsq, L+2, MPI_BYTE, 1, HMMMPIWORK3,
                    subcomm, useless); /* char * cast is dubious practice */
                SQD_DPRINTF1(("Slave unpacked a seq of %d bytes; beginning processing\n", L+2));
                /* Score sequence, do alignment (Viterbi), recover trace
                 */
                if (P7ViterbiSpaceOK(L, hmm->M, mx))
                    sc = P7Viterbi(dsq, L, hmm, mx, &tr);
                else
                    sc = P7SmallViterbi(dsq, L, hmm, mx, &tr);
                
                if (do_forward) {
                        sc  = P7Forward(dsq, L, hmm, NULL);
                        if (do_null2)   sc -= TraceScoreCorrection(hmm, tr, dsq);
                }
                pvalue = PValue(hmm, sc);
                evalue = Z ? (double) Z * pvalue : (double) nseq * pvalue;
                send_trace = (tr != NULL && sc >= globT && evalue <= globE) ? 1 : 0;

                /* return output
                 */
                SQD_DPRINTF1(("Slave has a result (sc = %.1f); sending back to master\n", sc));

                newresult.sid = my_idx;
                newresult.sc = sc;
                newresult.pvalue = pvalue;
                newresult.send_trace = send_trace;
                MPI_Send(&newresult, sizeof(newresult), MPI_BYTE, 1, HMMMPIRESULTS, subcomm);
                if (send_trace)
                        mpiSendTrace(tr, 1, subcomm);
                /* cleanup
                 */
                free(dsq);
                P7FreeTrace(tr);
        }

        /***********************************************
         * Cleanup, return.
         ***********************************************/
        SQD_DPRINTF1(("Slave is done; performing a normal exit.\n"));
        FreePlan7Matrix(mx);
        FreePlan7(hmm);
	SqdClean();
}
#endif

static char banner[] = "hmmsearch - search a sequence database with a profile HMM";

static char experts[] = "\
	--compat       : make best effort to use last version's output style\n\
	--cpu <n>      : run <n> threads in parallel (if threaded)\n\
	--cut_ga       : use Pfam GA gathering threshold cutoffs\n\
	--cut_nc       : use Pfam NC noise threshold cutoffs\n\
	--cut_tc       : use Pfam TC trusted threshold cutoffs\n\
	--domE <x>     : sets domain Eval cutoff (2nd threshold) to <= x\n\
	--domT <x>     : sets domain T bit thresh (2nd threshold) to >= x\n\
	--forward      : use the full Forward() algorithm instead of Viterbi\n\
	--informat <s> : sequence file is in format <s>\n\
	--null2        : turn OFF the post hoc second null model\n\
	--pvm          : run on a Parallel Virtual Machine (PVM)\n\
	--xnu          : turn ON XNU filtering of target protein sequences\n\
	--mpi          : run using Message Passing Interface on an MPI system\n\
	";

static struct opt_s OPTIONS[] = {
	{ "-h",        TRUE,  sqdARG_NONE },
	{ "-A",        TRUE,  sqdARG_INT  },
	{ "-E",        TRUE,  sqdARG_FLOAT},
	{ "-T",        TRUE,  sqdARG_FLOAT},
	{ "-Z",        TRUE,  sqdARG_INT  },
	{ "--compat",  FALSE, sqdARG_NONE },
	{ "--cpu",     FALSE, sqdARG_INT  },
	{ "--cut_ga",  FALSE, sqdARG_NONE },
	{ "--cut_nc",  FALSE, sqdARG_NONE },
	{ "--cut_tc",  FALSE, sqdARG_NONE },
	{ "--domE",    FALSE, sqdARG_FLOAT},
	{ "--domT",    FALSE, sqdARG_FLOAT},
	{ "--forward", FALSE, sqdARG_NONE },
	{ "--informat",FALSE, sqdARG_STRING},
	{ "--null2",   FALSE, sqdARG_NONE },
	{ "--pvm",     FALSE, sqdARG_NONE },
	{ "--mpi",     FALSE, sqdARG_NONE },
	{ "--xnu",     FALSE, sqdARG_NONE },

};
#define NOPTIONS (sizeof(OPTIONS) / sizeof(struct opt_s))


static void main_loop_serial(struct plan7_s *hmm, SQFILE *sqfp, struct threshold_s *thresh,
		int do_forward, int do_null2, int do_xnu,
		struct histogram_s *histogram, struct tophit_s *ghit,
		struct tophit_s *dhit, int *ret_nseq);
static void main_loop_pvm(struct plan7_s *hmm, SQFILE *sqfp, struct threshold_s *thresh,
		int do_forward, int do_null2, int do_xnu,
		struct histogram_s *histogram, struct tophit_s *ghit,
		struct tophit_s *dhit, int *ret_nseq);
static void main_loop_threaded(struct plan7_s *hmm, SQFILE *sqfp, struct threshold_s *thresh,
		int do_forward, int do_null2, int do_xnu, int num_threads,
		struct histogram_s *histogram, struct tophit_s *ghit,
		struct tophit_s *dhit, int *ret_nseq);

#ifdef HMMER_THREADS
/* POSIX threads version:
 * the threads share a workpool_s structure amongst themselves,
 * for obtaining locks on input HMM file and output histogram and
 * tophits structures.
 */
struct workpool_s {
	/* Shared configuration resources which don't change:
	 */
	struct plan7_s *hmm;		/* HMM to search with              */
	int    do_xnu;		/* TRUE to apply XNU filter  */
	int    do_forward;		/* TRUE to score using Forward     */
	int    do_null2;		/* TRUE to apply null2 ad hoc correction */
	struct threshold_s *thresh;   /* score/evalue threshold info     */

	/* Shared (mutex-protected) input resources:
	 */
	SQFILE *sqfp;                 /* ptr to open sequence file      */
	int nseq;			/* number of seqs searched so far */
	pthread_mutex_t input_lock;   /* mutex for locking input        */

	/* Shared (mutex-protected) output resources:
	 */
	struct tophit_s *ghit;        /* per-sequence top hits */
	struct tophit_s *dhit;        /* per-domain top hits */
	struct histogram_s *hist;     /* histogram of scores */
	pthread_mutex_t output_lock;  /* mutex for locking output */

	/* Thread pool information
	 */
	pthread_t *thread;            /* our pool of threads */
	int        num_threads;       /* number of threads   */
};
static struct workpool_s *workpool_start(struct plan7_s *hmm, SQFILE *sqfp,
		int do_xnu, int do_forward, int do_null2,
		struct threshold_s *thresh,
		struct tophit_s *ghit, struct tophit_s *dhit,
		struct histogram_s *hist, int num_threads);
static void  workpool_stop(struct workpool_s *wpool);
static void  workpool_free(struct workpool_s *wpool);
static void *worker_thread(void *ptr);
#endif /* HMMER_THREADS */


int main(int argc, char **argv)
{       char * seq;
        SQINFO  sqinfo;
	char    *hmmfile;             /* file to read HMM(s) from                */
	HMMFILE *hmmfp;               /* opened hmmfile for reading              */
	char    *seqfile;             /* file to read target sequence(s) from    */
        char    *index;
	SQFILE   *sqfp;               /* opened seqfile for reading              */
	int       format;             /* format of seqfile                       */
	int       i;
	struct plan7_s  *hmm;         /* HMM to search with                      */
	struct histogram_s *histogram;/* histogram of all scores                 */
	struct fancyali_s *ali;       /* displayed alignment info                */
	struct tophit_s   *ghit;      /* list of top hits for whole sequences    */
	struct tophit_s   *dhit;      /* list of top hits for domains            */

	float  globT;			/* T parameter: keep only hits > globT bits */
	double globE;			/* E parameter: keep hits < globE E-value   */
	int my_idx;			/* my slave index: 0..nslaves-1, master assigns */
	int L;			/* length of sequence */
	unsigned char *dsq;           /* digitized sequence 1..L */
	struct dpmatrix_s *mx;        /* growable DP matrix */
	struct p7trace_s *tr;         /* trace structure for a Viterbi alignment */

	int alphatype;		/* alphabet type */

	float   sc;	        /* score of an HMM search                  */
	double  pvalue;		/* pvalue of an HMM score                  */
	double  evalue;		/* evalue of an HMM score                  */
	double  motherp;		/* pvalue of a whole seq HMM score         */
	float   mothersc;		/* score of a whole seq parent of domain   */
	int     sqfrom, sqto;	/* coordinates in sequence                 */
	int     hmmfrom, hmmto;	/* coordinate in HMM                       */
	char   *name, *desc;        /* hit sequence name and description       */
	int     sqlen;		/* length of seq that was hit              */
	int     nseq;		/* number of sequences searched            */
	int     Z;			/* # of seqs for purposes of E-val calc    */
	int     domidx;		/* number of this domain                   */
	int     ndom;		/* total # of domains in this seq          */
	int     namewidth;		/* max width of sequence name              */
	int     descwidth;		/* max width of description */
	int     nreported;		/* # of hits reported in a list            */

	int    Alimit;		/* A parameter limiting output alignments   */
	struct threshold_s thresh;  /* contains all threshold (cutoff) info     */

	char *optname;              /* name of option found by Getopt()         */
	char *optarg;               /* argument found by Getopt()               */
	int   optind;               /* index in argv[]                          */
	int   do_null2;		/* TRUE to adjust scores with null model #2 */
	int   do_forward;		/* TRUE to use Forward() not Viterbi()      */
	int   do_xnu;		/* TRUE to filter sequences thru XNU        */
	int   do_pvm;		/* TRUE to run on Parallel Virtual Machine  */
	int   do_mpi;
	int   be_backwards;		/* TRUE to be backwards-compatible in output*/
	int   num_threads;		/* number of worker threads                 */
	int   threads_support;	/* TRUE if threads support compiled in      */
	int   pvm_support;		/* TRUE if PVM support compiled in          */
	int   rank;
	int   mpi_threads;
	int   mpi_support, real_num;
	MPI_Status temp;
        MPI_Status *useless = &temp;
	int   send_trace, offset;
	struct resultpkt newresult;
        FILE * indexfile;// = fopen("index.i", "w");
        MPI_Comm subcomm, mastercomm;
        MPI_Group wgroup, subgroup, empgroup, mastergroup;
        int ranks[][3] = {{0,0,1},{0,0,1}}; 
#ifdef HMMER_MPI
	int rc;

	rc = MPI_Init(&argc, &argv);

	if (rc != MPI_SUCCESS)
	{
		printf("MPI cannot be initialized. Terminating.\n");
		MPI_Abort(MPI_COMM_WORLD, rc);
	}

	MPI_Comm_size(MPI_COMM_WORLD, &mpi_threads);
	MPI_Comm_rank(MPI_COMM_WORLD, &rank);
        MPI_Comm_group(MPI_COMM_WORLD, &wgroup);

        ranks[0][0] = rank - rank%32;
        if(mpi_threads - ranks[0][0] > 31)
        ranks[0][1] = ranks[0][0] + 31;
        else ranks[0][1] = mpi_threads-1;

        real_num = ranks[0][1] - ranks[0][0] + 1;
        
        if(mpi_threads>1)
        {
            MPI_Group_range_incl(wgroup, 1, ranks, &subgroup);
            MPI_Comm_create(MPI_COMM_WORLD, subgroup, &subcomm);

            ranks[1][0] = 1;
            ranks[1][1] = mpi_threads;
            ranks[1][2] = 32;
            ranks[0][0] = 0;
            ranks[0][1] = 0;
            ranks[0][2] = 1;
            
            MPI_Group_range_incl(wgroup, 2, ranks, &mastergroup);
            MPI_Comm_create(MPI_COMM_WORLD, mastergroup, &mastercomm);
       } 
#else
	rank = 0;
#endif

	/***********************************************
	 * Parse command line
	 ***********************************************/

	format      = SQFILE_UNKNOWN;    /* default: autodetect seq file format  */
	do_forward  = FALSE;
	do_null2    = TRUE;
	do_xnu      = FALSE;
	do_pvm      = FALSE;
	do_mpi      = FALSE;
	Z           = 0;
	be_backwards= FALSE;

	pvm_support     = FALSE;
	threads_support = FALSE;
	mpi_support     = FALSE;
	num_threads     = 0;
#ifdef HMMER_THREADS
	num_threads     = ThreadNumber();
	threads_support = TRUE;
#endif
#ifdef HMMER_PVM
	pvm_support     = TRUE;
#endif
#ifdef HMMER_MPI
	mpi_support     = TRUE;
#endif
        //really all the ranks should do this instead of the master broadcasting this info

	if(rank==0 || rank%32==1)
	{
		Alimit         = INT_MAX;     /* no limit on alignment output       */
		thresh.globE   = 10.0;        /* use a reasonable Eval threshold;   */
		thresh.globT   = -FLT_MAX;    /*   but no bit threshold,            */
		thresh.domT    = -FLT_MAX;    /*   no domain bit threshold,         */
		thresh.domE    = FLT_MAX;     /*   and no domain Eval threshold.    */
		thresh.autocut = CUT_NONE;    /*   and no Pfam cutoffs used         */
		thresh.Z       = 0;           /* Z not preset; use actual # of seqs */
	}
	while (Getopt(argc, argv, OPTIONS, NOPTIONS, usage,
				&optind, &optname, &optarg))  {
		if      (strcmp(optname, "-A") == 0)        Alimit         = atoi(optarg);
		else if (strcmp(optname, "-E") == 0)        thresh.globE   = atof(optarg);
		else if (strcmp(optname, "-T") == 0)        thresh.globT   = atof(optarg);
		else if (strcmp(optname, "-Z") == 0)        thresh.Z       = atoi(optarg);
		else if (strcmp(optname, "--compat")  == 0) be_backwards   = TRUE;
		else if (strcmp(optname, "--cpu")     == 0) num_threads    = atoi(optarg);
		else if (strcmp(optname, "--cut_ga")  == 0) thresh.autocut = CUT_GA;
		else if (strcmp(optname, "--cut_nc")  == 0) thresh.autocut = CUT_NC;
		else if (strcmp(optname, "--cut_tc")  == 0) thresh.autocut = CUT_TC;
		else if (strcmp(optname, "--domE")    == 0) thresh.domE    = atof(optarg);
		else if (strcmp(optname, "--domT")    == 0) thresh.domT    = atof(optarg);
		else if (strcmp(optname, "--forward") == 0) do_forward     = TRUE;
		else if (strcmp(optname, "--null2")   == 0) do_null2       = FALSE;
		else if (strcmp(optname, "--pvm")     == 0) do_pvm         = TRUE;
		else if (strcmp(optname, "--xnu")     == 0) do_xnu         = TRUE;
		else if (strcmp(optname, "--mpi")     == 0) do_mpi         = TRUE;
		else if (strcmp(optname, "--informat") == 0 && rank == 0)
		{
			format = String2SeqfileFormat(optarg);
			if (format == SQFILE_UNKNOWN)
				Die("unrecognized sequence file format \"%s\"", optarg);
		}
		else if (strcmp(optname, "-h") == 0)
		{
			if (rank == 0)
			{
				HMMERBanner(stdout, banner);
				puts(usage);
				puts(experts);
			}
#ifdef HMMER_MPI
			MPI_Finalize();
#endif
			exit(0);
		}
	}
	if (argc - optind != 2)
	{

		Die("Incorrect number of arguments.\n%s\n", usage);

	}
	if(rank==0 || rank%32==1)
	{
		hmmfile = argv[optind++];
		seqfile = argv[optind++];
                
                if(rank==0)
                {
		    if (do_pvm && ! pvm_support)
		            Die("PVM support is not compiled into your HMMER software; --pvm doesn't work.");
	    	    if (num_threads && ! threads_support)
			    Die("POSIX threads support is not compiled into HMMER; --cpu doesn't have any effect");
                }
		/* Try to work around inability to autodetect from a pipe or .gz:
		 * assume FASTA format
		 */
		if (rank == 1 && format == SQFILE_UNKNOWN &&
				(Strparse("^.*\\.gz$", seqfile, 0) || strcmp(seqfile, "-") == 0))
			format = SQFILE_FASTA;

		/***********************************************
		 * Open sequence database (might be in BLASTDB or current directory)
		 ***********************************************/
                
                if(rank==0) 
	        {       if ((sqfp = SeqfileOpen(seqfile, format, "BLASTDB")) == NULL)
	                {
    
        	            Die("Failed to open sequence database file %s\n%s\n", seqfile, usage);
                        }
	        }
                if(rank ==0 )
                {   index = malloc ((sizeof(char))*strlen(seqfile)+3);
                    i=strlen(seqfile);
                    strcpy(index, seqfile);
                    index[i]='.';
                    index[i+1]='i';
                    index[i+2]='\0';
                    if((indexfile = fopen(index, "r"))==NULL)
                    {   printf("Creating index file: %s, this will only be done once:\n", index);
                        printf("|-----------------------------------------------|\n");
                        indexfile = fopen(index, "w");
                        printf("Done open\n");
                        fprintf(indexfile, "            \n");
                        i=0;
                        fprintf(indexfile, "%11d\n", ftell(sqfp->f));
	                while (ReadSeq(sqfp, sqfp->format, &seq, &sqinfo))
                        {   
                            fprintf(indexfile, "%11d\n", ftell(sqfp->f));
		            i++;
                            FreeSequence(seq, &sqinfo);
                        }
                        rewind(indexfile);
                        fprintf(indexfile, "%11d", i);
                        fclose(indexfile);
                    }
                    indexfile = fopen(index, "r");
                }
                
	/***********************************************
		 * Open HMM file (might be in HMMERDB or current directory).
		 * Read a single HMM from it. (Config HMM, if necessary).
		 * Alphabet globals are set by reading the HMM.
		 ***********************************************/
                if(rank==0 || rank%32==1) 
                {   
                    if ((hmmfp = HMMFileOpen(hmmfile, "HMMERDB")) == NULL)
		    {
			Die("Failed to open HMM file %s\n%s", hmmfile, usage);

		    }
                    if (!HMMFileRead(hmmfp, &hmm))
			Die("Failed to read any HMMs from %s\n", hmmfile);
		    if (hmm == NULL)
			Die("HMM file %s corrupt or in incorrect format? Parse failed", hmmfile);
		    P7Logoddsify(hmm, !do_forward);
                
		    if (do_xnu && Alphabet_type == hmmNUCLEIC)
			Die("The HMM is a DNA model, and you can't use the --xnu filter on DNA data");
                }
	}
	/*****************************************************************
	 * Set up optional Pfam score thresholds.
	 * Can do this before starting any searches, since we'll only use 1 HMM.
	 *****************************************************************/

	if (! SetAutocuts(&thresh, hmm))
		Die("HMM %s did not contain the GA, TC, or NC cutoffs you needed",
				hmm->name);

	/***********************************************
	 * Show the banner
	 ***********************************************/
	if(rank==0)
	{	HMMERBanner(stdout, banner);
		printf(   "HMM file:                   %s [%s]\n", hmmfile, hmm->name);
		printf(   "Sequence database:          %s\n", seqfile);
		if (do_pvm)
			printf( "PVM:                        ACTIVE\n");
		printf(   "per-sequence score cutoff:  ");
		if(do_mpi)
			printf( "MPI:                        ACTIVE\n");
		if (thresh.globT == -FLT_MAX) printf("[none]\n");
		else  {
			printf(">= %.1f", thresh.globT);
			if      (thresh.autocut == CUT_GA) printf(" [GA1]\n");
			else if (thresh.autocut == CUT_NC) printf(" [NC1]\n");
			else if (thresh.autocut == CUT_TC) printf(" [TC1]\n");
			else                               printf("\n");
		}
		printf(   "per-domain score cutoff:    ");
		if (thresh.domT == -FLT_MAX) printf("[none]\n");
		else  {
			printf(">= %.1f", thresh.domT);
			if      (thresh.autocut == CUT_GA) printf(" [GA2]\n");
			else if (thresh.autocut == CUT_NC) printf(" [NC2]\n");
			else if (thresh.autocut == CUT_TC) printf(" [TC2]\n");
			else                               printf("\n");
		}
		printf(   "per-sequence Eval cutoff:   ");
		if (thresh.globE == FLT_MAX) printf("[none]\n");
		else                  printf("<= %-10.2g\n", thresh.globE);

		printf(   "per-domain Eval cutoff:     ");
		if (thresh.domE == FLT_MAX) printf("[none]\n");
		else                 printf("<= %10.2g\n", thresh.domE);
		printf("- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -\n");

		if (pvm_support && do_pvm)
			main_loop_pvm(hmm, sqfp, &thresh, do_forward, do_null2, do_xnu,
					histogram, ghit, dhit, &nseq);
		else if (threads_support && num_threads)
			main_loop_threaded(hmm, sqfp, &thresh, do_forward, do_null2, do_xnu, num_threads,
					histogram, ghit, dhit, &nseq);
		else if (mpi_support && do_mpi)
		{ 	
                        histogram = AllocHistogram(-200, 200, 100);
                        dhit = AllocTophits(35000);
                        ghit = AllocTophits(35000);
			main_loop_master_mpi(hmm, indexfile, &thresh, do_forward, do_null2, do_xnu, 
					histogram, ghit, dhit, &nseq,
                                        mpi_threads, mastercomm);
                        //printf("done again\n");
		}
		else
		{	
		    histogram = AllocHistogram(-200, 200, 100);  /* keeps full histogram */
		    ghit      = AllocTophits(200);         /* per-seq hits: 200=lumpsize */
	    	    dhit      = AllocTophits(200);         /* domain hits:  200=lumpsize */
                    main_loop_serial(hmm, sqfp, &thresh, do_forward, do_null2, do_xnu,
					histogram, ghit, dhit, &nseq);
                }
         }
         else if(rank%32 == 1)
	 {	/***********************************************
		 * Search HMM against each sequence
		 ***********************************************/
		/* set up structures for storing output  */
		ghit      = AllocTophits(200);         /* per-seq hits: 200=lumpsize */
		dhit      = AllocTophits(200);         /* domain hits:  200=lumpsize */

		if (mpi_support && do_mpi)
		{ 	
                    #ifdef HMMER_MPI
                    main_loop_mpi(hmm, seqfile, format, &thresh, do_forward, do_null2, do_xnu, 
                        histogram, ghit, dhit, &nseq, real_num,
                        subcomm, rank, mastercomm);
                    #else
                        printf("No MPI Support, aborting...\n");
                        MPI_Abort();
                    #endif
		}
         }
         if(rank==0)
	 {	
                /***********************************************
		 * Process hit lists, produce text output
		 ***********************************************/

		/* Set the theoretical EVD curve in our histogram using
		 * calibration in the HMM, if available.
		 */
		if (hmm->flags & PLAN7_STATS)
			ExtremeValueSetHistogram(histogram, hmm->mu, hmm->lambda,
					histogram->lowscore, histogram->highscore, 0);
		if (!thresh.Z) thresh.Z = nseq;		/* set Z for good now that we're done. */

		/* Format and report our output
		 */
		/* 1. Report overall sequence hits (sorted on E-value) */
		if (be_backwards)
		{
			printf("\nQuery HMM: %s|%s|%s\n",
					hmm->name,
					hmm->flags & PLAN7_ACC  ? hmm->acc  : "",
					hmm->flags & PLAN7_DESC ? hmm->desc : "");
		}
		else
		{
			printf("\nQuery HMM:   %s\n", hmm->name);
			printf("Accession:   %s\n", hmm->flags & PLAN7_ACC  ? hmm->acc  : "[none]");
			printf("Description: %s\n", hmm->flags & PLAN7_DESC ? hmm->desc : "[none]");
		}

		if (hmm->flags & PLAN7_STATS)
			printf("  [HMM has been calibrated; E-values are empirical estimates]\n");
		else
			printf("  [No calibration for HMM; E-values are upper bounds]\n");
                
		FullSortTophits(ghit);
		namewidth = MAX(8, TophitsMaxName(ghit)); /* cannot truncate name. */
		descwidth = MAX(52-namewidth, 11);/* may truncate desc, but need strlen("Description") */
		printf("\nScores for complete sequences (score includes all domains):\n");
		printf("%-*s %-*s %7s %10s %3s\n", namewidth, "Sequence", descwidth, "Description", "Score", "E-value", " N ");
		printf("%-*s %-*s %7s %10s %3s\n", namewidth, "--------", descwidth, "-----------", "-----", "-------", "---");

		for (i = 0, nreported = 0; i < ghit->num; i++)
		{
                    char *safedesc;
                    GetRankedHit(ghit, i,
                                    &pvalue, &sc, NULL, NULL,
                                    &name, NULL, &desc,
                                    NULL, NULL, NULL,               /* sequence positions */
                                    NULL, NULL, NULL,               /* HMM positions      */
                                    NULL, &ndom,	                   /* domain info        */
                                    NULL);	                   /* alignment info     */
                    
            
                    evalue = pvalue * (double) thresh.Z;

                    /* safedesc is a workaround for an apparent Linux printf()
                     * bug with the *.*s format. dbmalloc crashes with a memchr() ptr out of bounds
                     * flaw if the malloc'ed space for desc is short. The workaround
                     * is to make sure the ptr for *.* has a big malloc space.
                     */
                    if (desc != NULL && strlen(desc) < 80)
                    {
                        safedesc = newMallocOrDie(sizeof(char) * 80);
                        strcpy(safedesc, desc);
                    }
                    else safedesc = Strdup(desc);

                    if (evalue <= thresh.globE && sc >= thresh.globT) {
                        printf("%-*s %-*.*s %7.1f %10.2g %3d\n",
                            namewidth, name,
                            descwidth, descwidth, safedesc != NULL ? safedesc : "",
                            sc, evalue, ndom);
                        nreported++;
                    }
                    free(safedesc);
		}
		if (nreported == 0) printf("\t[no hits above thresholds]\n");


		/* 2. Report domain hits (also sorted on E-value) */
		FullSortTophits(dhit);
		namewidth = MAX(8, TophitsMaxName(dhit));

		printf("\nParsed for domains:\n");
		printf("%-*s %7s %5s %5s    %5s %5s    %7s %8s\n",
				namewidth, "Sequence", "Domain ", "seq-f", "seq-t", "hmm-f", "hmm-t", "score", "E-value");
		printf("%-*s %7s %5s %5s    %5s %5s    %7s %8s\n",
				namewidth, "--------", "-------", "-----", "-----", "-----", "-----", "-----", "-------");

		for (i = 0, nreported = 0; i < dhit->num; i++)
		{
                    GetRankedHit(dhit, i,
                        &pvalue, &sc, &motherp, &mothersc,
                        &name, NULL, NULL,
                        &sqfrom, &sqto, &sqlen,            /* seq position info  */
                        &hmmfrom, &hmmto, NULL,            /* HMM position info  */
                        &domidx, &ndom,                    /* domain info        */
                        NULL);	                      /* alignment info     */
                    evalue = pvalue * (double) thresh.Z;

                    if (motherp * (double) thresh.Z > thresh.globE || mothersc < thresh.globT)
                        continue;
                    else if (evalue <= thresh.domE && sc >= thresh.domT) {
                        printf("%-*s %3d/%-3d %5d %5d %c%c %5d %5d %c%c %7.1f %8.2g\n",
                            namewidth, name,
                            domidx, ndom,
                            sqfrom, sqto,
                            sqfrom == 1 ? '[' : '.', sqto == sqlen ? ']' : '.',
                            hmmfrom, hmmto,
                            hmmfrom == 1 ? '[':'.', hmmto == hmm->M ? ']' : '.',
                            sc, evalue);
                        nreported++;
                    }
		}
		if (nreported == 0) printf("\t[no hits above thresholds]\n");


		/* 3. Alignment output, also by domain.
		 *    dhits is already sorted and namewidth is set, from above code.
		 *    Number of displayed alignments is limited by Alimit parameter;
		 *    also by domE (evalue threshold), domT (score theshold).
		 */
		if (Alimit != 0)
		{
                    printf("\nAlignments of top-scoring domains:\n");
                    for (i = 0, nreported = 0; i < dhit->num; i++)
                    {
                        if (nreported == Alimit) break; /* limit to Alimit output alignments */
                        GetRankedHit(dhit, i,
                            &pvalue, &sc, &motherp, &mothersc,
                            &name, NULL, NULL,
                            &sqfrom, &sqto, &sqlen,            /* seq position info  */
                            &hmmfrom, &hmmto, NULL,            /* HMM position info  */
                            &domidx, &ndom,                    /* domain info        */
                            &ali);	                      /* alignment info     */
                        evalue = pvalue * (double) thresh.Z;

                        if (motherp * (double) thresh.Z > thresh.globE || mothersc < thresh.globT)
                                continue;
                        else if (evalue <= thresh.domE && sc >= thresh.domT)
                        {
                            printf("%s: domain %d of %d, from %d to %d: score %.1f, E = %.2g\n",
                                            name, domidx, ndom, sqfrom, sqto, sc, evalue);
                            PrintFancyAli(stdout, ali);
                            nreported++;
                        }
                    }
                    if (nreported == 0)      printf("\t[no hits above thresholds]\n");
                    if (nreported == Alimit) printf("\t[output cut off at A = %d top alignments]\n", Alimit);
		}

		/* 4. Histogram output */
		printf("\nHistogram of all scores:\n");
		PrintASCIIHistogram(stdout, histogram);

		/* 5. Tophits summaries, while developing...
		 */
		printf("\nTotal sequences searched: %d\n", nseq);
		printf("\nWhole sequence top hits:\n");
		TophitsReport(ghit, thresh.globE, nseq);
		printf("\nDomain top hits:\n");
		TophitsReport(dhit, thresh.domE, nseq);

		/***********************************************
		 * Clean-up and exit.
		 ***********************************************/

		FreeHistogram(histogram);
		HMMFileClose(hmmfp);
		FreeTophits(ghit);
		FreeTophits(dhit);
	}
#ifdef HMMER_MPI
	else if(rank%32 != 1)
            main_loop_mpi_slave(rank%32, rank, num_threads, do_mpi, do_null2, do_forward, do_xnu, be_backwards, subcomm);
        MPI_Finalize();
#endif
	return 0;
}


/* Function: main_loop_serial()
 * Date:     SRE, Wed Sep 23 10:20:49 1998 [St. Louis]
 *
 * Purpose:  Search an HMM against a sequence database.
 *           main loop for the serial (non-PVM, non-threads)
 *           version.
 *
 *           In:   HMM and open sqfile, plus options
 *           Out:  histogram, global hits list, domain hits list, nseq.
 *
 * Args:     hmm        - the HMM to search with.
 *           sqfp       - open SQFILE for sequence database
 *           thresh     - score/evalue threshold info
 *           do_forward - TRUE to score using Forward()
 *           do_null2   - TRUE to use ad hoc null2 score correction
 *           do_xnu     - TRUE to apply XNU mask
 *           histogram  - RETURN: score histogram
 *           ghit       - RETURN: ranked global scores
 *           dhit       - RETURN: ranked domain scores
 *           ret_nseq   - RETURN: actual number of seqs searched
 *
 * Returns:  (void)
 */
static void
main_loop_serial(struct plan7_s *hmm, SQFILE *sqfp, struct threshold_s *thresh, int do_forward,
		int do_null2, int do_xnu,
		struct histogram_s *histogram,
		struct tophit_s *ghit, struct tophit_s *dhit, int *ret_nseq)
{
	struct dpmatrix_s *mx;        /* DP matrix, growable                     */
	struct p7trace_s *tr;         /* traceback                               */
	char   *seq;                  /* target sequence                         */
	unsigned char   *dsq;	        /* digitized target sequence               */
	SQINFO sqinfo;		/* optional info for seq                   */
	float  sc;	        	/* score of an HMM search                  */
	double pvalue;		/* pvalue of an HMM score                  */
	double evalue;		/* evalue of an HMM score                  */
	int    nseq, i;			/* number of sequences searched            */

	/* Create a DP matrix; initially only two rows big, but growable;
	 * we overalloc by 25 rows (L dimension) when we grow; not growable
	 * in model dimension, since we know the hmm size
	 */
	mx = CreatePlan7Matrix(1, hmm->M, 25, 0);
        printf("Out\n");
	nseq = 0;
	while (ReadSeq(sqfp, sqfp->format, &seq, &sqinfo))
	{       printf("Read seq!\n");
		/* Silently skip length 0 seqs.
		 * What, you think this doesn't occur? Welcome to genomics,
		 * young grasshopper.
		 */
		if (sqinfo.len == 0) continue;
		nseq++;
                printf("About to digitize\n");
		dsq = DigitizeSequence(seq, sqinfo.len);

		if (do_xnu && Alphabet_type == hmmAMINO) XNU(dsq, sqinfo.len);

                printf("About to Viterbi\n");
		/* 1. Recover a trace by Viterbi.
		 *    In extreme cases, the alignment may be literally impossible;
		 *    in which case, the score comes out ridiculously small (but not
		 *    necessarily <= -INFTY, because we're not terribly careful
		 *    about underflow issues), and tr will be returned as NULL.
		 */
		if (P7ViterbiSpaceOK(sqinfo.len, hmm->M, mx))
                    sc = P7Viterbi(dsq, sqinfo.len, hmm, mx, &tr);
                else
                    sc = P7SmallViterbi(dsq, sqinfo.len, hmm, mx, &tr);

		/* 2. If we're using Forward scores, calculate the
		 *    whole sequence score; this overrides anything
		 *    PostprocessSignificantHit() is going to do to the per-seq score.
		 */
		if (do_forward) 
                {
                    sc  = P7Forward(dsq, sqinfo.len, hmm, NULL);
                    if (do_null2)   sc -= TraceScoreCorrection(hmm, tr, dsq);
		}

#if DEBUGLEVEL >= 2
		P7PrintTrace(stdout, tr, hmm, dsq);
#endif

		/* 2. Store score/pvalue for global alignment; will sort on score,
		 *    which in hmmsearch is monotonic with E-value.
		 *    Keep all domains in a significant sequence hit.
		 *    We can only make a lower bound estimate of E-value since
		 *    we don't know the final value of nseq yet, so the list
		 *    of hits we keep in memory is >= the list we actually
		 *    output.
		 */
		pvalue = PValue(hmm, sc);
		evalue = thresh->Z ? (double) thresh->Z * pvalue : (double) nseq * pvalue;
                if (sc >= thresh->globT && evalue <= thresh->globE)
		{
			sc = PostprocessSignificantHit(ghit, dhit,
					tr, hmm, dsq, sqinfo.len,
					sqinfo.name,
					sqinfo.flags & SQINFO_ACC  ? sqinfo.acc  : NULL,
					sqinfo.flags & SQINFO_DESC ? sqinfo.desc : NULL,
					do_forward, sc,
					do_null2,
					thresh,
					FALSE); /* FALSE-> not hmmpfam mode, hmmsearch mode */
		}
		SQD_DPRINTF2(("AddToHistogram: %s\t%f\n", sqinfo.name, sc));
		AddToHistogram(histogram, sc);
		FreeSequence(seq, &sqinfo);
		P7FreeTrace(tr);
		free(dsq);
	}

	FreePlan7Matrix(mx);
	*ret_nseq = nseq;
	return;
}



#ifdef HMMER_PVM
/*****************************************************************
 * PVM specific functions
 ****************************************************************/

/* Function: main_loop_pvm()
 * Date:     SRE, Wed Sep 23 10:36:44 1998 [St. Louis]
 *
 * Purpose:  Search an HMM against a sequence database.
 *           main loop for the PVM version.
 *
 *           In:   HMM and open sqfile, plus options
 *           Out:  histogram, global hits list, domain hits list, nseq.
 *
 * Args:     hmm        - the HMM to search with. scoring form.
 *           sqfp       - open SQFILE for sequence database
 *           thresh     - score/evalue threshold information
 *           do_forward - TRUE to score using Forward()
 *           do_null2   - TRUE to use ad hoc null2 score correction
 *           do_xnu     - TRUE to apply XNU mask
 *           histogram  - RETURN: score histogram
 *           ghit       - RETURN: ranked global scores
 *           dhit       - RETURN: ranked domain scores
 *           ret_nseq   - RETURN: actual number of seqs searched
 *
 * Returns:  (void)
 */
static void
main_loop_pvm(struct plan7_s *hmm, SQFILE *sqfp, struct threshold_s *thresh, int do_forward,
		int do_null2, int do_xnu, struct histogram_s *histogram,
		struct tophit_s *ghit, struct tophit_s *dhit, int *ret_nseq)
{
	char *seq;                    /* target sequence */
	unsigned char *dsq;           /* digitized target seq */
	SQINFO sqinfo;                /* optional info about target seq */
	int   master_tid;		/* master's (my) PVM TID */
	int  *slave_tid;              /* array of slave TID's  */
	int   nslaves;		/* number of slaves      */
	int   code;			/* status code rec'd from a slave */
	int   nseq;			/* number of sequences searched */
	int   sent_trace;		/* TRUE if slave gave us a trace */
	unsigned char **dsqlist;      /* remember what seqs slaves are doing */
	char **namelist;              /* remember what seq names slaves are doing */
	char **acclist ;              /* remember what seq accessions slaves are doing */
	char **desclist;              /* remember what seq desc's slaves are doing */
	int   *lenlist;               /* remember lengths of seqs slaves are doing */
	int    slaveidx;		/* counter for slaves */
	float  sc;			/* score of an alignment */
	double pvalue;		/* P-value of a score of an alignment */
	struct p7trace_s *tr;         /* Viterbi traceback of an alignment */
	int    i;			/* generic counter */

	/* Initialize PVM.
	 */
	SQD_DPRINTF1(("Requesting master TID...\n"));
	master_tid = pvm_mytid();
#if DEBUGLEVEL >= 1
	pvm_catchout(stderr);		/* catch output for debugging */
#endif
	SQD_DPRINTF1(("Spawning slaves...\n"));
	PVMSpawnSlaves("hmmsearch-pvm", &slave_tid, &nslaves);
	SQD_DPRINTF1(("Spawned a total of %d slaves...\n", nslaves));

	/* Initialize the slaves by broadcast.
	 */
	SQD_DPRINTF1(("Broadcasting to %d slaves...\n", nslaves));
	pvm_initsend(PvmDataDefault);
	pvm_pkfloat(&(thresh->globT), 1, 1);
	pvm_pkdouble(&(thresh->globE), 1, 1);
	pvm_pkint(&(thresh->Z),          1, 1);
	pvm_pkint(&do_forward, 1, 1);
	pvm_pkint(&do_null2,   1, 1);
	pvm_pkint(&Alphabet_type, 1, 1);
	PVMPackHMM(hmm);
	pvm_mcast(slave_tid, nslaves, HMMPVM_INIT);
	SQD_DPRINTF1(("Slaves should be ready...\n"));

	/* Confirm slaves' OK status.
	 */
	PVMConfirmSlaves(slave_tid, nslaves);
	SQD_DPRINTF1(("Slaves confirm that they're ok...\n"));

	/* Alloc arrays for remembering what seq each
	 * slave was working on.
	 */
	namelist = newMallocOrDie(sizeof(char *) * nslaves);
	acclist  = newMallocOrDie(sizeof(char *) * nslaves);
	desclist = newMallocOrDie(sizeof(char *) * nslaves);
	dsqlist  = newMallocOrDie(sizeof(unsigned char *) * nslaves);
	lenlist  = newMallocOrDie(sizeof(int) * nslaves);

	/* Load the slaves.
	 * Give them all a sequence number and a digitized sequence
	 * to work on.
	 * A side effect of the seq number is that we assign each slave
	 * a number from 0..nslaves-1.
	 */
	for (nseq = 0; nseq < nslaves; nseq++)
	{
		if (! ReadSeq(sqfp, sqfp->format, &seq, &sqinfo)) break;
		if (sqinfo.len == 0) { nseq--; continue; }

		dsq = DigitizeSequence(seq, sqinfo.len);
		if (do_xnu && Alphabet_type == hmmAMINO) XNU(dsq, sqinfo.len);

		pvm_initsend(PvmDataDefault);
		pvm_pkint(&nseq, 1, 1);
		pvm_pkint(&(sqinfo.len), 1, 1);
		pvm_pkbyte((char *) dsq, sqinfo.len+2, 1); /* is this char * casting safe?  */
		pvm_send(slave_tid[nseq], HMMPVM_WORK);
		SQD_DPRINTF1(("sent a dsq : %d bytes\n", sqinfo.len+2));

		namelist[nseq] = Strdup(sqinfo.name);
		acclist[nseq]  = (sqinfo.flags & SQINFO_ACC)  ? Strdup(sqinfo.acc)  : NULL;
		desclist[nseq] = (sqinfo.flags & SQINFO_DESC) ? Strdup(sqinfo.desc) : NULL;
		lenlist[nseq]  = sqinfo.len;
		dsqlist[nseq]  = dsq;

		FreeSequence(seq, &sqinfo);
	}
	SQD_DPRINTF1(("%d slaves are loaded\n", nseq));

	/* main receive/send loop
	 */
	while (ReadSeq(sqfp, sqfp->format, &seq, &sqinfo))
	{
		if (sqinfo.len == 0) { continue; }
		nseq++;
		/* check slaves before blocking */
		PVMCheckSlaves(slave_tid, nslaves);

		/* receive output */
		SQD_DPRINTF1(("Waiting for a slave to give me output...\n"));
		pvm_recv(-1, HMMPVM_RESULTS);
		pvm_upkint(&slaveidx, 1, 1);     /* # of slave who's sending us stuff */
		pvm_upkfloat(&sc, 1, 1);         /* score   */
		pvm_upkdouble(&pvalue, 1, 1);    /* P-value */
		pvm_upkint(&sent_trace, 1, 1);   /* TRUE if trace is coming */
		tr = (sent_trace) ? PVMUnpackTrace() : NULL;
		SQD_DPRINTF1(("Slave %d finished %s for me...\n", slaveidx, namelist[slaveidx]));

		/* send new work */
		dsq = DigitizeSequence(seq, sqinfo.len);
		if (do_xnu) XNU(dsq, sqinfo.len);

		pvm_initsend(PvmDataDefault);
		pvm_pkint(&nseq, 1, 1);
		pvm_pkint(&(sqinfo.len), 1, 1);
		pvm_pkbyte((char *) dsq, sqinfo.len+2, 1); /* safety of char * cast unknown. */
		pvm_send(slave_tid[slaveidx], HMMPVM_WORK);

		/* process output */
		if (sent_trace)
		{
			sc = PostprocessSignificantHit(ghit, dhit,
					tr, hmm, dsqlist[slaveidx], lenlist[slaveidx],
					namelist[slaveidx], acclist[slaveidx], desclist[slaveidx],
					do_forward, sc,
					do_null2,
					thresh,
					FALSE); /* FALSE-> not hmmpfam mode, hmmsearch mode */
			P7FreeTrace(tr);
		}
		AddToHistogram(histogram, sc);

		/* record seq info for seq we just sent */
		free(namelist[slaveidx]);
		if (acclist[slaveidx]  != NULL) free(acclist[slaveidx]);
		if (desclist[slaveidx] != NULL) free(desclist[slaveidx]);
		free(dsqlist[slaveidx]);

		dsqlist[slaveidx]  = dsq;
		namelist[slaveidx] = Strdup(sqinfo.name);
		acclist[slaveidx]  = (sqinfo.flags & SQINFO_ACC)  ? Strdup(sqinfo.acc)  : NULL;
		desclist[slaveidx] = (sqinfo.flags & SQINFO_DESC) ? Strdup(sqinfo.desc) : NULL;
		lenlist[slaveidx]  = sqinfo.len;

		FreeSequence(seq, &sqinfo);
	}
	SQD_DPRINTF1(("End of receive/send loop\n"));

	/* Collect the output. All n slaves are still working.
	 */
	for (i = 0; i < nslaves && i < nseq; i++)
	{
		/* don't check slaves (they're exiting normally);
		   window of vulnerability here to slave crashes */
		/* receive output */
		pvm_recv(-1, HMMPVM_RESULTS);
		pvm_upkint(&slaveidx, 1, 1);     /* # of slave who's sending us stuff */
		pvm_upkfloat(&sc, 1, 1);         /* score   */
		pvm_upkdouble(&pvalue, 1, 1);    /* P-value */
		pvm_upkint(&sent_trace, 1, 1);   /* TRUE if trace is coming */
		tr = (sent_trace) ? PVMUnpackTrace() : NULL;
		SQD_DPRINTF1(("Slave %d finished %s for me...\n", slaveidx, namelist[slaveidx]));

		/* process output */
		if (sent_trace)
		{
			sc = PostprocessSignificantHit(ghit, dhit,
					tr, hmm, dsqlist[slaveidx], lenlist[slaveidx],
					namelist[slaveidx], acclist[slaveidx], desclist[slaveidx],
					do_forward, sc,
					do_null2,
					thresh,
					FALSE); /* FALSE-> not hmmpfam mode, hmmsearch mode */
			P7FreeTrace(tr);
		}
		SQD_DPRINTF2(("AddToHistogram: %s\t%f\n", namelist[slaveidx], sc));
		AddToHistogram(histogram, sc);

		/* free seq info */
		free(namelist[slaveidx]);
		if (acclist[slaveidx]  != NULL) free(acclist[slaveidx]);
		if (desclist[slaveidx] != NULL) free(desclist[slaveidx]);
		free(dsqlist[slaveidx]);

		/* send cleanup/shutdown flag to slave */
		pvm_initsend(PvmDataDefault);
		code = -1;
		pvm_pkint(&code, 1, 1);
		pvm_send(slave_tid[slaveidx], HMMPVM_WORK);
	}


	/* Cleanup; quit the VM; and return
	 */
	free(slave_tid);
	free(dsqlist);
	free(namelist);
	free(acclist);
	free(desclist);
	free(lenlist);
	pvm_exit();
	*ret_nseq = nseq;
	return;
}
#else /* HMMER_PVM off, no PVM support, dummy function: */
static void
main_loop_pvm(struct plan7_s *hmm, SQFILE *sqfp, struct threshold_s *thresh, int do_forward,
		int do_null2, int do_xnu, struct histogram_s *histogram,
		struct tophit_s *ghit, struct tophit_s *dhit, int *ret_nseq)
{
	Die("No PVM support");
}
#endif /*HMMER_PVM*/

#ifdef HMMER_THREADS
/*****************************************************************
 * POSIX threads implementation.
 *
 * API:
 *      workpool_start()   (makes a workpool_s structure. Starts calculations.)
 *      workpool_stop()    (waits for threads to finish.)
 *      workpool_free()    (destroys the structure)
 *
 * Threads:
 *      worker_thread()    (the actual parallelized worker thread).
 *****************************************************************/
/* Function: main_loop_threaded
 * Date:     SRE, Wed Feb 27 11:38:11 2002 [St. Louis]
 *
 * Purpose:  Search an HMM against a sequence database.
 *           main loop for the threaded version.
 *
 *           In:   HMM and open sqfile, plus options
 *           Out:  histogram, global hits list, domain hits list, nseq.
 *
 * Args:     hmm        - the HMM to search with.
 *           sqfp       - open SQFILE for sequence database
 *           thresh     - score/evalue threshold info
 *           do_forward - TRUE to score using Forward()
 *           do_null2   - TRUE to use ad hoc null2 score correction
 *           do_xnu     - TRUE to apply XNU mask
 *           num_threads- number of worker threads to start, >=1
 *           histogram  - RETURN: score histogram
 *           ghit       - RETURN: ranked global scores
 *           dhit       - RETURN: ranked domain scores
 *           ret_nseq   - RETURN: actual number of seqs searched
 *
 * Returns:  (void)
 */
static void
main_loop_threaded(struct plan7_s *hmm, SQFILE *sqfp, struct threshold_s *thresh, int do_forward,
		int do_null2, int do_xnu, int num_threads,
		struct histogram_s *histogram,
		struct tophit_s *ghit, struct tophit_s *dhit, int *ret_nseq)
{
	struct workpool_s *wpool;	/* pool of worker threads                  */
	int    nseq;			/* number of sequences searched            */

	wpool = workpool_start(hmm, sqfp, do_xnu, do_forward, do_null2, thresh,
			ghit, dhit, histogram, num_threads);
	workpool_stop(wpool);
	nseq = wpool->nseq;
	workpool_free(wpool);

	*ret_nseq = nseq;
	return;
}
/* Function: workpool_start()
 * Date:     SRE, Mon Oct  5 16:44:53 1998
 *
 * Purpose:  Initialize a workpool_s structure, and return it.
 *
 * Args:     sqfp       - open sequence file, at start
 *           do_xnu     - TRUE to apply XNU filter
 *           do_forward - TRUE to score using Forward
 *           do_null2   - TRUE to apply null2 ad hoc correction
 *           thresh     - score/evalue threshold info
 *           ghit       - per-seq hit list
 *           dhit       - per-domain hit list
 *           hist       - histogram (alloced but empty)
 *           num_threads- number of worker threads to run.
 *
 * Returns:  ptr to struct workpool_s.
 *           Caller must wait for threads to finish with workpool_stop(),
 *           then free the structure with workpool_free().
 */
static struct workpool_s *
workpool_start(struct plan7_s *hmm, SQFILE *sqfp, int do_xnu,
		int do_forward, int do_null2, struct threshold_s *thresh,
		struct tophit_s *ghit, struct tophit_s *dhit,
		struct histogram_s *hist, int num_threads)
{
	struct workpool_s *wpool;
	pthread_attr_t    attr;
	int i;
	int rtn;

	wpool             = newMallocOrDie(sizeof(struct workpool_s));
	wpool->thread     = newMallocOrDie(num_threads * sizeof(pthread_t));
	wpool->hmm        = hmm;

	wpool->do_xnu     = do_xnu;
	wpool->do_forward = do_forward;
	wpool->do_null2   = do_null2;
	wpool->thresh     = thresh;

	wpool->sqfp       = sqfp;
	wpool->nseq       = 0;
	if ((rtn = pthread_mutex_init(&(wpool->input_lock), NULL)) != 0)
		Die("pthread_mutex_init FAILED; %s\n", strerror(rtn));

	wpool->ghit       = ghit;
	wpool->dhit       = dhit;
	wpool->hist       = hist;
	if ((rtn = pthread_mutex_init(&(wpool->output_lock), NULL)) != 0)
		Die("pthread_mutex_init FAILED; %s\n", strerror(rtn));

	wpool->num_threads= num_threads;

	/* Create slave threads. See comments in hmmcalibrate.c at this
	 * step, regarding concurrency, system scope, and portability
	 * amongst various UNIX implementations of pthreads.
	 */
	pthread_attr_init(&attr);
#ifndef __sgi
#ifdef HAVE_PTHREAD_ATTR_SETSCOPE
	pthread_attr_setscope(&attr, PTHREAD_SCOPE_SYSTEM);
#endif
#endif
#ifdef HAVE_PTHREAD_SETCONCURRENCY
	pthread_setconcurrency(num_threads+1);
#endif
	/* pthread_attr_setscope(&attr, PTHREAD_SCOPE_SYSTEM); */
	for (i = 0; i < num_threads; i++)
		if ((rtn = pthread_create(&(wpool->thread[i]), &attr,
						worker_thread , (void *) wpool)) != 0)
			Die("Failed to create thread %d; return code %d\n", i, rtn);

	pthread_attr_destroy(&attr);
	return wpool;
}
/* Function: workpool_stop()
 * Date:     SRE, Thu Jul 16 11:20:16 1998 [St. Louis]
 *
 * Purpose:  Waits for threads in a workpool to finish.
 *
 * Args:     wpool -- ptr to the workpool structure
 *
 * Returns:  (void)
 */
	static void
workpool_stop(struct workpool_s *wpool)
{
	int i;
	/* wait for threads to stop */
	for (i = 0; i < wpool->num_threads; i++)
		if (pthread_join(wpool->thread[i],NULL) != 0)
			Die("pthread_join failed");
	return;
}

/* Function: workpool_free()
 * Date:     SRE, Thu Jul 16 11:26:27 1998 [St. Louis]
 *
 * Purpose:  Free a workpool_s structure, after the threads
 *           have finished.
 *
 * Args:     wpool -- ptr to the workpool.
 *
 * Returns:  (void)
 */
	static void
workpool_free(struct workpool_s *wpool)
{
	free(wpool->thread);
	free(wpool);
	return;
}


/* Function: worker_thread()
 * Date:     SRE, Mon Sep 28 10:48:29 1998 [St. Louis]
 *
 * Purpose:  The procedure executed by the worker threads.
 *
 * Args:     ptr  - (void *) that is recast to a pointer to
 *                  the workpool.
 *
 * Returns:  (void *)
 */
	void *
worker_thread(void *ptr)
{
	struct workpool_s *wpool;     /* our working threads structure   */
	char  *seq;                   /* target sequence                 */
	SQINFO sqinfo;		/* information assoc w/ seq        */
	unsigned char     *dsq;       /* digitized sequence              */
	struct dpmatrix_s *mx;        /* growable DP matrix              */
	struct p7trace_s  *tr;        /* traceback from an alignment     */
	float  sc;			/* score of an alignment           */
	int    rtn;			/* a return code from pthreads lib */
	double pvalue;		/* P-value of score                */
	double evalue;		/* E-value of score                */

	wpool = (struct workpool_s *) ptr;

	/* Init with a small DP matrix; we'll grow in the sequence dimension
	 * overalloc'ing by 25 rows (residues).
	 */
	mx = CreatePlan7Matrix(1, wpool->hmm->M, 25, 0);
	for (;;) {

		/* 1. acquire lock on sequence input, and get
		 *    the next seq to work on.
		 */
		/* acquire a lock */
		if ((rtn = pthread_mutex_lock(&(wpool->input_lock))) != 0)
			Die("pthread_mutex_lock failure: %s\n", strerror(rtn));
		if (! ReadSeq(wpool->sqfp, wpool->sqfp->format, &seq, &sqinfo))
		{	/* we're done. release lock, exit thread */
			if ((rtn = pthread_mutex_unlock(&(wpool->input_lock))) != 0)
				Die("pthread_mutex_unlock failure: %s\n", strerror(rtn));
			FreePlan7Matrix(mx);
			pthread_exit(NULL);
		}
		SQD_DPRINTF1(("a thread is working on %s\n", sqinfo.name));
		wpool->nseq++;
		/* release the lock */
		if ((rtn = pthread_mutex_unlock(&(wpool->input_lock))) != 0)
			Die("pthread_mutex_unlock failure: %s\n", strerror(rtn));

		if (sqinfo.len == 0) continue; /* silent skip of len=0 seqs (wormpep!?!) */

		dsq = DigitizeSequence(seq, sqinfo.len);
		if (wpool->do_xnu) XNU(dsq, sqinfo.len);

		/* 1. Recover a trace by Viterbi.
		 */
		if (P7ViterbiSpaceOK(sqinfo.len, wpool->hmm->M, mx))
			sc = P7Viterbi(dsq, sqinfo.len, wpool->hmm, mx, &tr);
		else
			sc = P7SmallViterbi(dsq, sqinfo.len, wpool->hmm, mx, &tr);

		/* 2. If we're using Forward scores, do another DP
		 *    to get it; else, we already have a Viterbi score
		 *    in sc.
		 */
		if (wpool->do_forward) {
			sc  = P7Forward(dsq, sqinfo.len, wpool->hmm, NULL);
			if (wpool->do_null2) sc -= TraceScoreCorrection(wpool->hmm, tr, dsq);
		}

		/* 3. Save the output in tophits and histogram structures, after acquiring a lock
		 */
		if ((rtn = pthread_mutex_lock(&(wpool->output_lock))) != 0)
			Die("pthread_mutex_lock failure: %s\n", strerror(rtn));
		SQD_DPRINTF1(("seq %s scores %f\n", sqinfo.name, sc));

		pvalue = PValue(wpool->hmm, sc);
		evalue = wpool->thresh->Z ? (double) wpool->thresh->Z * pvalue : (double) wpool->nseq * pvalue;

		if (sc >= wpool->thresh->globT && evalue <= wpool->thresh->globE)
		{
			sc = PostprocessSignificantHit(wpool->ghit, wpool->dhit,
					tr, wpool->hmm, dsq, sqinfo.len,
					sqinfo.name,
					sqinfo.flags & SQINFO_ACC  ? sqinfo.acc  : NULL,
					sqinfo.flags & SQINFO_DESC ? sqinfo.desc : NULL,
					wpool->do_forward, sc,
					wpool->do_null2,
					wpool->thresh,
					FALSE); /* FALSE-> not hmmpfam mode, hmmsearch mode */
		}
		SQD_DPRINTF2(("AddToHistogram: %s\t%f\n", sqinfo.name, sc));
		AddToHistogram(wpool->hist, sc);
		if ((rtn = pthread_mutex_unlock(&(wpool->output_lock))) != 0)
			Die("pthread_mutex_unlock failure: %s\n", strerror(rtn));

		P7FreeTrace(tr);
		FreeSequence(seq, &sqinfo);
		free(dsq);
	} /* end 'infinite' loop over seqs in this thread */
}
#else /*HMMER_THREADS off; no threads support; dummy stub: */
static void
main_loop_threaded(struct plan7_s *hmm, SQFILE *sqfp, struct threshold_s *thresh, int do_forward,
		int do_null2, int do_xnu, int num_threads,
		struct histogram_s *histogram,
		struct tophit_s *ghit, struct tophit_s *dhit, int *ret_nseq)
{
	Die("No threads support");
}
#endif /* HMMER_THREADS */

